# Channels

#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  Paul Francis Harrison
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import os,os.path,string,re,cStringIO,traceback

from settings import *
from utility import *
from error import Error
import hash

class Channels(Task_manager):
    def __init__(self,app):
        Task_manager.__init__(self)
    
        self.app = app
        self.node = app.node

        self.list = get_config("channels", {'#welcome': {'muted': 0}})
        self.ads = {}
        self.subs = {}

    def start(self):
        Task_manager.start(self)

        #self.node.add_handler("channel subscribe query",self)
        for ch in self.list.keys():
            if not self.list[ch]['muted']:
                self.do_sub(ch)

    def stop(self):
        Task_manager.stop(self)

        #self.node.remove_handler("channel subscribe query")

        for ch in self.list.keys():
            if not self.list[ch]['muted']:
                self.do_unsub(ch)

        self.save()

    def save(self):
        set_config("channels", self.list);

    def mute(self, channel, mute=1):
        if self.list[channel]['muted']!=mute:
            self.list[channel]['muted']=mute
            self.app.chat.refresh_userlist()
            if mute:
                self.do_unsub(channel)
            else:
                self.do_sub(channel)

    def sub(self, channel, muted_on_create=0):
        if not self.list.has_key(channel):
            self.list[channel]={'muted': muted_on_create}
            if not muted_on_create:
                self.do_sub(channel)
            self.app.chat.refresh_userlist()

    def unsub(self, channel):
        self.lock.acquire()
        try:
            if self.list.has_key(channel):
                    del self.list[channel]
                    try:
                        self.do_unsub(channel)
                    except Error:
                        pass
                    self.app.chat.refresh_userlist()
            else:
                pass #should report an error here
        finally:
            self.lock.release()

    def set_listening(self, list):
        for item in list:
            self.sub(item)
        for key in self.list.keys():
            self.mute(key, key not in list)

    def get_listening(self):
        list = [ ]
        for item in self.list.keys():
            if not self.list[item]['muted']:
                list.append(item)
        return list

    def is_listening_to(self, channel):
        self.lock.acquire()
        try:
            if not self.list.has_key(channel):
                return 0
            return not self.list[channel]['muted']
        finally:
            self.lock.release()

    def check_join(self, dests):
        for dest in dests:
            if dest[0]=='#':
                self.sub(dest)
                self.mute(dest,0)

    def sub_list_pipe(self, channel):
        return self.node.retrieve(hash.hash_of("channel subscribe "+channel), channel_redundancy,1)
        
        #res = []
        #pipe = self.node.retrieve(hash.hash_of("channel subscribe "+channel), channel_redundancy)
        #try:
        #  while not pipe.finished():
        #    for item in pipe.read_all():
                #  res.append(item)
                #time.sleep(1)
        #finally:
        #  pipe.stop()
        #return res

    def edit_settings(self, channel):
        import gtk

        self.lock.acquire()
        try:
            window = gtk.GtkWindow()
            window.set_default_size(300,10)
            window.set_border_width(10)
            window.set_title(_('Channel: ')+channel)
            vbox = gtk.GtkVBox(0,5)
            window.add(vbox)

            label = gtk.GtkLabel(channel)
            vbox.add(label)

            #label = GtkLabel('(Here should go the subscriber list)')
            label = gtk.GtkLabel('')
            vbox.add(label)

            hbox = gtk.GtkHBox(0,5)

            button = gtk.GtkButton(_("Unsubscribe"))
            def unsub(_b,_ch=channel,_s=self,_w=window):
                _s.unsub(_ch)
                _w.destroy()
            button.connect("clicked",unsub)
            hbox.pack_start(button, 1,1,0)

            button = gtk.GtkButton(_("List subscribers"))
            def list(_b,_ch=channel,_s=self):
                import search
                search.search_for_people_on_channel(_ch, _s.node,_s.app)
            button.connect("clicked",list)
            hbox.pack_start(button, 1,1,0)

            button = gtk.GtkButton(_("Cancel"))
            button.connect("clicked",lambda _b,_window=window:_window.destroy())
            hbox.pack_start(button, 1,1,0)

            vbox.pack_start(hbox, 0,0,0)
            
            show_window(window, _('channel'))

        finally:
            self.lock.release()

    def context_menu(self, channel, event):
        import gtk

	menu = []

        if self.list.has_key(channel):

            if channel in self.app.chat.channel:
                def untalk(data, self=self, channel=channel):
                    if channel in self.app.chat.channel:
                        self.app.chat.channel.remove(channel)
                        self.app.chat.refresh_userlist()
                        self.app.chat.set_normal_prompt()

                menu.append((_("Stop talking on %s")%channel, untalk))
            else:
                if self.list[channel]['muted']:
                    def unmute(data, self=self, channel=channel):
                        self.mute(channel, 0)
                    menu.append((_("Listen to %s")%channel, unmute))
                else:
                    def mute(data, self=self, channel=channel):
                        self.mute(channel, 1)
                    menu.append((_("Mute %s")%channel, mute))

                def talk(data, self=self, channel=channel):
                    if not channel in self.app.chat.channel:
                        self.mute(channel, 0)
                        self.app.chat.channel.append(channel)
                        self.app.chat.refresh_userlist()
                        self.app.chat.set_normal_prompt()
                menu.append((_("Talk on %s")%channel, talk))

            if self.app.chat.channel!=[] and self.app.chat.channel!=[channel]:
                def talkonly(data, self=self, channel=channel):
                    self.mute(channel, 0)
		    self.app.chat.channel=[channel]
		    self.app.chat.refresh_userlist()
		    self.app.chat.set_normal_prompt()
                menu.append((_("Talk only on %s")%channel, talkonly))

            def unsub(data, _s=self, _ch=channel):
                _s.unsub(_ch)
            menu.append((_("Unsubscribe from %s")%channel,unsub))

        else:

            def sub(data, _s=self, _ch=channel):
                _s.sub(_ch)
            menu.append((_("Subscribe to %s")%channel,sub))

        def list(_b,_ch=channel,_s=self):
	    import search
	    search.search_for_people_on_channel(_ch, _s.node,_s.app)
        menu.append((_("List %s subscribers")%channel, list))

	menuwidget = gtk.GtkMenu()
	for (text, action) in menu:
	    mi = gtk.GtkMenuItem(text)
	    mi.connect("activate", action)
	    mi.show()
	    menuwidget.append(mi)
	menuwidget.popup(None,None,None,event.button,event.time)

    def do_sub(self,ch):
        if self.subs.has_key(ch) or self.ads.has_key(ch):
            raise Error(_("erroneous subscription"))
        self.ads[ch]={'type': 'channel exists',
                                    'name': ch,
                                    'keywords':[]}

        #me = self.app.name_server.get_info()
        #self.subs[ch]={'channel': ch}
        #for field in ('type', 'key', 'name'):
        #  self.subs[ch][field]=me[field]
        #del self.subs[ch]['channel']
        #me = self.app.name_server.get_info()
        
        self.subs[ch] = lambda self=self: self.app.name_server.get_info()
        # - has to be a distinct object to allow unpublish
        
        self.node.publish(hash.hash_of('channel exists'),self.ads[ch],
                                            channel_redundancy)
        self.node.publish(hash.hash_of('channel subscribe '+ch),self.subs[ch],
                                            channel_redundancy)

    def do_unsub(self,ch):
        if not self.subs.has_key(ch) or not self.ads.has_key(ch):
            raise Error(_("erroneous unsubscription"))
        self.node.unpublish(self.subs[ch])
        self.node.unpublish(self.ads[ch])
        del self.subs[ch]
        del self.ads[ch]

