# Chat

#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  Paul Francis Harrison
#    Copyright (C) 2002  Monash University
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Yolande:
#  blue is a bad choice for online people - it looks like a clickable link.
#  the arrows aren't an obvious visual clue to their behaviour.  would checkboxes work better?
#  arrow keys are a bit yucky.
#  gossip

# TODO: (njh)
#   - clickable urls in chat window(James Mollison)
#   - ability to remove people from click-list(Michale Kaufman)

# TODO: detect when someone drops off the network badly

# TODO: float pickle

# TODO: improve /recall : select from list, only recall from some people for single message

# TODO: offline message polling, in case of net split
#  !app.chat.retrieve_cached_messages()

# All messages in the chat widget are displayed in "fields",
# eg a field may display a message indicating some task is in progress, 
# then later change to say that it has finished

import os, os.path, string, cStringIO, traceback, select

from types import *

from settings import *
from utility import *
from utility import mainthread_call
import utility                          # migrating to imports
from error import Error
import chat_commands, safe_pickle, hash
import calendar
#import time
#import check


# When publishing timestamps, we standardize on the Unix epoch of 1970.
host2standard_time_offset = calendar.timegm((1970, 1, 1, 0, 0, 0, 3, 1, -1))

if time.gmtime(host2standard_time_offset)[:6] != (1970, 1, 1, 0, 0, 0):
    check.show_bug(_('Bug in calculating epoch offset.'))


def host2standard_timestamp(host_tstamp):
    check.check_has_type(host_tstamp, types.FloatType)
    
    ret = long(host_tstamp + host2standard_time_offset)
    
    check.check_has_type(ret, types.LongType)  #=@E16
    return ret

def standard2host_timestamp(std_tstamp):
    check.check_has_type(std_tstamp, types.LongType)  #=@R36
    
    ret = float(std_tstamp - host2standard_time_offset)
    
    check.check_has_type(ret, types.FloatType)  #=@E17
    return ret


def english_list(list):
    if len(list) == 1:
        return list[0]
    else:
        return string.join(list[:-1],', ')+' and '+list[-1]

def chat_tell_task(chat,dests,message, field, attachment=None):
    std_timestamp = host2standard_timestamp(time.time())
    
    mainthread_call(field.show, _('Sending message...\n'))

    result = ''
    result2 = ''
    old_result2 = ''
    try:
        dest_list = [ ]
        dest_address  = [ ]
        dest_names = [ ]
        dest_bad_names = [ ]
        offline_list  = [ ]
        offline_names = [ ]
        recipient_list = [ ]
        channel_pipes = { }
        
        for dest in dests:
            if dest[0]=='#':
                channel_pipes[dest] = chat.app.channels.sub_list_pipe(dest)
                recipient_list.append((None,None,dest))
                continue
            acq = chat.app.name_server.locate(dest)
            acq.start_watching(chat.node,1)

            acq.lock.acquire()
            online = acq.online
            address = acq.address
            name = acq.name
            username = acq.info['name']
            acq.lock.release()

            if not online:
                offline_list.append(acq)
                offline_names.append(dest)
                recipient_list.append((name,None,username))
                #result = result+dest+ ' is not connected.\n'
            else:
                dest_list.append(acq)
                dest_address.append(address)
                dest_names.append(dest)
                recipient_list.append((name,address,username))
        
        #if result:
        #  mainthread_call(field.show, result)

        my_details = (
            chat.app.name_server.public_key_name,
            chat.app.name_server.name
        )
        if attachment == None:
            sig = chat.app.name_server.sign(message)
            message = ('chat message',my_details,recipient_list,message,sig)
        else:
            sig = chat.app.name_server.sign(message+attachment)
            message = ('chat message',my_details,recipient_list,message,attachment,sig)

        for i in range(len(dest_list)):
            succ = 1
            try:
                ret_value = chat.node.call(dest_address[i],message)
            except Error:
                succ = 0
                result = result + dest_names[i] + _(' could not be contacted.\n')

            if succ:
              try:
                #result = result + 'Message sent to '+ dest_names[i] +'.\n'
                
                #Type checking?
                if type(ret_value) <> TupleType:
                    raise Error(_('unexpected value returned from call'))
                
                if ret_value[0]:
                    if ret_value[1] <> '':
                        result = result + dest_names[i] + ' ' + ret_value[1] + '\n'
                else:
                    if ret_value[1] == '':
                        result = result + dest_names[i] + _(' is quiet.\n')
                    else:
                        result = result + dest_names[i] + _(' is quiet: ') + ret_value[1] + '\n'
              except Error:
                succ = 0
                result = result + dest_names[i] + _(' sent bad reply.\n')
            if not succ:
                offline_list.append(dest_list[i])
                offline_names.append(dest_names[i])
                dest_bad_names.append(dest_names[i])

        for name in dest_bad_names:
            dest_names.remove(name)

        #prev_result = result
        if dest_names:
            result = _('Message sent to %s.')%english_list(dest_names) + '\n' + result

        mainthread_call(field.show, result + _('Sending message...\n'))
            
        #for pipe in channel_pipe_list:
        #  chcount=0

        chan_cnt = { }
        for key in channel_pipes.keys():
            chan_cnt[key] = 0

        while channel_pipes:
            for chan_name,chan_pipe in channel_pipes.items():
                if chan_pipe.finished():
                    chan_pipe.stop()
                    del channel_pipes[chan_name]
                    continue
                
                def chan_task(chat,address,chan_name, message=message, chan_cnt=chan_cnt):
                    try:
                        ret_value = chat.node.call(address,message)
                        
                        #Type checking?
                        if type(ret_value) <> TupleType:
                            raise Error(_('unexpected value returned from call'))

                        chat.lock.acquire()
                        chan_cnt[chan_name] += 1
                        chat.lock.release()

                    except Error:
                        pass

                for address in chan_pipe.read_all():
                    if address in dest_address:
                        continue
                    dest_address.append(address)
                    Task(chan_task,chat,address,chan_name).start()
                                        
            time.sleep(0.1)
            result2 = ''
            for chan_name,cnt in chan_cnt.items():
                if chan_name in channel_pipes.keys():
                    ellipsis='...'
                else:
                    ellipsis='.'
                if cnt == 1:
                    result2 += (_('Message sent to %d person on channel %s%s\n') % (cnt,chan_name,ellipsis))
                else:
                    result2 += (_('Message sent to %d people on channel %s%s\n') % (cnt,chan_name,ellipsis))
            if result2!= old_result2:
                mainthread_call(field.show, result+result2)
                old_result2 = result2

        result = result+ result2 # huh?
        mainthread_call(field.show, result+result2)


            #dest_names.append(ch + ' (' + ("%d"%chcount) + ')')
            #result = 'Message sent to ' + english_list(dest_names) + '.\n' + prev_result
            #mainthread_call(field.show, result+'Sending message...\n')


        recall_list = [ ]
        for i in range(len(offline_list)):
            key   = random_bytes(name_bytes)
            lock  = hash.hash_of(key)
            crypt = offline_list[i].encrypt((key, message, std_timestamp))
            data  = {
                'type' : 'offline message',
                'crypt': crypt
            }
            name  = hash.hash_of('offline message '+offline_list[i].name)

            recall_list.append((offline_list[i].nickname,name,data,key))

            redundancy = chat.app.cache.publish([name],data,lock)

            if redundancy == cache_redundancy:
                str = offline_names[i] + _(' will get your message when they log on.\n')
            elif redundancy == 0:
                str = _('Could not store message for ') + offline_names[i] + '!\n'
            else:
                str = offline_names[i] + _(' will probably get your message when they log on:\n' +\
                    '  Message only stored with redundancy %d.\n')%redundancy

            result = result + str
            mainthread_call(field.show, result + '\n')
        else:
            mainthread_call(field.show, result + '\n')
            
        chat.recall_list = recall_list
    except Error, error:
        mainthread_call(field.show,
                        _('Error sending message:\n') + error.message + '\n')

    mainthread_call(field.close)


# (If you change these then search for uses and update the
# proofs, or at least add a todo note.)
incoming_message_history_item_tmpl = ('any', types.FloatType)
unread_message_list_item_tmpl = ((types.StringType,),
                                 'opt-address',
                                 types.FloatType)


def chat_receive_task(chat_obj, request,address, field1,quiet,send_time, add_to_history=1):
    """send_time should be as if from time.time() (i.e. a float expressing seconds
       since the local machine's epoch) rather than the standardized timestamps
       sent over the network (i.e. a long expressing seconds since the Unix epoch
       of 1970)."""

    check.check_is_opt_address(address)  #=@R33  fixme callers
    check.check_has_type(send_time, types.FloatType)  #=@R35  fixme callers
    # fixme: should probably require that request be a tuple with 5+ elems.
    
    #TODO: search for people not in acq list
    try:
        if not chat_obj.widget_visible:
            chat_obj.n_unseen_messages = chat_obj.n_unseen_messages + 1
            if chat_obj.n_unseen_messages == 1:
                title = _('1 unseen message')
            else:
                title = _('%d unseen messages') % chat_obj.n_unseen_messages
            chat_obj.app.set_title(title)
    
        has_attachment = (len(request) == 6)


        if type(request[1]) == type(()):
            sender_key_name = request[1][0]
            sender_name = request[1][1]
        else:
            # Used not to send username with key name
            sender_key_name = request[1]
            sender_name = 'unknown'

        try:
            sender = chat_obj.app.name_server.identify(sender_key_name, address)
            # Proof of @R32: @R33, not reassigned since then, deeply immutable
            # from @E21.
        except Error:
            sender = None

        if sender:
            if has_attachment:
                sender.verify(request[3]+request[4],request[5])
            else:
                sender.verify(request[3],request[4])

        mainthread_call(field1.show,'[','blue')
        if sender:
            sender_nickname = sender.nickname
            field_sender = field1.child(name_field_left_click,name_field_right_click)
            mainthread_call(field_sender.show,sender_nickname,'blue')
        else:
            sender_nickname = '<'+sender_name+'>'
            field_sender = field1.child()
            mainthread_call(field_sender.show,sender_nickname)

        mainthread_call(field1.close)
        mainthread_call(field_sender.close)
        field_middle = field_sender.child()
        mainthread_call(field_middle.show,' >>','blue')

        field_right = field_middle.child()
        mainthread_call(field_right.show,' ...]','blue')

        field = field_right.child()
        text=' '+request[3]
        while text:
            list = [ text.find('http:'), text.find('ftp:'), text.find('circle-file:') ]
            while list.count(-1):
                list.remove(-1)
            if list:
                list.sort()
                l = list[0]
                if text[l:l+4]=='http':
                    prefix = 'http:'
                elif text[l:l+3]=='ftp':
                    prefix = 'ftp:'
                else:
                    prefix = 'circle-file:'
                
                n = min
                
                s,ss = text.split(prefix,1)
                sss = prefix+ss
                if sss.find(' ')<>-1:
                    url,text = sss.split(' ',1)
                    text = ' '+text
                else:
                    url = sss
                    text = ''
                mainthread_call(field.show,s,'black')
                field_url = field.child()
                mainthread_call(field.close)
                field_url.set_action(url_field_left_click)
                mainthread_call(field_url.show,url,'blue')
                field = field_url.child()
                mainthread_call(field_url.close)
            else:
                mainthread_call(field.show,text,'black')
                break
                
        field3 = field.child()
        mainthread_call(field.close)
                        
        if has_attachment:
            attachment = safe_pickle.loads(request[4])
            chat_obj.lock.acquire()
            i = 1
            while chat_obj.exec_vars.has_key("%s%d" % (sender_nickname, i)):
                i = i + 1
            chat_obj.exec_vars["%s%d"        % (sender_nickname, i)] = attachment
            chat_obj.exec_vars["%s%d_source" % (sender_nickname, i)] = address
            chat_obj.lock.release()
            if type(attachment) == type({}) and attachment.get('type',None) == 'file':
                field3_extra = _('\nContains attachment; type `!%s%d\' to see its details, `/take %s%d\' to retrieve.') % (sender_nickname,i,sender_nickname,i)
            else:
                field3_extra = _('\nContains attached python object; type `!%s%d\' to see.') % (sender_nickname,i)
        else:
            field3_extra = ''

        now_date  = time.strftime(_('%A %B %d, %Y'),time.localtime(time.time()))
        then_date = time.strftime(_('%A %B %d, %Y'),time.localtime(send_time))

        if now_date <> then_date:
            date_str = '\n'+then_date+' '
        else:
            date_str = '  '
        date_str = date_str + time.strftime(_('%I:%M %p'),time.localtime(send_time))

        mainthread_call(field3.show, date_str + field3_extra + '\n\n', 'light_grey') 

        if not quiet and chat_obj.app.config['beep_on_message']:
            mainthread_call(utility.beep)

        previous_field = field_middle
        for item in request[2]:
            if not check.matches(item, ('any', 'opt-address')):
                sys.stderr.write('chat_receive_task passed bad item in request[2]: '
                                 + `item` + '.  Ignoring this item.\n')
                continue

            field_recipient = previous_field.child()
            previous_field = field_recipient
            recipient_key_name,recip_address = item[0],item[1]
            if recipient_key_name==None and len(item)>=3 and item[2][0]=='#':
                if chat_obj.app.config['auto_list_channels']:  # kc5tja, 22jul2002
                    chat_obj.app.channels.sub(item[2], 1)
                
                field_recipient.set_action(channel_field_left_click,channel_field_right_click)
                mainthread_call(field_recipient.show,' '+item[2],'blue') 
                mainthread_call(field_recipient.close)
                continue

            try:
                recipient = chat_obj.app.name_server.identify(recipient_key_name,
                                                              recip_address)
                # Proof of @R32: recip_address is assigned from item[1] above.
                # item was checked above to be a tuple whose second elem is
                # an opt-address.  Neither item nor recip_address are
                # reassigned.  Deep immutability from @E21.
                
                #recipient.lock.acquire()
                #chat_obj.refresh_user(recipient.nickname)
                #recipient.lock.release()

                field_recipient.set_action(name_field_left_click,name_field_right_click)
                mainthread_call(field_recipient.show,' ' + recipient.nickname,'blue') 
                mainthread_call(field_recipient.close)

            except Error,error :
                if len(item) >= 3:
                    str = ' <'+item[2]+'>'
                else:
                    str = ' ?'
                mainthread_call(field_recipient.show,str) 
                mainthread_call(field_recipient.close)

            #Avoid excessive updating
            #mainthread_call(field1.show, str+' ...]', 'blue')
        mainthread_call(field_right.show, ']','blue')
    except Error:
        mainthread_call(chat_obj.show,
                        _('Bad message received from ') + repr(address) + '\n')

    mainthread_call(field_middle.close) 
    mainthread_call(field_right.close)
    mainthread_call(field3.close)

    if add_to_history:
        chat_obj.lock.acquire()
        try:
            new_item = (request, send_time)
            check.check_matches(new_item, incoming_message_history_item_tmpl)
            # Proof: @R35, immutability of float, send_time not reassigned
            # in this function.
            # Relevance: @I16.
            chat_obj.incoming_message_history.append(new_item)

            while len(chat_obj.incoming_message_history) > incoming_message_history_size:
                del chat_obj.incoming_message_history[0]
        finally:
            chat_obj.lock.release()
            
class Chat_field_text_group:
    def __init__(self, chat):
        self.chat = chat
        self.members = [ ]
        self.n_open = 0

    def add(self, member,after=None):
        if after:
            self.members.insert(self.members.index(after)+1,member)
        else:
            self.members.append(member)
        self.n_open = self.n_open + 1

    def closure(self):
        self.n_open = self.n_open - 1

        if not self.n_open:
            self.chat.text_hide_command_line()
            for member in self.members:
                sys.stdout.write(member.text)
            sys.stdout.flush()
            self.chat.text_show_command_line()
            self.chat = None

class Chat_field_text:
    def __init__(self,chat=None,group=None,after=None):
        if not group:
            group = Chat_field_text_group(chat)
        self.text = ''
        self.group = group
        group.add(self,after)
        
    def child(self,left_action=None,right_action=None):
        # Only function that can be called from other thread
        return Chat_field_text(group=self.group,after=self)

    def show(self,str,color='grey'):
        self.text = str

    def close(self):
        self.group.closure()

    def set_action(self,left_action=None,right_action=None):
        pass

class Chat_field_gtk:
    def __init__(self,chat, start=None,end=None,index=None,left_action=None,right_action=None):
        chat.lock.acquire()
        self.left_action = left_action
        self.right_action = right_action
        self.text=''
        try:
            if start:
                self.start = start
                self.end   = end
            else:
                self.start = chat.insertion_point
                self.end   = chat.insertion_point
            self.chat  = chat

            if index:
                chat.fields.insert(index,self)
            else:
                chat.fields.append(self)
        finally:
            chat.lock.release()

    def set_action(self,left_action=None,right_action=None):
        self.left_action =left_action
        self.right_action=right_action

    def on_click(self,w,e):
        if e.button == 1 and self.left_action<>None:
            self.left_action(self,self.chat.app,w,e)
        elif e.button == 3 and self.right_action<>None:
            self.right_action(self,self.chat.app,w,e)

    def child(self,left_action=None,right_action=None):
        # Only function that can be called from other thread
        # FIXME: self.chat can be None.
        self.chat.lock.acquire()
        try:
            return Chat_field_gtk(self.chat,self.end,self.end,self.chat.fields.index(self)+1,\
                                  left_action=left_action, right_action=right_action)
        finally:
            self.chat.lock.release()

    def show(self,str,color='grey'):
        import gtk
        import widgets

        self.text = str
        self.chat.lock.acquire()
        try:
            if self.chat.widget.flags(gtk.DESTROYED):
                return

            if not self in self.chat.fields:
                return

            pos = self.chat.widget.get_position()

            self.chat.widget.set_point(self.start)

            self.chat.widget.freeze()
            #self.chat.widget.insert_defaults(str)
            
            #if bold:
                #font = self.chat.bold_font
            #else:
                #font = self.chat.normal_font
            
            self.chat.widget.forward_delete(self.end-self.start)
            
            style = self.chat.widget.get_style()
            self.chat.widget.insert(
                #font,
                self.chat.font,
                #style.fg[STATE_NORMAL],
                getattr(widgets,color),
                style.base[gtk.STATE_NORMAL],
                str)
            
            self.chat.widget.thaw()

            change = len(str)-(self.end-self.start)
            
            pos = pos + change
            self.chat.horizon         = self.chat.horizon         + change
            self.chat.insertion_point = self.chat.insertion_point + change

            index = self.chat.fields.index(self)

            for item in self.chat.fields[index+1:]:
                item.start = item.start + change
                item.end   = item.end   + change
        
            self.chat.widget.set_position(pos)

            self.end = self.end + change
        finally:
            self.chat.lock.release()


    def close(self):
        #do not close clickable fields...
        if self.left_action<>None or self.right_action<>None:
            return
        
        if self in self.chat.fields:
            self.chat.fields.remove(self)
        self.chat = None


class Chat(Task_manager):
    """ Base class of Chat_gtk and Chat_text, provides basic chat functions. """

    def __init__(self,app, use_gtk=1):
        Task_manager.__init__(self)
    
        self.node = app.node
        self.app  = app
        self.use_gtk = use_gtk

        # Visible status
        self.quiet       = 0
        self.activity    = ''

        # Detect repeated offline messages
        self.offline_message_buffer = [ ]

        # Store a copy of recent messages
        self.incoming_message_history = [ ]

        # While quiet, messages are buffered for later display.
        # Tuples of form (request, opt-address, time.time()).
        # address for offline-sent messages is None.
        # The time item must be wrt this client's epoch.
        self.unread_message_list = [ ]
        
        config = utility.get_checked_config("chat", types.DictionaryType, { })
        for name,tmpl in [('quiet', 'any'),
                          ('activity', 'any'),
                          ('offline_message_buffer', 'any'),
                          ('incoming_message_history', [('any', 'any timestamp')]),
                          ('unread_message_list', [((types.StringType,),
                                                    'opt-address',
                                                    'any timestamp')])]:
            if config.has_key(name):
                config_val = config[name]
                if check.matches(config_val, tmpl):
                    setattr(self, name, config_val)
                else:
                    print (_("Warning: ignoring chat config item %s, which doesn't match %s.")
                           % (name, tmpl))
        # The current version of Circle always writes these timestamps in
        # standardized form, but previous versions wrote a mixture of floats
        # and longs.
        for i in xrange(len(self.incoming_message_history)):
            request,tstamp = self.incoming_message_history[i]
            if type(tstamp) == types.LongType:
                self.incoming_message_history[i] = (request,
                                                    standard2host_timestamp(tstamp))
            check.check_matches(self.incoming_message_history[i],
                                incoming_message_history_item_tmpl)
            # Loose proof: definition of check.is_any_timestamp, @E17.
            # Relevance: @R.I16.
        for i in xrange(len(self.unread_message_list)):
            request,addr,tstamp = self.unread_message_list[i]
            if type(tstamp) == types.LongType:
                self.unread_message_list[i] = (request, addr,
                                               standard2host_timestamp(tstamp))
            check.check_matches(self.unread_message_list[i],
                                unread_message_list_item_tmpl)
            # Loose proof: as above.
            # Relevance: @R.I15.
        
        self.recall_list = [ ]

        # Detect repeated messages
        self.message_buffer = [ ]

        self.channel = [ ]

        self.history = [ ] # Commands
        self.history_position = 0

        # Results of /find
        self.search_list = [ ]

        # Python environment
        self.exec_vars = { 'app' : app, 'node' : app.node }

        # Command registry
        self._commands = { }  # name -> (min n. params, max n. params, function)

        chat_commands.register(self)

        self.check_invar()
    
    def check_invar(self):
        check.check_matches(self.unread_message_list,
                            [unread_message_list_item_tmpl])  #=@I15
        check.check_matches(self.incoming_message_history,
                            [incoming_message_history_item_tmpl])  #=@I16
    
    def repeated(self, address,call_id):
        if (address,call_id) in self.message_buffer:
            return 1
        self.message_buffer.append((address,call_id))
        if len(self.message_buffer) > 50:
            del self.message_buffer[0]
        return 0

    def start(self):
        Task_manager.start(self)
        
        self.app.name_server.set_status('chat',(not self.quiet,self.activity))
        
        self.set_normal_prompt()

        self.node.add_handler('chat message',self, (('any', # ns.public_key_name
                                                     'any'), # ns.name
                                                    [('any', 'any', 'string')],
                                                    'any', # message
                                                    'any')) # attachment or sig
        self.node.add_handler('chat look', self)
        # The expected call/return signatures are () / (types.IntType, 'text')
        # respectively.  But don't bother checking return type, as callers
        # already do their own checking.
        
        self.app.name_server.add_watch_callback(self)

        self.refresh_userlist()

        if self.app.gossip:
            def on_click(field,app,widget,event):
                mainthread_call(field.show,field.text,'purple')
                app.gossip.show_gossip()
                
            title_field = self.get_field() 
            title_field.show('Fetching gossip...\n')
            gossip_fields = []
            for i in range(5):
                field = self.get_field()
                field.show('\n')
                gossip_fields.append(field)
            def on_gossiped(title_field=title_field,gossip_fields=gossip_fields,self=self,gossip=self.app.gossip,on_click=on_click):
                gossip.lock.acquire()
                n = 0
                for item in gossip.sorted_wodges()[:100]:
                    if not item[1].collapsed:
                        gossip_fields[n].set_action(on_click)
                        mainthread_call(gossip_fields[n].show,'   ' + item[1].wodge['subject'] + '\n','blue')
                        n = n + 1
                        if n >= 5: 
                            break
                for i in range(5-n):
                    mainthread_call(gossip_fields[4-i].show,'')
                    mainthread_call(gossip_fields[4-i].close)
                
                gossip.lock.release()

                if n == 0:
                    str = ''
                else:
                    str="Today's hot gossip:\n"
                mainthread_call(title_field.show, str)
                mainthread_call(title_field.close)

            self.app.gossip.request_update(on_gossiped)

        self.retrieve_cached_messages()

        if self.quiet or self.activity <> '':
            self.show('\n')
            self.show_status()

    def stop(self):
        Task_manager.stop(self)
        self.change_activity('removing callbacks')
        self.node.remove_handler('chat message')
        self.node.remove_handler('chat look')
        self.app.name_server.remove_watch_callback(self)

        self.change_activity('preparing config')
        config = { }
        for item in ['quiet','activity','offline_message_buffer']:
            config[item] = getattr(self,item)
        # Convert timestamps to standardized epoch.  The intent is to allow
        # sharing/copying a .circle directory between different operating
        # systems (in particular, between operating systems that have a
        # value of time.gmtime(0), i.e. different epochs).
        config['incoming_message_history'] \
          = map(lambda item: (item[0], host2standard_timestamp(item[1])),
                self.incoming_message_history)
        config['unread_message_list'] \
          = map(lambda item: (item[0], item[1], host2standard_timestamp(item[2])),
                self.unread_message_list)

        self.change_activity('saving to disk')
        set_config("chat",config)
        self.change_activity('')

    def handle_watch(self, acq, because_of, **extras):
        acq.lock.acquire()
        try:
            nickname = acq.nickname
            watch    = acq.watch
        finally:
            acq.lock.release()

        if because_of == 'edit':
            self.lock.acquire()
            try:
                new_channel = [ ]
                for item in self.channel:
                    if item[:1] == '#' or \
                         self.app.name_server.nicknames.has_key(item):
                        new_channel.append(item)
            finally:
                self.lock.release()

            self.set_channel(new_channel)

        elif acq <> self.app.name_server.me:
            self.refresh_userlist()
            
        if not watch or (because_of <> 'connect' and because_of <> 'disconnect'):
            return

        self.lock.acquire()
        try:
            if not self.quiet:
                if because_of == 'connect':
                    action = _(' just arrived.\n')
                else:
                    if extras['message']:
                        action = _(' just left: ') + extras['message'] + '\n'
                    else:
                        action = _(' just left.\n')
                mainthread_call(self.show,
                                time.strftime(_('%I:%M %p') + ' ',
                                              time.localtime(time.time())))
                mainthread_call(self.show,nickname,'blue',name_field_left_click,name_field_right_click)
                mainthread_call(self.show, action,'blue')
        finally:
            self.lock.release()

    def handle(self, request, address,call_id):
        check.check_matches(request, (types.StringType,))  #=@R34
        check.check_is_af_inet_address(address)  #=@R31
        
        self.lock.acquire()
        try:
            if request[0] == 'chat look':
                return (not self.quiet, self.activity)
            if request[0] == 'chat message':
                for recipient in request[2]:
                    if recipient[0] == self.app.name_server.public_key_name:
                        break
                    if recipient[2][:1] == '#' and \
                         self.app.channels.is_listening_to(recipient[2]) and \
                         request[1][0] <> self.app.name_server.public_key_name:
                        break
                else:
                    return None
            
                if not self.repeated(address,call_id):
                    if self.quiet:
                        new_item = (request, address, time.time())
                        check.check_matches(new_item,
                                            unread_message_list_item_tmpl)
                        # Proof: @R34,@R31, and not having modified request or
                        # address, and the relevant types being immutable
                        # (partly by @E11).  time.time() returns float.
                        # Relevance: @R.I15.
                        self.unread_message_list.append(new_item)
                        mainthread_call(self.set_normal_prompt)
                        del new_item
                    else:
                        Task(chat_receive_task,self,request,address, self.get_field(),self.quiet,time.time()).start()
                        # Proof of @R33: @R31; deep immutability by @E11.
                        # Proof of @R35: time.time() returns float.

                return (not self.quiet, self.activity)

        finally:
            self.lock.release()

    def refresh_userlist(self):
        pass

    def set_prompt(self, str):
        if not self.use_gtk:
            self.lock.acquire()
            try:
                if str == self.prompt:
                    return

                self.text_hide_command_line()
                self.prompt = str
                self.text_show_command_line()
            finally:
                self.lock.release()
            return

        # Kind of hacked...
        if self.quiet:
            color = 'green'
        else:
            color = 'blue'

        self.lock.acquire()
        try:
            insertion_point = self.insertion_point
            field = Chat_field_gtk(self, insertion_point,self.horizon)
            field.show(str,color)
            field.start = field.end
            field.show(' ','black')
            field.close()
            self.insertion_point = insertion_point
            self.horizon = field.end
        finally:
            self.lock.release()
        
    def set_normal_prompt(self):
        if self.shell:
            str = ''
        elif not self.use_gtk and self.command_text_lines:
            str = '>'
        else:
            if not self.channel:
                str = '['+self.app.config['name']+']'
            else:
                str = '['+self.app.config['name']+' >> '+string.join(self.channel,' ')+']'
                self.set_prompt(str)

            l = len(self.unread_message_list)
            if l:
                str = str + _(' (%d unread)')%l

        if not self.use_gtk:
            status = get_config('text_status',{ })
            status['unread'] = len(self.unread_message_list)
            set_config('text_status',status)

        self.set_prompt(str)

    def get_field(self):
        raise Error("define in subclass")
    
    def show(self, str,color='grey',left_action=None,right_action=None):
        field = self.get_field(left_action=left_action,right_action=right_action)
        field.show(str,color)
        field.close()
        #return field

    def retrieve_cached_messages(self):
        def retr_msg_task(self):
            field = self.get_field()
            mainthread_call(field.show,
                            _('Retrieving any messages sent while you were offline...\n'))
        
            pipe = self.node.retrieve(hash.hash_of('offline message '+self.app.name_server.public_key_name),cache_redundancy)
            pipe_reads = [ ]
            if not pipe.finished():
                while 1:
                    pipe_reads.extend(pipe.read_all())
                    if pipe.finished():
                        break
                    time.sleep(1)
            pipe.stop()
            # pjm 2002-08-05: I've changed the above to sleep only if
            # one read_all call isn't enough.  However, I don't know why
            # sleep is wanted in the first place, or whether a different
            # duration might be better.  (Python sleep allows its
            # argument to be fractional, implemented in terms of
            # select.)
            
            unique_messages = [ ]
            for item in pipe_reads:
                # I think we're guaranteed that item matches ('af_inet_address', 'any').
                if not check.matches(item[1], {'type' : types.StringType,
                                               'crypt' : ('any', 'any')}):
                    # print bad peer
                    pass
                elif item[1] not in unique_messages:
                    unique_messages.append(item[1])

            message_list = [ ]
            for raw_msg in unique_messages:
                    if type(raw_msg) == type({}) \
                       and raw_msg.get('type','') == 'offline message' \
                       and raw_msg.has_key('crypt'):
                        try:
                            decrypt = self.app.name_server.decrypt(raw_msg['crypt'])
                            if not check.matches(decrypt, ('any',
                                                           (types.StringType,),
                                                           types.LongType)):
                                raise Error('bad decrypted reply')
                            
                            # Remove from caches
                            for thing in pipe_reads:
                                if thing[1] == raw_msg:
                                    try:
                                        self.node.call(thing[0],('data cache remove',decrypt[0]))
                                    except Error:
                                        pass
                                    
                            message_list.append((decrypt[2],decrypt[1]))
                        except Error:
                            pass

            message_list.sort()
            self.lock.acquire()
            try:
                any = 0
                for msg in message_list:
                    if msg not in self.offline_message_buffer:
                        self.offline_message_buffer = [msg] + self.offline_message_buffer[:50]
                        #Task(chat_receive_task,self,msg[1],None, self.get_field(),1,msg[0]).start()
                        new_item = (msg[1], None,
                                    standard2host_timestamp(msg[0]))
                        # Proof of @R36: msg is taken from message_list.
                        # message_list is local to this method, and is not
                        # passed to any other method (so is not shared with any
                        # other thread).  message_list starts as empty and is
                        # written to solely as (decrypt[2],decrypt[1]) pairs,
                        # and only where decrypt has already been found to
                        # match ('any', ('string',), 'long').  The relevant
                        # types are immutable.
                        check.check_matches(new_item,
                                            unread_message_list_item_tmpl)
                        # Proof: the @R36 proof just above also shows that
                        # msg[1] (i.e. decrypt[1]) matches ('string',) and
                        # is immutable.  new_item[1] matches because
                        # is_opt_address(None).  new_item[2] matches from @E17.
                        # Relevance: @R.I15
                        self.unread_message_list.append(new_item)
                        any = 1
                    else:
                        print _("Duplicate offline message.")
            finally:
                self.lock.release()

            if any:
                mainthread_call(field.show,
                                _('\nYou received messages while you were offline. Type /read to read them.\n'))
                mainthread_call(self.set_normal_prompt)
            else:
                mainthread_call(field.show, '\n')
            mainthread_call(field.close)

        Task(retr_msg_task, self).start()

    def do_who(self):
        def who_task(self, acq, field, field_dots, n):
            acq.lock.acquire()
            name = acq.nickname
            acq.lock.release()

            acq.start_watching(self.node, 1)

            acq.lock.acquire()
            online = acq.online
            address = acq.address
            acq.lock.release()

            if online:
                mainthread_call(field.show, name+' ')
                try:
                    result = self.node.call(address,('chat look',))
                    if not check.matches(result, (types.IntType, 'text')):
                        mainthread_call(field.show, '?'+name+' ')
                    elif not result[0]:
                        mainthread_call(field.show, '*'+name+' ')
                except Error:
                    pass
            else:
                mainthread_call(field.show, '')
            mainthread_call(field.close)

            self.lock.acquire()
            n[0] = n[0] - 1
            if n[0] == 0:
                mainthread_call(field_dots.show, '\n')
                mainthread_call(field_dots.close)
            self.lock.release()

        field = self.get_field()
        field.show(_('Watched people currently online:\n'))
        new_field = field.child()
        field.close()
        field = new_field

        self.app.name_server.lock.acquire()
        values = self.app.name_server.acquaintances.values()
        self.app.name_server.lock.release()
        
        n = [0]
        fields = { }
        for item in values:
            if item.watch:
                n[0] = n[0]+1
                fields[item], field = field, field.child()

        if n[0]:
            field.show(' ...\n')
            for item in fields.keys():
                Task(who_task, self, item, fields[item], field, n).start()
        else:
            field.show('\n')
            field.close()

    def show_status(self):
        if self.quiet:
            if self.activity == '':
                self.show(_('You are quiet.\n'))
            else:
                self.show(_('You are quiet: ')+self.activity+'\n')
        else:
            if self.activity == '':
                self.show(_('You are listening.\n'))
            else:
                self.show(_('What you are doing: %s')%self.activity+'\n')

        if self.unread_message_list:
            self.show(_('You have unread messages. Type /read to read them.\n'))

    def set_channel(self, dests):
        self.lock.acquire()
        old_channel = self.channel
        self.channel = dests

        self.app.channels.check_join(dests)

        #for item in old_channel:
        #  self.refresh_user(item,0)
        #for item in self.channel:
        #  self.refresh_user(item,1)
        self.refresh_userlist()
        self.set_normal_prompt()
        self.lock.release()

    def set_status(self, quiet,activity):
        self.lock.acquire()
        self.quiet = quiet
        self.activity = activity
        self.lock.release()
        
        self.app.name_server.set_status('chat',(not quiet,activity))

    def send_message(self, dests, message, attachment=None):
        if self.quiet or self.activity <> '':
            self.show_status()
        Task(chat_tell_task,self,dests,message,self.get_field(),attachment).start()

    def register(self, name, min_param, max_param, function):
        """ Register a command. 
        
                A command takes a certain number of parameters separated by spaces,
                up to max_param, and not less than min_param. The last parameter may
                contain spaces.

                function may assert that its argument matches ['text'] (i.e. is
                a list of items each of type string or unicode) and that the
                length of this list is in the range [min_param, max_param].

                E.g.:
                
                def mycommand(chat, params):
                    check.check_matches(params, ['text'])
                    check.check_assertion(1 <= len(params) <= 2)
                    
                    str = 'You typed ' + params[0]
                    if len(params) == 2:
                        str += '; ' + params[1]
                    chat.show(str + '\n')
                app.chat.register('mycommand', 1, 2, mycommand)
                
        """
        check.check_has_type(name, types.StringType)
        check.check_has_type(min_param, types.IntType)
        check.check_has_type(max_param, types.IntType)
        check.check_is_callable(function)
        check.check_assertion(0 <= min_param)
        check.check_assertion(min_param <= max_param)
        
        self.lock.acquire()
        self._commands[name] = (min_param, max_param, function)
        self.lock.release()
        
    def do(self, str):
        """ Execute a chat command. """
        str = string.strip(str)

        #chat_commands.do(self, str)
        try:
            if not str:
                pass
		
            elif str[:2] == '//':
                self.run_shell(str[2:])

            elif str[0] == '/':
                # pjm: I'm not sure what we should do if str is just "/" or
                # is "/ blah", and whether the behaviour should depend
                # on whether "/blah" would be a valid command.
                # Usually the user would mean "/say /" (or "/say / blah").
                # Versions of circle up to 0.31 implicitly treat "/ blah"
                # the same as "/blah" (due to the behaviour of string.split).
                if ((str == '/') or (str[1] in string.whitespace)):
                    list = ['', string.lstrip(str[2:])]
                else:
                    list = string.split(str[1:], maxsplit=1)
                command = self._commands.get(list[0])
                if command is None:
                    raise Error(_("Unknown command /%s.\n(Did you mean `/say %s...' ?)\nType /help for help on available commands.") % (list[0], str[:8]))

                if len(list) == 2:
                    remainder = list[1]
                else:
                    remainder = ''

                param = string.split(remainder,maxsplit=command[1]-1)

                if len(param) < command[0]:
                    raise Error(_('/%s requires at least %d parameters.\nType /help for help on command usage.')%(list[0],command[0]))

                check.check_matches(param, ['text'])
                check.check_assertion(command[0] <= len(param) <= command[1])
                apply(command[2],(self,param))
            
            elif str[0] == '!':
                file = cStringIO.StringIO()
                stdout = sys.stdout
                stderr = sys.stderr
                sys.stdout = file
                sys.stderr = file
                try:
                    try:
                        result = eval(str[1:],self.exec_vars)
                        if result is not None:
                            print `result`
                    except SyntaxError:
                        exec str[1:]+'\n' in self.exec_vars
                except:
                    traceback.print_exc()
                    if len(str) > 10:
                        short_str = str[:7] + '...'
                    else:
                        short_str = str
                    print _("(Perhaps you meant `/say %s' ?)" % short_str)
                sys.stdout = stdout
                sys.stderr = stderr
                self.show(file.getvalue())
                
            elif self.channel == [ ]:
                #self.show("Use /join to specify someone to talk to.\n")
                self.show("Before sending messages, you need to select a person or channel to talk to in the list to the right. If the person is not in the list, you need to search for them in the search box above and add them to your list of contacts.\n")

            else:
                self.send_message(self.channel[:],str)
        
        except Error, error:
            self.show(error.message+'\n')
        
        self.set_normal_prompt()
    
    def handle_user_entry(self, str):
        """ Handler user typing some text. """
        self.lock.acquire()
        try:
            if self.shell:
                self.set_prompt('')

                str = str + '\n'
                try:
                    while str:
                        n = os.write(self.shell_master,str)
                        str = str[n:]
                except OSError:
                    pass
                return
        finally:
            self.lock.release()

        #try:
        if str:
            self.history.append(str)
            self.history_position = len(self.history)

        self.do(str)

        #finally: 
        #  self.set_normal_prompt()

class Chat_text(Chat):
    """ Text mode chat interface. """
    
    def __init__(self,app):
        Chat.__init__(self,app,0)

        self.prompt = ''
        self.command_text_lines = [ ]
        self.command_text = ''
        self.input_queue = ''
        self.escape_queue = [ ]
        self.command_visible = 0
        
        self.widget_visible = 1

        self.shell = None

    def stop(self):
        # main thread only!
        stdin_echo(0)
        self.text_hide_command_line()
        stdin_echo(1)

        Chat.stop(self)

    def run_shell(self, command):
        self.text_hide_command_line()
        stdin_echo(1)
        os.system(command)
        stdin_echo(0)
        self.text_show_command_line()

    def get_field(self,left_action=None,right_action=None):
        return Chat_field_text(self)

    def text_read_input(self):
        char = os.read(0,1)
        self.input_queue = self.input_queue + char

        if (char >= 'a' and char <= 'z') or \
             (char >= 'A' and char <= 'Z'):
            pos = string.find(self.input_queue,chr(27)+'[')
            if pos <> -1:
                self.escape_queue.append(self.input_queue[pos+2:])
                self.input_queue = self.input_queue[:pos]

    def text_get_escape(self,code):
        while 1:
            for i in range(len(self.escape_queue)):
                if self.escape_queue[i][-1] == code:
                    return self.escape_queue.pop(i)
            self.text_read_input()

    def text_backspace(self, n):
        while n:
            sys.stdout.write(chr(27)+'[6n')
            sys.stdout.flush()

            escape = self.text_get_escape('R')

            try:
                column = int(string.split(escape[:-1],';')[1]) - 1
            except:
                print _('Badly formatted escape sequence '),escape
                return

            if column == 0:
                sys.stdout.write(chr(27)+'[A'+chr(27)+'[999C'+chr(27)+'[K')
                sys.stdout.flush()
                n = n - 1
                continue
            
            if n < column:
                move = n
            else:
                move = column
            sys.stdout.write(chr(27)+'[%dD'%move+chr(27)+'[K')
            sys.stdout.flush()

            n = n - move

    def text_hide_command_line(self):
        #sys.stdout.write( (chr(8)+' '+chr(8))*(len(self.prompt)+len(self.command_text)) )
        #sys.stdout.flush()
        self.command_visible = self.command_visible - 1
        if self.command_visible == 0:
            self.text_backspace( len(self.prompt)+len(self.command_text)+1 )

    def text_show_command_line(self):
        self.command_visible = self.command_visible + 1
        if self.command_visible == 1:
            sys.stdout.write( self.prompt+' '+self.command_text )
            sys.stdout.flush()

    def text_advance_command_line(self):
        sys.stdout.write('\n')
        self.command_text = ''
        self.command_visible = 0

    #def text_mainloop(self):
    #  while 1:
    #    try:
    #      self.text_mainloop_inner()
#       return
#      except Error, error:
#        if error.message == 'sigint':
#         self.command_text = ''
#         self.command_text_lines = [ ]
#         self.set_normal_prompt()
#       else:
#         raise error

    def text_mainloop(self):
        self.command_text = ''

        set_config('text_status',{'running':1})

        stdin_echo(0)
        self.text_show_command_line()
        try:
            threadnice_mainloop_fake_start()
            while threadnice_mainloop_fake_is_running():
                threadnice_mainloop_fake_iteration()
                    
                if select.select([0],[ ],[ ],0.1)[0]:
                    self.text_read_input()

                while self.escape_queue:
                    item = self.escape_queue.pop(0)

                    if self.command_text_lines:
                        continue

                    if item == 'A' and self.history_position > 0:
                        self.history_position = self.history_position - 1
                        self.text_hide_command_line()
                        self.command_text = self.history[self.history_position]
                        self.text_show_command_line()

                    if item == 'B' and self.history_position < len(self.history):
                        self.history_position = self.history_position + 1
                        self.text_hide_command_line()
                        if self.history_position < len(self.history):
                            self.command_text = self.history[self.history_position]
                        else:
                            self.command_text = ''
                        self.text_show_command_line()

                while self.input_queue and self.input_queue <> chr(27) and self.input_queue[:2] <> chr(27)+'[':
                    char = self.input_queue[0]
                    self.input_queue = self.input_queue[1:]

                    if char == '\n':
                        line = self.command_text
                        self.text_advance_command_line()

                        if line[-1:] == '\\':
                            self.command_text_lines.append(line[:-1])
                            self.set_normal_prompt()
                            self.text_show_command_line()
                        else:
                            lines = self.command_text_lines
                            self.command_text_lines = [ ]
                            lines.append(line)
                            self.handle_user_entry(string.strip(string.join(lines,'\n')))
                            self.text_show_command_line()
                        continue

                    if char == chr(4):
                        self.text_hide_command_line()
                        self.show(_('Entering shell (Circle is still running)\n'))

                        self.lock.acquire()
                        old_quiet = self.quiet
                        old_activity = self.activity
                        self.lock.release()

                        if not old_quiet:
                            self.set_status(1,'(shell)')
        
                        command = 'exec $SHELL'
                        status = get_config('text_status',{ })
                        if status.get('directory',None):
                            command = 'cd '+quote_for_shell(status['directory'])+' ; '+command
                            check_directory = 0
                        else:
                            check_directory = 1

                        stdin_echo(1)
                        try:
                            self.do('//'+command)
                            self.show('\n')
                        finally:
                            stdin_echo(0)
                            self.text_show_command_line()
                        
                        if check_directory:
                            status = get_config('text_status',{ })
                            if not status.has_key('directory'):
                                self.show(_("Current working directory was not saved.\n"+\
                                                    "Suggest adding `circle -s` to your shell prompt (see /help).\n"))

                        if not old_quiet:
                            self.set_status(old_quiet,old_activity)

                        continue

                    if ord(char) < 32:
                        continue
                    
                    if char == chr(127):
                        if self.command_text:
                            self.command_text = self.command_text[:-1]
                            #sys.stdout.write(chr(8)+' '+chr(8))
                            #sys.stdout.flush()
                            self.text_backspace(1)
                    else:
                        sys.stdout.write(char)
                        sys.stdout.flush()
                        self.command_text = self.command_text + char

        finally:
            set_config('text_status',{'running':0})
            threadnice_mainloop_fake_stop()
            self.text_hide_command_line()
            stdin_echo(1)







class Chat_gtk(Chat):
    """ GTK chat interface. """

    def __init__(self,app):
        import gtk

        Chat.__init__(self,app,1)
        
        self.fields = [ ]
        self.widget = gtk.GtkText()
        self.widget.set_usize(10,10)
        self.widget.set_word_wrap(1)
        self.widget.set_editable(1)
        self.horizon = 0
        self.insertion_point = 0

        self.userlist_widget = gtk.GtkScrolledWindow()
        
        self.userlist_widget.ensure_style()
        font = self.userlist_widget.get_style().font
        
        self.userlist_widget.set_policy(gtk.POLICY_NEVER,gtk.POLICY_AUTOMATIC)
        self.whiteboard=None


        def on_event(_w,_e,self=self):
            length = self.widget.get_length() - self.horizon
            command = self.widget.get_chars(self.horizon,self.horizon+length)
            position = self.widget.get_position() - self.horizon
            
            if position < 0:
                return 1

            if _e.keyval == ord('c') and (_e.state & 5): # Ctrl-C
                self.lock.acquire()
                try:
                    if self.shell <> None:
                        os.kill(self.shell,2)
                except OSError:
                    pass
                self.lock.release()
                self.widget.emit_stop_by_name("key-press-event")
            
            if position == 0 and _e.keyval == 65361: # Left arrow
                self.widget.emit_stop_by_name("key-press-event")
            
            if string.find(command[:position],'\n') == -1 and \
                 _e.keyval == 65362: # Up arrow
                if self.history_position > 0:
                    self.history_position = self.history_position - 1
                    self.widget.set_position(self.horizon)
                    self.widget.freeze()
                    self.widget.set_point(self.horizon)
                    self.widget.forward_delete(length)
                    self.widget.insert_defaults(self.history[self.history_position])
                    self.widget.thaw()
                    self.widget.set_position(self.widget.get_length())
                self.widget.emit_stop_by_name("key-press-event")

            if string.find(command[position:],'\n') == -1 and \
                 _e.keyval == 65364: # Down arrow
                if self.history_position < len(self.history):
                    self.history_position = self.history_position + 1
                    if self.history_position < len(self.history):
                        str = self.history[self.history_position]
                    else:
                        str = ''
                    self.widget.set_position(self.horizon)
                    self.widget.freeze()
                    self.widget.set_point(self.horizon)
                    self.widget.forward_delete(length)
                    self.widget.insert_defaults(str)
                    self.widget.thaw()
                    self.widget.set_position(self.widget.get_length())
                self.widget.emit_stop_by_name("key-press-event")

            return 1
        self.widget.connect("key-press-event",on_event)

        def on_event(_w,_e,self=self):
            position = self.widget.get_position()
            if position < self.horizon:
                chars = self.widget.get_chars(position,self.horizon)
                if string.find(chars,'\n') == -1:
                    self.widget.set_position(self.horizon)
            return 1
        self.widget.connect_after("key-press-event",on_event)

        def on_click(_w,_e,_self=self):
            for field in _self.fields:
                if field.end > _w.get_position() and field.start <= _w.get_position():
                    field.on_click(_w,_e)
                    break
            return 1
        self.widget.connect("button-release-event",on_click)


        def on_delete(_w,start,end,_self=self):
            if start < _self.horizon:
                _self.widget.emit_stop_by_name("delete-text")
        self.widget.connect("delete-text",on_delete)

        def on_insert(_w,str,len,position, _self=self):
            import gtk
            position = _self.widget.get_position()
            if position < _self.horizon:
                _self.widget.emit_stop_by_name("insert-text")
                _self.widget.set_point(_self.widget.get_length())
                style = _self.widget.get_style()
                _self.widget.insert(
                    #_self.bold_font,
                    _self.font,
                    style.fg[gtk.STATE_NORMAL],
                    style.base[gtk.STATE_NORMAL],
                    str[:len])
                return
            if len == 1 and str[0] == '\n':
                if position <> _self.horizon and \
                     _self.widget.get_chars(position-1,position) == '\\':
                     return
                command = _self.widget.get_chars(_self.horizon,_self.widget.get_length())
                _self.widget.set_position(_self.widget.get_length())
                _self.horizon = _self.widget.get_position()
                _self.insertion_point = _self.horizon
                _self.show('\n')
                _self.widget.emit_stop_by_name("insert-text")

                command = string.replace(command,'\\\n','\n')
                _self.handle_user_entry(command)
                    
        self.widget.connect("insert-text",on_insert)

        self.widget_visible = 1
        self.n_unseen_messages = 0

        def on_draw(widget, event, self=self):
            import gtk
            self.widget_visible = (event.state <> gtk.GDK.VISIBILITY_FULLY_OBSCURED)
            if self.widget_visible:
                self.n_unseen_messages = 0
                self.app.set_title()
        self.widget.connect("visibility-notify-event",on_draw)
        self.widget.set_events(gtk.GDK.VISIBILITY_NOTIFY_MASK)

        # Shell command support
        self.shell = None   # (pid, write_pipe)

        try:
            openpty = None
            try:
                from pty import openpty
            except:
                from os import openpty
            self.shell_master, self.shell_slave = openpty()
            def callback(source, condition, fd=self.shell_master,self=self):
                try:
                    str = os.read(fd,256)
                except OSError:
                    str = ''
                str = string.replace(str,'\r','')
                if string.find(str,chr(7)) <> -1:
                    str = string.replace(str,chr(7),'')
                    utility.beep()
                self.show(str)
            self.shell_tag = gtk.input_add(self.shell_master,5,callback)
        
            import termios

            # TERMIOS deprecated in 2.x series
            try:
                termios.ECHO
                TERMIOS = termios
            except:
                import TERMIOS

            attr = termios.tcgetattr(self.shell_slave)
            attr[3] = attr[3] & ~TERMIOS.ECHO
            termios.tcsetattr(self.shell_slave,TERMIOS.TCSANOW,attr)
        except:
            self.shell_master = None
            self.shell_slave  = None

    def start(self):
        import widgets

        self.font = self.widget.get_style().font

        self.blue_style   = self.widget.get_style().copy()
        for i in range(5):
            self.blue_style.fg[i] = widgets.blue
        self.green_style = self.widget.get_style().copy()
        for i in range(5):
            self.green_style.fg[i] = widgets.green

        Chat.start(self)
        
        # Note: 'chat whiteboard' not present in 0.32 and earlier,
        # and not present for text interface.
        self.node.add_handler('chat whiteboard', self, ('any',))

    def stop(self):
        import gtk

        self.lock.acquire()
        if self.shell_master <> None:
            gtk.input_remove(self.shell_tag)
            os.close(self.shell_master)
            os.close(self.shell_slave)

        try:
            if self.shell:
                os.kill(self.shell,2)
        except OSError:
            pass
        self.lock.release()

        Chat.stop(self)
        self.node.remove_handler('chat whiteboard')

    def handle(self, request, address, call_id):
        check.check_matches(request, (types.StringType,))  #=@R40
        check.check_is_af_inet_address(address)

        self.acquire_lock('hdl')
        try:
            if request[0] == 'chat whiteboard':
                # TODO: Events can be received in a different order from how
                # they were sent (and can even be received multiple times).
                # We ought to pass draw_string more info about order.
                # The (address,call_id) pair is enough to show order of
                # events from a given client, but not enough to show what
                # events that client had already received.
                def chat_board_task(self,request):
                    if self.whiteboard is not None:
                        self.whiteboard.draw_string(request)
                Task(chat_board_task,self,request).start()
                
                # The initial check-in of whiteboard.py checked that
                # the return type is a tuple.
                return ()
        finally:
            self.release_lock('hdl')

        return Chat.handle(self, request, address, call_id)
    
    
    def run_shell(self, command):
        # only call from main thread

        if self.shell_master == None:
            mainthread_call(self.show, _("No openpty in pty module.\n"))
            return

        try:
            pid = os.fork()
        except:
            event.set()
            mainthread_call(self.show, _("Could not fork.\n"))

        if not pid:
            os.dup2(self.shell_slave, 0)
            os.dup2(self.shell_slave, 1)
            os.dup2(self.shell_slave, 2)
            os.execlp('/bin/sh','/bin/sh','-c',command)
            os._exit(0)
                
        self.lock.acquire()
        self.shell = pid
        self.lock.release()

        def timeout(self=self,pid=pid):
            result = os.waitpid(pid,os.WNOHANG)
            if result[0] == pid:
                self.lock.acquire()
                self.shell = None
                self.lock.release()
                #self.set_prompt('['+self.app.config['name']+']')
                self.set_normal_prompt()
                return 0
            return 1

        utility.schedule_mainthread(250.0, timeout)

    def refresh_userlist(self):
        """Uses gtk."""
        def refr_usr_task(self):
            import gtk
            import widgets

            widget = gtk.GtkVBox(0,0)

            channels = self.app.channels.list.keys()
            channels.sort()
            for name in channels:
                try:
                    ch = self.app.channels.list[name]
                except ValueError:
                    continue

                hbox = gtk.GtkHBox(0,0)

                #mute = GtkCheckButton()
                mute = widgets.Toggle_icon_widget('<')
                mute.set_relief(gtk.RELIEF_NONE)
                def on_toggle(mute, name=name,self=self):
                    self.lock.acquire()
                    self.app.channels.mute(name,not mute.get_active())
                    mainthread_call(self.widget.grab_focus)
                    self.lock.release()
                mute.connect("toggled", on_toggle)
                self.lock.acquire()
                mute.set_active(not ch['muted'])
                self.lock.release()
                hbox.pack_start(mute,0,0,0)
                widgets.tooltips.set_tip(mute,_("listen to channel %s")%name,"")
                
                #checker = GtkCheckButton()
                checker = widgets.Toggle_icon_widget('>')
                checker.set_relief(gtk.RELIEF_NONE)
                def on_toggle(checker, name=name,self=self):
                    self.lock.acquire()
                    if checker.get_active():
                        if name not in self.channel:
                            self.channel.append(name)
                            self.app.channels.mute(name,0)
                    else:
                        if name in self.channel:
                            self.channel.remove(name)
                    mainthread_call(self.widget.grab_focus)
                    self.set_normal_prompt()
                    self.lock.release()
                checker.connect("toggled", on_toggle)
                self.lock.acquire()
                checker.set_active(name in self.channel)
                self.lock.release()
                hbox.pack_start(checker,0,0,0)
                widgets.tooltips.set_tip(checker,_("talk on %s")%name,"")

                n = name
                if len(n) > 36:
                    n = name[0:33]+"..."
                
                button = widgets.Text_button_widget(n) 
                button.set_style(self.blue_style)
                button.set_relief(gtk.RELIEF_NONE)
                def on_click(button, name=name,self=self):
                    self.app.channels.edit_settings(name)
                def on_press(button, ev, name=name,self=self):
                    if ev.button == 3:
                        self.app.channels.context_menu(name, ev)
                button.connect("clicked",on_click)
                button.connect("button-press-event",on_press)
                hbox.pack_start(button,1,1,0)
                widgets.tooltips.set_tip(button,_("view %s settings")%name)

                #label = GtkLabel(name)
                #label.set_alignment(0,0)
                #button.add(label)
         
                widget.pack_start(hbox,0,0,0)
 
            #widget.pack_start(GtkHSeparator(), FALSE,FALSE,0)

            self.app.name_server.lock.acquire()
            try:
                list = [ ]
                for item in self.app.name_server.nicknames.items():
                    if item[1] is not self.app.name_server.me:
                        if item[1].watch:
                            dist = item[1].distance
                            if dist == None: dist = 100.0
                        else:
                            dist = 101.0
                        if item[1].online:
                            top_sorter = 0
                        else:
                            top_sorter = 1
                        list.append(((top_sorter,dist),item[0],item[1]))
                list.sort()
            finally:
                self.app.name_server.lock.release()

            for item in list:
                item[2].lock.acquire()
                try:
                    name    = item[2].nickname
                    online  = item[2].online
                    status  = item[2].status.get('chat',(1,''))
                finally:
                    item[2].lock.release()

                addendum = ''
                if status[0] and status[1]:
                    addendum = '\n('+name+' '+status[1]+')'
                elif not status[0] and status[1]:
                    addendum = '\n('+name+_(' is quiet: ')+status[1]+')'
                
                hbox = gtk.GtkHBox(0,0)
 
                # NOTE (saf2):
                #
                # Here is where you would insert the code to display a mute
                # button next to the person's name.  This doesn't address the
                # issue of mute logic, however.  See the code to insert the
                # mute button logic above for inspiration.

                #checker = GtkCheckButton()
                checker = widgets.Toggle_icon_widget('>', 3)
                checker.set_relief(gtk.RELIEF_NONE)
                def on_toggle(checker, name=name,self=self):
                    self.lock.acquire()
                    if checker.get_active():
                        if name not in self.channel:
                            self.channel.append(name)
                    else:
                        if name in self.channel:
                            self.channel.remove(name)
                    mainthread_call(self.widget.grab_focus)
                    self.set_normal_prompt()
                    self.lock.release()
                checker.connect("toggled", on_toggle)
                self.lock.acquire()
                checker.set_active(name in self.channel)
                self.lock.release()
                hbox.pack_start(checker,0,0,0)
                widgets.tooltips.set_tip(checker,_("talk to ")+name+addendum,"")
                    
                button = widgets.Text_button_widget(name) 
                button.set_relief(gtk.RELIEF_NONE)
                def on_click(button, name=name,self=self):
                    self.app.name_server.lock.acquire()
                    try:
                        if self.app.name_server.nicknames.has_key(name):
                            self.app.name_server.nicknames[name].edit(self.app.name_server)
                    finally:
                        self.app.name_server.lock.release()

                def on_press(button, ev, self=self, name=name):
                    name_context_menu(button,ev,self,name)
                    
                button.connect("clicked",on_click)
                button.connect("button-press-event",on_press)
                hbox.pack_start(button,1,1,0)
                widgets.tooltips.set_tip(button,_("view %s's details")%name+addendum)
                         
                #label = GtkLabel(name)
                if online:
                    if status[0]:
                        button.set_style(self.blue_style)
                    else:
                        button.set_style(self.green_style)
                #label.set_alignment(0,0)
                #button.add(label)
                    
                widget.pack_start(hbox,0,0,0)
             

            for child in self.userlist_widget.children():
                self.userlist_widget.remove(child)

            self.userlist_widget.add_with_viewport(widget)
            widget.show_all()

        mainthread_call(refr_usr_task, self)

    def get_field(self,left_action=None,right_action=None):
        return Chat_field_gtk(self,left_action=left_action,right_action=right_action)

    def clear(self):
        self.widget.freeze()
        self.widget.set_point(self.horizon)
        self.widget.backward_delete(self.widget.get_length())
        self.widget.thaw()
        self.widget.set_position(self.widget.get_length())
        self.lock.acquire()
        self.horizon = 0
        self.insertion_point = 0
        self.fields = [ ]
        self.lock.release()
        self.show("\n")


    def do_whiteboard(self,data):
        if self.whiteboard is None:
            import whiteboard
            self.whiteboard=whiteboard.Whiteboard(self)


def name_field_left_click(field,app,widget,event):
    name=field.text.strip()
    if not app.name_server.nicknames.has_key(name):
        try:
            for item in app.chat.search_list:
                if item['name'] == name:
                    app.name_server.make_acquaintance(item)
                    break
        except Error:
            return

    if not app.name_server.nicknames.has_key(name):
        return

    app.chat.lock.acquire()
    if name not in app.chat.channel:
        app.chat.channel.append(name)
        app.chat.refresh_userlist()
    else:
        app.chat.channel.remove(name)
        app.chat.refresh_userlist()
    app.chat.set_normal_prompt()
    app.chat.widget.set_position(app.chat.widget.get_length())
    app.chat.lock.release()

def name_field_right_click(field,app,widget,event):
    name_context_menu(widget,event,app.chat,field.text.strip())

def channel_field_left_click(field,app,widget,event):
    name=field.text.strip()
    app.chat.lock.acquire()
    if name not in app.chat.channel:
        app.chat.channel.append(name)
    else:
        app.chat.channel.remove(name)
    if app.channels.list.has_key(name):
        app.chat.refresh_userlist()
    app.chat.set_normal_prompt()
    app.chat.widget.set_position(app.chat.widget.get_length())
    app.chat.lock.release()

def channel_field_right_click(field,app,widget,event):
    app.channels.context_menu(field.text.strip(), event)

def url_field_left_click(field,app,widget,event):
    try:
        if field.text[0:12]=='circle-file:':
            import search
            search.search_for_name(hash.url_to_hash(field.text),'Files matching url', app.node,app)
        else:
            browse_url(field.text)
    except Error,error:
        error.show()
    field.show(field.text,'purple')


def name_context_menu(widget,ev,chat,name):
    if ev.button == 3:
        import gtk
        menu = []

        if not chat.app.name_server.nicknames.has_key(name):
            try:
                for item in chat.search_list:
                    if item['name'] == name:
                         chat.app.name_server.make_acquaintance(item)
                         break
            except Error:
                return
        if not chat.app.name_server.nicknames.has_key(name):
            return

        if name in chat.channel:
            def untalk(data, chat=chat, name=name):
                if name in chat.channel:
                    chat.channel.remove(name)
                    chat.refresh_userlist()
                    chat.set_normal_prompt()
            menu.append((_("Stop talking to %s")%name,untalk))
        else:
            def talk(data, chat=chat, name=name):
                if name not in chat.channel:
                    chat.channel.append(name)
                    chat.refresh_userlist()
                    chat.set_normal_prompt()
            menu.append((_("Talk to %s")%name,talk))
        if chat.channel!=[] and chat.channel!=[name]:
            def talkonly(data, chat=chat, name=name):
                chat.channel = [name]
                chat.refresh_userlist()
                chat.set_normal_prompt()
            menu.append((_("Talk only to %s")%name,talkonly))
        def details(data, chat=chat, name=name):
            chat.app.name_server.lock.acquire()
            try:
                chat.app.name_server.nicknames[name].edit(chat.app.name_server)
            finally:
                chat.app.name_server.lock.release()
        menu.append((_("View %s's details")%name, details)),
        if chat.app.name_server.nicknames[name].online:
            def browse(data, chat=chat, name=name):
                import search
                search.search_browse_files(
                    chat.app.name_server.nicknames[name].address,
                    chat.app.name_server.node,
                    chat.app)
            menu.append((_("Browse %s's files")%name, browse)),
        if not chat.app.name_server.nicknames[name].remember:
            def remember(data, chat=chat, name=name):
                acq = chat.app.name_server.nicknames[name]
                acq.lock.acquire()
                acq.remember = 1
                acq.lock.release()
            menu.append((_("Remember %s")%name, remember)),
        menuwidget = gtk.GtkMenu()
        for (text, action) in menu:
            mi = gtk.GtkMenuItem(text)
            mi.connect("activate", action)
            mi.show()
            menuwidget.append(mi)
        menuwidget.popup(None,None,None,ev.button,ev.time)

# vim: set expandtab :
