# chat_commands.py
#    Copyright (C) 2002  Monash University

# Issues:
#  - Need to be passed a "self" object, so ~/.circle/user_commands.py needs
#    to be given that self thing.
#
#  - Ideally want to make it hard for the user to shoot self in foot.
#

import os, re, string
from types import *

from utility import split_list
from utility import mainthread_call
import chat, search, hash
#from chat import chat_shell
#from chat import Chat_field

from settings import *
import safe_pickle
from file_server import *
from error import Error



def find_task(self, list, online_only=0):
    if not list:
        pipe = self.app.node.retrieve(hash.hash_of('service identity'), identity_redundancy)
    else:
        largest = list[0]
        for item in list[1:]:
            if len(item) > len(largest):
                largest = item

        pipe = self.app.node.retrieve(hash.hash_of('identity-name '+largest), identity_redundancy)

    self.lock.acquire()
    self.search_list = [ ]
    self.lock.release()

    mainthread_call(self.show, _('Searching...\n'))
    while not pipe.finished() and self.running:
        # TODO: In many of the places where we have `continue', we should
        # probably write some debugging text to stdout.
        for item in pipe.read_all():
            if online_only:
                demangled = item[1]
                if type(demangled) != types.DictionaryType:
                    continue
            else:
                try:
                    demangled = search.check_and_demangle_item(item[1])
                except:
                    continue

            if demangled.get('type') not in ('identity',
                                             'identity offline demangled'):
                continue

            keywords = demangled.get('keywords',[ ])
            bad = 0
            for keyword in list:
                if keyword not in keywords:
                    bad = 1
            if bad:
                continue

            #address_name = hostname_of_ip(item[0][0])

            self.lock.acquire()
            try:
                for list_item in self.search_list:
                    if demangled.get('key') == list_item.get('key'):
                        break
                else:
                    self.search_list.append(demangled)
                    mainthread_call(self.show,'%d. '%(len(self.search_list)))
                    mainthread_call(self.show,'%s'  %demangled.get('name'),'blue',\
                                    chat.name_field_left_click,chat.name_field_right_click)
                    mainthread_call(self.show,' (%s)\n'%demangled.get('human-name'))
            finally:
                self.lock.release()
        time.sleep(0.5)

    pipe.stop(1)

    mainthread_call(self.show, _('Search complete.\n'))

def do_join(self, param):
    if not param:
        if not self.channel:
            self.show(_('Use /join <nick> to open communications with nick.\n'))
        self.set_channel([ ])
        return

    param = param[0]
    if param[0] in '+-':
        prefix = param[0]
        param = param[1:]
    else:
        prefix = ''
    
    new_list = split_list(param)
    old_list = self.channel[:]

    if prefix == '+':
        for item in old_list:
            if item not in new_list:
                new_list.append(item)

    elif prefix == '-':
        for item in new_list:
            if item in old_list:
                old_list.remove(item)
            else:
                self.show(_('Not joined to %s\n')%item)
    
        new_list = old_list

    self.set_channel(new_list)
    self.show(_('Joined %s\n') % string.join(self.channel,',') )

def do_listen(self, param):
    if not param:
        self.show(_('Listening to %s\n') %
	  string.join(self.app.channels.get_listening(),',') )
        return

    param = param[0]
    if param[0] in '+-':
        prefix = param[0]
        param = param[1:]
    else:
        prefix = ''
 
    new_list = split_list(param) 
    for item in new_list:
        if item[:1] <> '#':
            self.show(_('All channel names must begin with a # character.\n'))
            return
    old_list = self.app.channels.get_listening()

    if prefix == '+':
        for item in old_list:
            if item not in new_list:
                new_list.append(item)

    elif prefix == '-':
        for item in new_list:
            if item in old_list:
                old_list.remove(item)
            else:
                self.show(_('Not listening to %s\n')%item)
    
        new_list = old_list
        
    self.app.channels.set_listening(new_list)
    self.show(_('Listening to %s\n') % string.join(new_list,',') )

def do_msg(self, param):
    self.send_message(split_list(param[0]),param[1])
    
def _giveto(self, dest, param):
    expression = param[0]

    if expression[:1] == '!':
        # Python expression
        try:
            result = eval(expression[1:],self.exec_vars)
        except:
            self.show(_('Error:\n')+\
                string.join(apply(traceback.format_exception,sys.exc_info())))
            return
    else:
        # File
        if not self.use_gtk:
            self.show(_("This feature does not currently work with the text interface, sorry.\n"))
            return

        try:
            if not os.path.isfile(expression):
                raise Error(_('The file %s does not exist.') % expression)
            
            expression = os.path.abspath(expression)
            #self.app.file_server.add_root(expression)
            
            #result = self.app.file_server.get_reference(expression)
            result = self.app.file_server.add_private(expression)
        except Error, error:
            self.show(error.message + '\n')
            return

    if len(param) == 2:
        message = param[1]
    else:
        message = ''

    try:
        attachment = safe_pickle.dumps(result)
    except:
        self.show(_("Couldn't SafePickle your object."))
        return
        
    self.send_message(dest,message,attachment)
    return

def do_give(self, param):
    _giveto(self, self.channel[:], param)

def do_giveto(self, param):
    _giveto(self, split_list(param[0]), param[1:])

def do_take(self, param):
    if not self.use_gtk:
        self.show(_("This feature does not currently work with the text interface, sorry.\n"))
        return
    
    what = param[0]
    self.lock.acquire()
    try:
        if not self.exec_vars.has_key(what) or \
             not self.exec_vars.has_key(what+'_source'):
            self.show(_('No such reference.\n'))
            return
        
        ref = self.exec_vars[what]
        source = self.exec_vars[what+'_source']
        if type(ref) <> type({}) or \
             ref.get('type',None) <> 'file' or \
             type(ref.get('filename',None)) <> type('') or \
             type(ref.get('name',None)) <> type(''):
            self.show(_('Not a reference to a file.\n'))
            return

        try:
            self.app.download_manager.download(ref,[source])
            self.show(_('Download of %s started.\n') % ref['filename'])
        except Error, error:
            self.show(_('Problem: ') + error.message + '\n')
    finally:
        self.lock.release()

def do_recall(self, param):
    def recall_task(self,recall_list,field):
        if recall_list == [ ]:
            mainthread_call(field.show, _('No message to recall.\n'))
        else:
            str = ''
            for item in recall_list:
                mainthread_call(field.show,
                                str+('Recalling message to %s...\n') % item[0])
                pipe = self.node.retrieve(item[1],cache_redundancy)
                n    = 0
                while not pipe.finished():
                    for pair in pipe.read_all():
                        if pair[1] == item[2]:
                            try:
                                self.node.call(pair[0],('data cache remove',item[3]))
                                n = n + 1
                            except Error:
                                pass
                    time.sleep(0.25)

                if n == cache_redundancy:
                    str = str + _('Recalled all %d copies of message to %s.\n') % (n,item[0])
                elif n == 0:
                    str = str + _('Could not recall message to %s.\n') % item[0]
                else:
                    str = str + _('Recalled %d copies of message to %s.\n') % (n,item[0])
            
            mainthread_call(field.show, str)
        mainthread_call(field.close)

    Task(recall_task, self, self.recall_list, self.get_field()).start()
 
def do_quiet(self, param):
    if param:
        self.set_status(1,param[0])
    else:
        self.set_status(not self.quiet, '')
    self.show_status()

def do_me(self, param):
    if param:
        activity = param[0]
    else:
        activity = ''

    self.set_status(0,activity)
    self.show_status()

def do_look(self, param):
    def look_task(self,nick,field):
        try:
            try:
                acq = self.app.name_server.locate(nick)
            except Error, error:
                mainthread_call(field.show, error.message+'\n')
                return
            acq.start_watching(self.node, 1)

            acq.lock.acquire()
            online  = acq.online
            address = acq.address
            name    = acq.name
            acq.lock.release()

            if online:
                result = self.node.call(address,('chat look',))
                if not check.matches(result, (types.IntType, 'text')):
                    str = _(' is using a bad Circle client')
                elif result[0]:
                    if result[1] == '':
                        str = _(' is listening')
                    else:
                        str = ' '+result[1]
                else:
                    if result[1] == '':
                        str = _(' is quiet')
                    else:
                        str = _(' is quiet: %s') % result[1]

                str = str + _(' (on %s)') % hostname_of_ip(address[0])
                mainthread_call(field.show, nick+str+'\n')
        except Error:
            pass

        mainthread_call(field.close)

    if param:
        Task(look_task, self, param[0], self.get_field()).start()
    else:
        self.app.name_server.lock.acquire()
        values = self.app.name_server.acquaintances.values()
        self.app.name_server.lock.release()
        for item in values:
            if item.watch:
                Task(look_task, self, item.nickname, self.get_field()).start()

def do_who(self, param):
    self.do_who()
    
def do_help(self, param):
    if param:
        # Discard all but first word.
        words = string.split(param[0], maxsplit=1)
        section = words[0]
        
        if section[0] == '/':
            section = section[1:]
        section = '#' + section
    else:
        section = ''
    
    if self.use_gtk:
        try:
            browse_file("chat_commands.html", section)
        except Error, error:
            error.show(self.app.window)
    else:
        stdin_echo(1)
        try:
            browse_file("chat_commands.html", section)
        except Error, error:
            self.show(error.message+'\n')
        stdin_echo(0)

def do_post(self, param):
    if not self.use_gtk:
        self.show(_("This feature does not currently work with the text interface, sorry.\n"))
        return

    self.app.gossip.prompt_for_wodge()

def do_gossip(self, param):
    if not self.use_gtk:
        self.show(_("This feature does not currently work with the text interface, sorry.\n"))
        return

    self.app.gossip.update_and_show_gossip()

def do_stats(self, param):
    node = self.app.node
    node.acquire_read_lock('stat')
    try:
        tot = node.network_usage
        prof = node.network_profile
        npeers = len(node.peers)
    finally:
        node.release_read_lock('stat')
    str = (_('Network usage so far: %s\nNumber of known peers: %d\n')
           % (human_size(tot), npeers))
    if param and param[0] in ('-p', '--profile'):
        str += _('Profile:\n')
        prof_items = map(lambda item: (item[1][1], item[1][0], item[0]),
                         prof.items())
        prof_items.sort()
        for nbytes,count,key in prof_items:
            str += (' ' + key + ': \t' + human_size(nbytes)
                    + _(' in %d items\n') % count)
    self.show(str)

def do_find(self, param):
    if param:
        param = param[0]
    else:
        param = ''

    list = string.split(string.lower(param))
    online_only = 0
    while '-o' in list:
        list.remove('-o')
        online_only = 1
    while '--online' in list:
        list.remove('--online')
        online_only = 1

    Task(find_task,self,list,online_only).start()

def do_add(self, param):
    try:
        value = int(param[0])
    except ValueError:
        raise Error(_("/add requires a number (ie a result from /find)."))

    self.lock.acquire()
    try:
        if value > len(self.search_list):
            raise Error(_("The list of search results isn't that long."))
        if value < 1:
            raise Error(_("/add requires a positive number."))
        value = self.search_list[value-1]
    finally:
        self.lock.release()

    def add_task(self,value=value):
        try:
            acq = self.app.name_server.make_acquaintance(value, None)

            acq.lock.acquire()
            acq.watch = 1
            acq.remember = 1
            name = acq.nickname
            acq.lock.release()
            mainthread_call(self.show,
                            _('%s has been added to your contact list.\n')%name)
            acq.start_watching(self.app.node)
        except Error, error:
            self.show(error.message+'\n')
    Task(add_task,self).start()

def do_forget(self, param):
    acq = self.app.name_server.nicknames.get(param[0],None)
    if not acq:
        self.show(_('%s is not in your contact list.\n') % param[0])
        return

    acq.lock.acquire()
    acq.watch = 0
    acq.remember = 0
    acq.lock.release()
    self.show(_('%s has been removed from your contact list.\n')%param[0])

def do_repeat(chat_obj, param):
    try:
        value = int(param[0])
    except ValueError:
        raise Error(_("/repeat requires a number."))

    if value <= 0:
        return

    chat_obj.lock.acquire()
    try:
        list = map((lambda msg,ix: (msg[1], ix, msg[0])),
                   chat_obj.incoming_message_history[-value:], range(value))
        # The map is in preparation for sorting.
        if list:
            fields = [chat_obj.get_field()]
            while len(fields) < len(list):
                fields.append(fields[-1].child())

            list.sort()
            def repeat_task(chat_obj, list=list, fields=fields):
                for i in range(len(list)):
                    chat.chat_receive_task(chat_obj, list[i][2], None,
                                           fields[i], 1, list[i][0], 0)
                    # Proof of @R33: check.is_opt_address(None).
                    # Proof of @R35: list[i][0] is assigned from a reordering
                    # (see the lambda) of [a subset of]
                    # chat_obj.incoming_message_history; @I16.
            Task(repeat_task, chat_obj).start()
    finally:
        chat_obj.lock.release()

def do_read(self, param):
    self.show(_('Displaying unread messages.\n'))
    self.lock.acquire()
    try:
        list = map((lambda msg,ix: (msg[2], ix, msg[0], msg[1])),
                   self.unread_message_list, range(len(self.unread_message_list)))
        if list:
            list.sort()
            fields = [self.get_field()]
            while len(fields) < len(list):
                fields.append(fields[-1].child())
            for msg_info in list:
                new_item = (msg_info[2], msg_info[0])
                # request, time
                check.check_matches(new_item,
                                    chat.incoming_message_history_item_tmpl)
                # Proof: list taken from self.unread_message_list; @I15.  list unmodified
                # other than sorting, and is local to this method and not passed to any
                # other objects (hence not shared with other threads that could write
                # into the list).
                # Relevance: @R.I16.
                self.incoming_message_history.append(new_item)

            def read_task(self,list=list,fields=fields):
                for i in range(len(list)):
                    # chat_obj, request, address,
                    # receptacle widget, quiet:yes, time, add_to_history:no
                    chat.chat_receive_task(self, list[i][2], list[i][3],
                                           fields[i], 1, list[i][0], 0)
                    # fixme prove that list[i][0] is like time():
                    # prove of chat.unread_message_list items.
            Task(read_task,self).start()
        # Don't clear the list of unread messages unless they've all been
        # successfully shown (or at least that no exceptions are thrown
        # in this thread).  Of course, for this to be safe, we must hold the
        # lock from before reading the list until after clearing the list, and
        # all updates to the list must hold the lock.
        self.unread_message_list = [ ]
    finally:
        self.lock.release()

def do_set(self, param):
    if not param: 
        for item in self.app.config_keys:
            self.show(item[1] + ' = ' + self.app.config[item[0]] + '\n')
        return

    match = re.match(r"(\S*)\s*(=|\s)?\s*(.*)",param[0],re.DOTALL)
    if not match:
        raise Error(_("Couldn't understand your /set command."))

    for item in self.app.config_keys:
        if item[1] == match.group(1):
            self.app.config[item[0]] = match.group(3)
            break
    else:
        self.show(_("No configuration item called %s")%match.group(1)+"\n")
        return

    self.app.sync_config()

def do_say(self, param):
    if not self.channel:
        raise Error(_("Before sending messages, you need to select a person or channel to talk to in the list to the right. If the person is not in the list, you need to search for them in the search box above and add them to your list of contacts."))

    self.send_message(self.channel[:],param[0])

def do_exit(self,param):
    check.check_matches(param, ['text'])
    
    if param:
        message = param[0]
    else:
        message = ''
    self.app.shutdown(message, 0)

def do_logout(self,param):
    check.check_matches(param, ['text'])
    
    if param:
        message = param[0]
    else:
        message = ''
    self.app.shutdown(message, 1)

def do_clear(chat_obj, ignored__param):
    if use_gtk:
        chat_obj.clear()

def register(self):
    """ Register commands. """

    self.register("join", 0,1, do_join)
    self.register("listen", 0,1, do_listen)
    self.register("msg", 2,2, do_msg)
    self.register("give", 1,2, do_give)
    self.register("giveto", 2,3, do_giveto)
    self.register("take", 1,1, do_take)
    self.register("recall", 0,0, do_recall)
    self.register("quiet", 0,1, do_quiet)
    self.register("quietly", 0,1, do_quiet)
    self.register("me", 0,1, do_me)
    self.register("look", 0,1, do_look)
    self.register("who", 0,0, do_who)
    self.register("help", 0,1, do_help)
    self.register("post", 0,0, do_post)
    self.register("gossip", 0,0, do_gossip)
    self.register("stats", 0,1, do_stats)
    self.register("find", 0,1, do_find)
    self.register("add", 1,1, do_add)
    self.register("forget", 1,1, do_forget)
    self.register("repeat", 1,1, do_repeat)
    self.register("read", 0,0, do_read)
    self.register("set", 0,1, do_set)
    self.register("say", 1,1, do_say)
    self.register("exit", 0,1, do_exit)
    self.register("logout", 0,1, do_logout)
    self.register("clear", 0,0, do_clear)


# vim: set expandtab :
