# Circle interface to eDonkey and Overnet

# This file based heavily on file_server.py:
#    Copyright (C) 2001,2002  Paul Francis Harrison
#
# The modifications from file_server.py are
#    Copyright (C) 2002  Thomas Voegtlin
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# TODO :
#        Use free software edonkey client: http://www.freesoftware.fsf.org/mldonkey/
#        handle edonkey files with the player
#        interface to mldonkey 
#        checksum the binary file after it is downloaded
# 
# bugs : if hashing takes too much time, it is interpreted by circle as a deadlock
#

import os,os.path,string,traceback,threading
from types import *
from utility import Task_manager
from error import Error
import file_server
import utility
import widgets
import time


def donkey_retrieve_task(pipe,donkey):
    
    donkey.results = [ ]
    donkey.waiting_results = 1
    donkey.reading_results = 0

    while donkey.waiting_results:
        time.sleep(0.1) 

    time.sleep(1) #wait here because the overnet client is not always ready
                  #also allows to get more results
    donkey.lock.acquire()
    donkey.reading_results = 1
    donkey.lock.release()

    donkey.tell('vr\n')
    time.sleep(0.3)

    donkey.lock.acquire()
    donkey.reading_results=0
    donkey.lock.release()

    for item in donkey.results:
        ss2 = item.split('\t')      #all fields
        ss = ss2[0].split(' ',1)    #name and number, first field
        if len(ss2)>1:
            pipe.write((donkey.source,{ \
                        'keywords' : [ donkey.key ], \
                        'length' : int(ss2[1]), \
                        'type' : 'file', \
                        'name' : 'file:'+ss[0][1:-1], \
                        'index' : ss[0][1:-1], \
                        'filename' : ss[1] \
                        }))


class Donkey(Task_manager):
    """ class Donkey. """

    def __init__(self,app,source):

        Task_manager.__init__(self)
        self.node = app.node
        self.app  = app
        self.pid = None
        self.results = [ ]
        self.key = None
        self.downloads = [ ]
        self.running = 0
        self.connected = 0
        self.connecting = 0
        self.reading_message = 0
        self.message = ''
        self.exit=0
        
        self.reading_results = 0
        self.reading_downloads = 0
        self.reading_status = 0
        self.waiting_results = 0
        self.chat_field = None
        self.source = source

        self.tempdir=os.path.join(self.app.config['download'],'.edonkey_temp/')
        if self.source=='Overnet':
            self.client_file    =os.path.join(utility.config_dir,'overnet0.29.1')
            self.connected = 1
        elif self.source=='eDonkey':
            self.client_file    =os.path.join(utility.config_dir,'donkey_s_59-3')
            self.serverlist_file=os.path.join(utility.config_dir,'server.met')
        else:
            raise Error('unknown client')

        if os.path.isdir(self.tempdir):
            list = os.listdir(self.tempdir)
            for item in list:
                if item[-4:]=='.met' and not list.count(item[:-4]):
                    os.unlink(os.path.join(self.tempdir,item))

            for item in list:
                if item[-5:]=='.part' and not list.count(item+'.met'):
                    os.unlink(os.path.join(self.tempdir,item))
        
    def display_message(self,msg,color='grey'):
        if self.app.chat.running:
            self.app.chat.lock.acquire()
            if self.chat_field ==None or self.chat_field not in self.app.chat.fields:
                self.chat_field = self.app.chat.get_field()
                if self.source=='Overnet':
                    self.chat_field.set_action(self.overnet_menu,self.overnet_menu)
                else:
                    self.chat_field.set_action(self.edonkey_menu,self.edonkey_menu)
            self.chat_field_color=color
            utility.mainthread_call(self.chat_field.show,msg,color)
            self.app.chat.lock.release()

    def start(self):

        def poll_downloads_task(self):

            while self.running:
                time.sleep(0.5)
                if not self.reading_results and not self.reading_status and not self.reading_downloads:
                    self.lock.acquire() #  otherwise downloads get written in results... 
                    self.vd()
                    self.reading_downloads = 1
                    self.lock.release()
                if self.app.chat.running and self.chat_field<>None \
                   and self.chat_field not in self.app.chat.fields:
                    self.app.chat.lock.acquire()
                    text=self.chat_field.text
                    self.display_message(text,self.chat_field_color)
                    self.app.chat.lock.release()

        def read_messages_task(self):
            buffer = ""
            while self.running:
                try:
                    str=""
                    c=""
                    while c<>'\n':
                        c = os.read(self.shell_master,1)
                        str=str+c
                except:
                    self.running=0
                    return

                str = string.replace(str,'\r','')
                str = string.replace(str,'> ','')
                str = string.replace(str,chr(7),'')

                if not self.reading_downloads:
                    #print self.source+'##'+str[:-1]
                    if self.reading_message:
                        self.message = self.message + str

                if self.reading_status:
                    self.status+=str
                    if string.find(str,'Users') <> -1:
                        self.reading_status=0
                    
                if string.find(str,'Invalid Result number') <> -1:
                    raise Error('eDonkey error: '+str)

                if string.find(str,'Connected to:') <> -1:
                    self.display_message(self.source+' client c'+str[1:],'green')
                    if self.connecting:
                        self.reading_message = 1
                        self.message = ''
                    else:
                        self.reading_message = 0
                    self.connecting = 0
                    self.connected = 1

                elif string.find(str,'Connecting to') <> -1 and self.connecting:
                    self.display_message( self.source+' client c'+str[1:],'orange')
                elif string.find(str,'Disconnected') <> -1:
                    self.display_message( self.source+' client disconnected.\n','red')
                    self.connected = 0

                if self.waiting_results:
                    if str[0:16] == 'Results returned' or str[0:11] == 'Got results':
                        self.waiting_results = 0

                if self.reading_results:
                    if str[0:1] == '(':
                        self.results.append(str)

                if self.reading_downloads:
                    if string.find(str,'File Name') <> -1:
                        self.downloads = [ ]

                    if string.find(str,'Total') <> -1:
                        self.lock.acquire()
                        self.reading_downloads = 0
                        self.lock.release()

                        for downloader in self.app.download_manager.downloaders:
                            downloader.lock.acquire()
                            if downloader.node==self and not downloader.neverseen:
                                downloader.success=1

                        for item in self.downloads:
                            number = self.downloads.index(item)
                            has_downloader=0
                            for downloader in self.app.download_manager.downloaders:
                                if string.find(item,downloader.data['filename']) == -1:
                                    continue
                                else:
                                    downloader.number = number
                                    downloader.success=0
                                    downloader.neverseen=0
                                    has_downloader=1
                                try:
                                    downloader.bytes_downloaded = int(self.downloads[number+1].split('\t')[4][:-1])*1024
                                    downloader.speed_kBs = float(self.downloads[number+1].split('\t')[5]) 
                                except:
                                    break

                            if not has_downloader:
                                try:
                                    data=( { 'keywords' : [ '?' ], \
                                             'length' : int(self.downloads[number+1].split('\t')[3][:-1])*1024, \
                                             'type' : 'file', \
                                             'name' : 'file:?', \
                                             'index' : '?', \
                                             'filename' : self.downloads[number].split(' ',1)[1][:-2] \
                                             })
                                except:
                                    data=None
                                if data:
                                    downloader = file_server.Donkey_Downloader(self,data,self.app.config['download'])
                                    downloader.number = number
                                    downloader.success=0
                                    downloader.neverseen=0
                                    utility.mainthread_call(self.app.download_manager.assert_visible)
                                    self.app.download_manager.list.append([downloader.basename,''])
                                    self.app.download_manager.downloaders.append(downloader)
                                    utility.mainthread_call(downloader.start)
                                    downloader.lock.acquire()

                        for downloader in self.app.download_manager.downloaders:
                            downloader.lock.release()

                    if string.find(str,'Downloading:') <> -1:
                        str=None
                    if str:
                        self.downloads.append(str)

        if self.running:
            return
        Task_manager.start(self)
        self.running=1
        self.results = [ ]
        self.downloads = [ ]
        self.reading_results = 0
        self.reading_downloads = 0
        self.waiting_results = 0

        self.shell_master, self.shell_slave = os.openpty()

        import termios
        try:
            termios.ECHO
            TERMIOS = termios
        except:
            import TERMIOS

        attr = termios.tcgetattr(self.shell_master)
        attr[3] = attr[3] & ~TERMIOS.ECHO
        termios.tcsetattr(self.shell_master,TERMIOS.TCSANOW,attr)

        attr = termios.tcgetattr(self.shell_slave)
        attr[3] = attr[3] & ~TERMIOS.ECHO
        termios.tcsetattr(self.shell_slave,TERMIOS.TCSANOW,attr)

        try:
            self.pid = os.fork()
        except:
            raise Error('eDonkey : cannot fork')
            return


        if not self.pid:
            time.sleep(1)
            os.chdir(utility.config_dir)
            os.dup2(self.shell_slave, 0) 
            os.dup2(self.shell_slave, 1) 
            os.dup2(self.shell_slave, 2)

            os.execlp('/bin/sh','/bin/sh','-c',self.client_file)
            os._exit(0)

    
        self.set_options()
        self.connect()
        self.display_message(self.source+' client started.\n','green')

        utility.Task(poll_downloads_task,self).start(0)
        utility.Task(read_messages_task,self).start()

    def stop(self):
        if not self.running:
            return

        if self.app.chat.running:
            self.display_message('waiting for '+ self.source+' client to finish...\n','orange')
            self.chat_field.set_action(None,None)

        def timeout(self=self):
            result = os.waitpid(self.pid,os.WNOHANG)
            if result[0] == self.pid:
                self.lock.acquire()
                self.pid = None
                self.lock.release()
                return 0
            return 1

        utility.schedule_mainthread(250.0, timeout)

        while self.pid <> None:
            self.tell('q\n')
            time.sleep(0.3)
            self.tell('y\n')
            time.sleep(0.3)
            
        os.close(self.shell_master)
        os.close(self.shell_slave)
            
        self.connected=0
        self.running = 0
        if self.app.chat.running:
            self.display_message( self.source+' client stopped.\n','grey')
            self.chat_field.set_action(None,None)

        for downloader in self.app.download_manager.downloaders:
            if downloader.node==self:
                downloader.stop()

        Task_manager.stop(self)

    def connect(self):
        self.connecting=1
        self.tell('c\n')

    def disconnect(self):
        self.tell('x\n')
        self.connected=0
        self.connecting=0

    def vm(self):
        self.tell('vm\n')        

    def download(self,number):
        self.tell( 'd %s\n'%number)

    def cancel_download(self,number):
        self.lock.acquire() 
        self.tell('m %s c\n'%number)
        time.sleep(0.5)  #this prevents the temporary reapparition in the downloads list...
        self.lock.release() 

    def set_options(self):
        self.tell('in %s\n'%self.app.config['download'])
        self.tell('temp %s\n'%self.tempdir)
        self.tell('name circle_user_(http://thecircle.org.au)\n')
        #self.tell('verbose\n')
        #self.tell('asr\n')

    def vp (self):
        self.tell('vp\n')

    def vf (self):
        self.tell('vf\n')

    def help(self):
        self.tell('?\n')        

    def vd(self):
        self.tell('vd\n')

    def retrieve(self,key,pipe):
        self.tell('temp %s\n'%self.tempdir)
        self.key = key
        if self.running and self.connected: 
            self.tell('s %s\n'%self.key)
            pipe.start(donkey_retrieve_task,self)
        
    def tell(self,command):
        if self.running and self.pid <> None:
            os.write(self.shell_master,command)

    def edonkey_menu(self,field,app,widget,ev):
        import gtk
        menu = []

        if self.connected:
            def disconnect(data,self=self):
                self.disconnect()
            menu.append(("Disconnect",disconnect))
            def view_message(data,self=self):
                self.app.chat.show(self.message)
            menu.append(("View greetings from server",view_message))
        else:
            if self.connecting:
                def retry(data,self=self):
                    self.disconnect()
                    time.sleep(0.1)
                    self.connect()
                menu.append(("Try another server",retry))
                def abort(data,self=self):
                    self.disconnect()
                    self.display_message( 'eDonkey client disconnected.\n','red')
                menu.append(("Abort connection",abort))
            else:
                def connect(data,self=self):
                    self.connect()
                menu.append(("Reconnect",connect))

        menuwidget = gtk.GtkMenu()
        for (text, action) in menu:
            mi = gtk.GtkMenuItem(text)
            mi.connect("activate", action)
            mi.show()
            menuwidget.append(mi)
        menuwidget.popup(None,None,None,ev.button,ev.time)

    def overnet_menu(self,field,app,widget,ev):
        import gtk
        menu = []

        def view_status(data,self=self):
            self.reading_status=1
            self.status=''
            time.sleep(0.05)
            self.tell('g\n')
            time.sleep(0.1)
            self.reading_status=0
            self.app.chat.show('Overnet status : \n'+self.status)
        menu.append(("View status",view_status))

        def boot(data,_self=self):
            import gtk
            
            window = gtk.GtkWindow()
            window.set_border_width(10)        
            window.set_default_size(300,10)
            window.set_title(_('Connect to peer'))

            window.set_transient_for(_self.app.window)
            window.set_policy(gtk.FALSE,gtk.FALSE,gtk.TRUE)

            vbox = gtk.GtkVBox(gtk.FALSE, 5)
            window.add(vbox)

            label = widgets.Helpful_label_widget(_('Address of overnet node to boot from.'),\
                                         _("In order to use Overnet for the first time, you need to enter the adress and port number of an active Overnet node. Example: 192.168.0.4:7893  Overnet nodes addresses can be found on the web, see for example http://members.lycos.co.uk/appbyhp/connect.html"))
            vbox.pack_start(label.widget, gtk.FALSE,gtk.FALSE,0)
        
            entry = gtk.GtkEntry()
            vbox.pack_start(entry, gtk.FALSE,gtk.FALSE,0)

            hbox = gtk.GtkHBox(gtk.FALSE,5)
            vbox.pack_end(hbox, gtk.FALSE,gtk.FALSE,0)

            button = gtk.GtkButton(_("Cancel"))
            button.connect("clicked",\
                           lambda _b, _window=window: _window.destroy())
            button.set_flags(gtk.CAN_DEFAULT)
            hbox.pack_end(button, gtk.TRUE,gtk.TRUE,0)

            def greet(_b, _entry=entry, _window=window, _self=_self):
                _self.tell('boot %s\n'%_entry.get_text().replace(':',' '))
                _window.destroy()
        
            button = gtk.GtkButton(_("Connect"))
            button.connect("clicked",greet)
            button.set_flags(gtk.CAN_DEFAULT)
            hbox.pack_end(button, gtk.TRUE,gtk.TRUE,0)

            utility.show_window(window, 'greet')
            button.grab_default()
            
        menu.append(("Boot",boot))

        menuwidget = gtk.GtkMenu()
        for (text, action) in menu:
            mi = gtk.GtkMenuItem(text)
            mi.connect("activate", action)
            mi.show()
            menuwidget.append(mi)
        menuwidget.popup(None,None,None,ev.button,ev.time)


def init_donkey(app):

    if hasattr(app,'overnet') and app.overnet<>None: 
        app.config['use_donkey']=0
        return

    if hasattr(app,'donkey') and app.donkey<>None: 
        app.config['use_donkey']=0
        return

    serverlist_file=os.path.join(utility.config_dir,'server.met')
    client_file    =os.path.join(utility.config_dir,'donkey_s_59-3')

    if not os.path.isfile(client_file) or not os.path.isfile(serverlist_file):

        if not os.path.isfile(client_file):
            ok = threading.Semaphore()
            ok.acquire()
            utility.mainthread_call(app.dialog_install_file,client_file+'.gz',\
                                    "http://users.aber.ac.uk/tpm01/donkey_s_59-3.gz",\
                                    'Install the eDonkey client.', \
                                    "The eDonkey core client was not found. In order to connect to the eDonkey network, you need to install this file. If you click on 'Install', a binary file will be downloaded from http://users.aber.ac.uk/tpm01/donkey_s_59-3.gz and installed it in your .circle directory. Alternately, you may choose to install this file manually and click on 'Cancel'.  Note: The eDonkey client is not part of Circle, and it is proprietary software. Running malicious files from untrusted sources might damage your system. For more information about eDonkey, visit http://www.edonkey2000.com/",ok)
            ok.acquire()
            ok.release()

        if os.path.isfile(client_file+'.gz'):
            os.system('/bin/gunzip '+client_file+'.gz')
            os.chmod(client_file,0755)

        if os.path.isfile(client_file) and not os.path.isfile(serverlist_file):
            ok = threading.Semaphore()
            ok.acquire()
            utility.mainthread_call(app.dialog_install_file,serverlist_file,\
                                    "http://ocbmaurice.dyndns.org/pl/slist.pl?download",\
                                    'Install the eDonkey server list', \
                                    "You do not seem to have 'server.met', a file that contains a list of servers to connect to. In order to connect to the eDonkey network, you need such a list. If you click on 'Install', this file will be downloaded from http://ocbmaurice.dyndns.org/pl/slist.pl?download and installed in your .circle directory. You may choose to install this file manually if you already have a 'server.met' file containing your favorite servers. ",ok)
            ok.acquire()
            ok.release()

    if not os.path.isfile(client_file) or not os.path.isfile(serverlist_file):
        app.config['use_donkey']=0
        app.donkey = None
        if hasattr(app,'use_donkey'):
            app.use_donkey.set_active(0)
    else:
        def create_donkey(app=app):
            app.donkey = Donkey(app,'eDonkey')
            app.donkey.start()
        utility.mainthread_call(create_donkey,app)


def init_overnet(app):

    if hasattr(app,'donkey') and app.donkey<>None: 
        app.config['use_overnet']=0
        return

    if hasattr(app,'overnet') and app.overnet<>None: 
        app.config['use_overnet']=0
        return

    client_file    =os.path.join(utility.config_dir,'overnet0.29.1')

    if not os.path.isfile(client_file):
        ok = threading.Semaphore()
        ok.acquire()
        utility.mainthread_call(app.dialog_install_file,client_file+'.gz',\
                                "http://www.overnet.com/files/overnet0.29.1.gz",\
                                'Install the Overnet client.', \
                                "The Overnet core client was not found. In order to connect to the Overnet network, you need to install this file. If you click on 'Install', a binary file will be downloaded from http://www.overnet.com/files/overnet0.29.1.gz and installed it in your .circle directory. Alternately, you may choose to install this file manually and click on 'Cancel'.  Note: The overnet client is not part of Circle, and it is proprietary software. Running malicious files from untrusted sources might damage your system. For more information about Overnet, visit http://www.overnet.com",ok)
        ok.acquire()
        ok.release()

    if os.path.isfile(client_file+'.gz'):
        os.system('/bin/gunzip '+client_file+'.gz')
        os.chmod(client_file,0755)


    if not os.path.isfile(client_file):
        app.config['use_overnet']=0
        app.overnet = None
        if hasattr(app,'use_overnet'):
            app.use_overnet.set_active(0)
    else:
        def create_overnet(app=app):
            app.overnet = Donkey(app,'Overnet')
            app.overnet.start()
        utility.mainthread_call(create_overnet,app)



# vim: expandtab:shiftwidth=4 :
