# Error class

#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  Paul Francis Harrison
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

class Error:
    """ An Error class that is used throughout Circle. """

    def __init__(self, str):
        """ Create an Error with a specific message. 
        
                The message might either be an internal message,
                such as "no reply", or a more verbose message to
                be displayed to the used.
                """
        self.message = str

    def __repr__(self):
        return '<Error: '+self.message+'>'

    def show(self, window=None, action=None):  
        """ Open a GTK window to display the message. """
        import gtk

        import utility

        mywindow = gtk.GtkWindow()
        mywindow.set_border_width(10)        
        mywindow.set_default_size(300,10)
        if window:
            mywindow.set_transient_for(window)
        mywindow.set_title('Problem')

        vbox = gtk.GtkVBox(0, 0)
        mywindow.add(vbox)

        align = gtk.GtkAlignment(0,0,0,0)
        vbox.pack_start(align, 0,0,0)

        label = gtk.GtkLabel(self.message + '\n\n')
        label.set_justify(gtk.JUSTIFY_LEFT)
        align.add(label)

        button = gtk.GtkButton("Close")
        button.connect("clicked",lambda _b,_window=mywindow: _window.destroy())
        vbox.pack_end(button, 0,0,0)

        if action:
            mywindow.connect("destroy",action)
        
        utility.show_window(mywindow, 'problem')


