#    File server
#
#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  Paul Francis Harrison
#    Copyright (C) 2002  Monash University
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os,sys,threading,md5,string

from utility import *
import utility
from settings import *
from error import Error
import hash

import classify

# Flags for file scan

flags_full = { 'keywords':1, 'filename':1, 'name':1, 'subdirectories':1 }
flags_deb  = { 'keywords':0, 'filename':0, 'name':1, 'subdirectories':0 }

# File access locking

files_locked = [ ]
files_locked_lock = threading.RLock()

def acquire_file_lock(filename,for_writing=0):
    files_locked_lock.acquire()
    try:
        filename = os.path.abspath(filename)
        if for_writing and os.path.exists(filename): 
            return 0

        if filename in files_locked:
            return 0

        files_locked.append(filename)
        return 1
    finally:
        files_locked_lock.release()

def release_file_lock(filename):
    files_locked_lock.acquire()
    try:
        filename = os.path.abspath(filename)
        files_locked.remove(filename)
    finally:
        files_locked_lock.release()

# Downloading files

def circle_downloader_task(dl):
    try:
        while 1:
            position, length = dl.get_ticket()

            while 1:
                try:
                    link = dl.get_link()
                except Error:
                    return

                if not link:
                    time.sleep(2)
                    continue
 
                try:
                    str = dl.node.call(link, ('download chunk',dl.data['name'],position,length))
                except Error:
                    dl.link_failed(link)
                    continue
                break

            dl.link_succeeded(link)
 
            dl.lock.acquire()
            dl.chunks[position] = str
            dl.bytes_downloaded = dl.bytes_downloaded + length

            running = dl.running
            dl.lock.release()

            if not running:
                break
    except Error: 
        pass
 
def circle_downloader_refresh_sources_task(self):
    check.check_isinstance(self, Downloader)
    
    wait_time = 60
    while 1:
        pipe = self.node.retrieve(self.data['name'],1,1)
        try:
            list = [ ]
            while 1:
                list.extend(pipe.read_all())
                if pipe.finished():
                    break
                if not self.running:
                    return
                time.sleep(1)
        finally:
            pipe.stop()

        self.lock.acquire()
        try:
            for item in list:
                if item not in self.links:
                    self.links.append(item)
        finally:
            self.lock.release()

        for i in xrange(wait_time):
            if not self.running:
                return
            time.sleep(1)

        wait_time = wait_time+60
        
 
class Downloader(Task_manager):
    def __init__(self, data, directory):

        Task_manager.__init__(self)
        self.data = data
        self.directory = directory
        self.bytes_downloaded = 0L
        self.comment = ''
        self.remaining_time = 0.0
        self.success=0
        
        if not directory:
            raise Error('No download directory configured yet.')
        self.filename = os.path.join(self.directory, os.path.basename(self.data['filename']))
        self.basename = os.path.basename(self.filename)

    def start(self):
        Task_manager.start(self)
        self.current_event = threading.Event()
        self.current_event.set()
        self.position = 0
        self.length_remaining = self.data['length']

        utility.schedule_mainthread(500.0, self.update) 
        self.update()

    def stop(self):
        if not self.running:
            return
        Task_manager.stop(self, 1)

    def get_bytes_downloaded(self):
        return self.bytes_downloaded

    def get_bytes_total(self):
        return self.data['length']

    def get_speed(self):
        """Some sort of average bytes per second."""
        pass
    
    def get_remaining_time(self):
        tot_seconds = self.remaining_time
        if not(tot_seconds >= 0.0):
            return ''
        if tot_seconds < 60*60.0:
            return _('%dmin%02ds') % divmod(int(tot_seconds), 60)
        elif tot_seconds < 48*60*60.0:
            return _('%dh%02dmin') % divmod(int(tot_seconds) / 60, 60)
        else:
            tot_days = tot_seconds / (24.0 * 60 * 60)
            if tot_days < 100:
                return _('%d days') % int(tot_days)
            else:
                return _('(months)')

    def get_sources(self):
        pass

    def cancel_download(self):
        pass

    def update(self):
        pass


class Circle_Downloader(Downloader):

    def __init__(self, node, data,links, directory, on_complete=None):
        Downloader.__init__(self, data, directory)

        self.node = node
        self.links = links[:]
        self.chunks = { }
        self.bytes_written    = 0L
        self.speed_sum = 0
        self.speed_sum_weights = 0.01
        self.on_completion = on_complete

        if not acquire_file_lock(self.filename,1):
            #raise Error('That file is already in your download directory.')
            i = 1
            while not acquire_file_lock(self.filename + "-%d"%i,1):
                i = i + 1
            self.comment = '\n\nNote: not overwriting existing ' + os.path.basename(self.filename)
            self.filename = self.filename + "-%d"%i

        try:
            self.file = open(self.filename,'wb')
        except:
            raise Error("Couldn't open the file to write to.")

    def get_sources(self):
        if len(self.links) ==0:
            return " (waiting for source)"
        elif len(self.links) == 1:
            return " (1 source)"
        else:
            return " (%d sources)" % len(self.links) 

    def get_speed(self):
        """Some sort of average bytes per second.
           (Circle_Downloader uses a form of exponential smoothing.)"""
        return self.speed_sum / self.speed_sum_weights

    def start(self):
        Downloader.start(self)

        for i in range(download_threads):
            Task(circle_downloader_task,self).start(0)

        Task(circle_downloader_refresh_sources_task,self).start(0)

        #Multiple calls to timeout_add crash windows gtk
        #timeout_add(100,self.update)

    def get_ticket(self):
        try:
            self.lock.acquire()

            if self.length_remaining == 0:
                raise Error('finished')

            if self.length_remaining < download_chunk:
                fetch_length = self.length_remaining
            else:
                fetch_length = download_chunk
            
            result = (self.position, fetch_length)
            self.length_remaining = self.length_remaining - fetch_length
            self.position = self.position + fetch_length
        finally:
            self.lock.release()

        return result

    def get_link(self):
        self.lock.acquire()
        try:
            if not self.links:
                if not self.running:
                    raise Error("abort")
                else:
                    return None # just twiddle your thumbs for a bits...

            link = self.links.pop(0)
            self.links.append(link)
            return link
        finally:
            self.lock.release()

    def link_failed(self, link):
        self.lock.acquire()
        if link in self.links:
            self.links.remove(link)
        self.lock.release()

    def link_succeeded(self, link):
        pass

    def cancel(self):
        # nyi for Circle downloader, but Donkey_Downloader does handle this method.
        pass

    def stop(self):
        if not self.running:
            return

        Downloader.stop(self)

        self.file.close()
        if self.bytes_written < self.data['length']:
            try:
                os.unlink(self.filename)
            except OSError:
                pass
            #if self.links == [ ]:
            #  Error('Ran out of sources while downloading ' + self.data['filename']).show()
            self.success = 0
        else:
            self.success = 1
        release_file_lock(self.filename)

        if self.success and self.on_completion:
            self.on_completion()

    def update(self):
        if not self.running:
            return 0

        self.lock.acquire()
        try:
            total=0
            while self.chunks.has_key(self.bytes_written):
                chunk = self.chunks[self.bytes_written]
                del self.chunks[self.bytes_written]
                self.file.write(chunk)
                self.bytes_written = self.bytes_written + len(chunk)
                total=total+len(chunk)

            # Update the weighted average for "speed", where speed is taken to
            # be the number of half-bytes read in this update.
            # (Why is this, is update called every 0.5 seconds?  This assumption
            # should probably be modified given that we can't guarantee that it's
            # 0.5 seconds since the last update; TODO.)
            # The weights used are 1.0 for this update, 0.9 for the previous
            # one, 0.81 for the update before that, etc. (0.9**i for the i'th
            # most recent update).  This is more accurate than the simple
            # implementation of exponential smoothing that doesn't keep track
            # of how much weight of real data we have.
            self.speed_sum = 0.9 * self.speed_sum + (2 * total)
            self.speed_sum_weights = 0.9 * self.speed_sum_weights + 1.0
            
            done = (self.bytes_written == self.data['length'])
            if self.speed_sum != 0:
                # This is a hack with no theoretical basis.
                self.remaining_time \
                  = (0.8 * (self.remaining_time - 0.4)
                     + (0.2 * (self.data['length'] - self.bytes_downloaded)
                        / self.get_speed()))
        finally: 
            self.lock.release()
        
        if done:
            self.stop()
            
        return 1


class Donkey_Downloader(Downloader):

    def __init__(self, node, data, directory):
        Downloader.__init__(self, data, directory)
        self.node = node
        self.neverseen=1
        self.speed_kBs = 0

    def get_sources(self):
        return "(%s)"%self.node.source

    def get_speed(self):
        """Some sort of average bytes per second."""
        return self.speed_kBs * 1024

    def start(self):
        Downloader.start(self)

        if self.data['index'] <> '?':
            self.node.download(self.data['index'])
            
    def cancel_download(self):
        try:
            self.node.cancel_download('%d'%((self.number+1)/2))
        except:
            pass

    def update(self):
        if not self.running:
            return 0

        if self.success:
            self.stop()

        if self.speed_kBs != 0:
            self.remaining_time \
              = (0.99 * self.remaining_time
                 + 0.01 * ((self.data['length'] - self.bytes_downloaded)
                           / (self.speed_kBs * 1024.0)))
        return 1


class Download_manager:
    def __init__(self, app):
        import gtk

        self.app = app

        self.window = gtk.GtkWindow()
        self.window.set_title("Download")
        self.window.set_border_width(10)
        self.window.set_default_size(600,150)

        vbox = gtk.GtkVBox(0,5)
        self.window.add(vbox)
        vbox.show()

        scrolly = gtk.GtkScrolledWindow()
        scrolly.set_policy(gtk.POLICY_NEVER,gtk.POLICY_AUTOMATIC)
        vbox.pack_start(scrolly, 1,1,0)
        scrolly.show()

        self.list = gtk.GtkCList(2)
        self.downloaders = [ ] # parallels list
        self.list.set_column_width(0,300)
        self.list.set_selection_mode(gtk.SELECTION_SINGLE)
        scrolly.add(self.list)
        self.list.show()

        hbox = gtk.GtkHBox(0,5)
        vbox.pack_start(hbox, 0,0,0)
        hbox.show()

        self.cancel_button = gtk.GtkButton('Cancel download')
        hbox.pack_start(self.cancel_button, 1,1,0)
        
        #button = gtk.GtkButton('Close')
        #hbox.pack_start(button, 1,1,0)
        #button.show()

        def on_delete(window, event, self=self):
            while self.downloaders:
                self.downloaders[0].stop()
                self.on_complete(0)
            return 1
        self.window.connect("delete_event",on_delete)

        def on_select(list,index,column,event,self=self):
            self.selected_row = index
            self.cancel_button.show()
        self.list.connect("select-row",on_select)

        def on_unselect(list,index,column,event,self=self):
            self.cancel_button.hide()
        self.list.connect("unselect-row",on_unselect)

        def on_cancel(button, self=self):
            downloader=self.downloaders[self.selected_row]
            downloader.cancel_download()
            downloader.stop()
            self.on_complete(self.selected_row)
        self.cancel_button.connect("clicked",on_cancel)

        self.visible = 0
        
    def on_complete(self,row):
        self.list.remove(row)
        del self.downloaders[row]
        if len(self.downloaders) == 0:
            self.window.hide()
            self.visible = 0

    def assert_visible(self):
        if not self.visible:
            show_window(self.window, 'download', show_all=0)

            # Obscure GTK bug workaround
            # thomasV : I think this is useless, at least on linux ... added the test
            if sys.platform <> 'linux2':
                import gtk
                while gtk.events_pending():
                    gtk.mainiteration()
                
            self.visible = 1

    def start(self):
        #Multiple calls to timeout_add crash windows gtk
        #timeout_add(100,self.update)
        utility.schedule_mainthread(500.0, self.update)
        self.update()

    def stop(self):
        """ Only call from main thread. """

        for downloader in self.downloaders:
            downloader.stop()

        self.window.hide()


    def complete_last(self):
        i = len(self.downloaders) - 1
        del self.downloaders[i]
        self.list.remove(i)
    
    def download(self, item, sources):
        if sources == [ 'eDonkey' ]:
            if self.app.donkey.running:
                try:
                    name = self.app.donkey.results[int(item['index'])-1]
                except:
                    name = ''

                if string.find(name, item['filename']) <> -1 :
                    self.assert_visible()
                    downloader = Donkey_Downloader(self.app.donkey, item, self.app.config['download'])
                    self.list.append([downloader.basename,''])
                    self.downloaders.append(downloader)
                    downloader.start()
                else:
                    raise Error('This search item is out of date: Please do another search')
            else:
                raise Error('eDonkey client was stopped unexpectedly')
        elif sources == [ 'Overnet' ]:
            if self.app.overnet.running:
                try:
                    name = self.app.overnet.results[int(item['index'])-1]
                except:
                    name = ''

                if string.find(name, item['filename']) <> -1 :
                    self.assert_visible()
                    downloader = Donkey_Downloader(self.app.overnet, item, self.app.config['download'])
                    self.list.append([downloader.basename,''])
                    self.downloaders.append(downloader)
                    downloader.start()
                else:
                    raise Error('This search item is out of date: Please do another search')
            else:
                raise Error('Overnet client was stopped unexpectedly')
        else:
            downloader = Circle_Downloader(self.app.node,item,sources,self.app.config['download'])

            self.assert_visible()
            self.list.append([downloader.basename,''])
            self.downloaders.append(downloader)
            downloader.start()


    def update(self):
        if not self.app.running :
            return 0
        for i in range(len(self.downloaders)):
            downloader= self.downloaders[i]
            #downloader.update()
            if not downloader.running:
                self.on_complete(i)
                break
            else:
                str = ((" %d%% " % ((downloader.get_bytes_downloaded()+1)*100
                                    / (downloader.get_bytes_total()+1)))
                       + 'of ' + utility.human_size(downloader.get_bytes_total())
                       + '  ' + utility.human_size(downloader.get_speed()) + '/s'
                       + ' '  + downloader.get_remaining_time()
                       + '  ' + downloader.get_sources())
                self.list.set_text(i,1,str)
        return 1
        

# File server

class File:
    # Members:
    #   is_dir
    #   mtime
    #   info 
    #   hash
    #   length
    #   names
    pass

def build_file(path, mtime,flags,old_cache,new_cache):
    file = File()
    file.is_dir = 0

    if old_cache.has_key((path,mtime)):
        file.hash, file.length = old_cache[(path,mtime)]
    else:
        try:
            f = open(path,'rb')
            m = md5.new()
            file.length = 0L
            while 1:
                str = f.read(1<<20)
                if str == '':
                    break

                m.update(str)
                file.length = file.length + len(str)

            f.close()
            file.hash = m.digest()
        except IOError:
            raise Error('bad file')

    new_cache[(path,mtime)] = (file.hash, file.length)

    basename = os.path.basename(path)

    file.info = {
        'type'     : 'file',
        'filename' : basename,
        'name'     : file.hash,
        'length'   : file.length,
    }

    str = string.lower(basename)

    keywords = [ ]
    
    if flags.get('filename'):
        keywords.append(str)

    if flags.get('keywords'):
        for char in '-_.,?!':
            str = string.replace(str,char," ")

        keywords.extend(string.split(str)) 

        if string.lower(path[-4:]) in ('.mp3','.ogg'):
            try:
                f = open(path,'rb')
                f.seek(-128,2)
                id3block = f.read(128)
                f.close()

                if id3block[0:3] == 'TAG':
                    def strip_zeros(str):
                        index = string.find(str,'\000')
                        if index == -1:
                            return string.strip(str)
                        return string.strip(str[0:index])
                    file.info['music_title']  = strip_zeros(id3block[3:33])
                    file.info['music_artist'] = strip_zeros(id3block[33:63])
                    file.info['music_album']  = strip_zeros(id3block[63:93])
                    keywords.extend(string.split(string.lower(file.info['music_title'])))
                    keywords.extend(string.split(string.lower(file.info['music_artist'])))
                    keywords.extend(string.split(string.lower(file.info['music_album'])))
            except:
                pass

    file.names = [ ]

    if flags.get('name'):
        file.names.append(file.hash) 

    file.info['keywords'] = [ ]

    information = classify.classifier.information(path)
    for key in information.keys():
        if len(information[key]) >= min_search_keyword_len:
            # Generic is what is returned by the file command
            if key == 'generic':
                file.info['keywords'].append(information[key])
                file.names.append(hash.hash_of(information[key]))
            else:
                file.info[key] = information[key]
                file.names.append(hash.hash_of(information[key]))

    for word in keywords:
        if len(word) >= min_search_keyword_len and word not in file.info['keywords']:
            file.info['keywords'].append(word)
            file.names.append(hash.hash_of(word))

    return file

class Directory:
    # Members:
    #   is_dir
    #   mtime
    #   info 
    #   files
    pass

def build_directory(path,flags):
    dir = Directory()
    dir.is_dir = 1

    basename = os.path.basename(path)

    dir.info = {
        'type'     : 'directory',
        'filename' : basename,
    }

    names = os.listdir(path)

    dir.files = [ ]
    for item in names:
        if item[0] != '.':
            dir.files.append(os.path.join(path,item))

    return dir

def build_entry(path, last_entry, flags,old_cache,new_cache):
    try:
        if not acquire_file_lock(path):
            raise Error('Could not lock')
        try:
            mtime = os.path.getmtime(path)
            if last_entry and mtime == last_entry.mtime:
                if not last_entry.is_dir:
                    new_cache[(path,mtime)] = (last_entry.hash, last_entry.length)
                return last_entry

            if os.path.isdir(path):
                if os.path.basename(path) != '.':
                    result = build_directory(path,flags)
            else:
                # TODO: flocking
                #now = time.time()
                #if mtime > now-10 and mtime <= now:
                #  raise Error('File being updated')

                result = build_file(path, mtime,flags,old_cache,new_cache)

            result.mtime = mtime
            return result
        finally:
            release_file_lock(path)
    except OSError:
        raise Error('No such file')
    except IOError:
        raise Error('Bad shit going down')
    
def file_server_poller(server):
    #roots = [ ]
    
    #cache = get_config("hash_cache",{ })

    #while 1:

    server.lock.acquire()
    try:
        if not server.running:
            return
    
        roots = server.roots
    finally:
        server.lock.release()

    new_entries = { }
    new_cache   = { }
    new_paths   = { }
    new_names   = { }
    todo = [ ]
    for item in roots:
        todo.append((os.path.abspath(item[0]),item[1]))

    running = 1
    n_new = 0

    while todo and n_new < 10:
        running = server.running
        if not running:
            break
        
        item = todo.pop()

        if new_entries.has_key(item[0]):
            continue

        try:
            result = build_entry(item[0],server.entries.get(item[0],None),item[1],server.cache,new_cache)
            new_entries[item[0]] = result
            if result.is_dir:
                for file in result.files:
                    if item[1].get('subdirectories') or \
                         not os.path.isdir(file):
                        todo.append((file,item[1]))
            else:
                new_paths[result.hash] = (item[0], result.mtime)
                new_names[item[0]]        = result.hash
                if not server.cache.has_key((item[0],result.mtime)):
                    n_new = n_new + 1
        except Error, error:
            pass

    if not running:
        return

    for key in new_entries.keys():
        if not new_entries[key].is_dir:
            if not server.entries.has_key(key):
                for item in new_entries[key].names:
                    server.node.publish(item, new_entries[key].info)
            elif server.entries[key].mtime <> new_entries[key].mtime:
                for item in new_entries[key].names:
                    server.node.publish(item, new_entries[key].info)
                server.node.unpublish(server.entries[key].info)
    
    for key in server.entries.keys():
        if not server.entries[key].is_dir:
            if not new_entries.has_key(key):
                server.node.unpublish(server.entries[key].info)

    server.cache   = new_cache

    server.lock.acquire()
    server.paths   = new_paths
    server.names   = new_names
    server.entries = new_entries
    server.fresh = 1
    server.lock.release()

    if todo and server.running:
        utility.mainthread_call(server.run_poller_now)

        #for i in range(30):
        #  if not server.running or roots <> server.roots:
        #    break
        #  time.sleep(1)

    # Clean up
    #for key in server.entries.keys():
    #  if not server.entries[key].is_dir:
    #    server.node.unpublish(server.entries[key].info)
    #
    #set_config("hash_cache", cache)
            
class File_server(Task_manager):
    def __init__(self, node):
        Task_manager.__init__(self)

        self.paths          = { }  # hash -> (path, mtime)
        self.names          = { }  # path -> hash
        self.entries        = { }  # path -> entry object
        self.roots          = [ ]

        self.private_paths  = { }  # hash -> (path, mtime)

        self.node           = node

        self.fresh          = 0
    
        self.cache = get_config("hash_cache",{ })

    def start(self):
        Task_manager.start(self)
        
        self.node.add_handler('download chunk',self)
        self.node.add_handler('files available',self)

        #Task(file_server_task,self).start()
        self.start_poller(120, 30, file_server_poller,self)

    def stop(self):
        self.change_activity('removing handlers')
        self.node.remove_handler('download chunk')
        self.node.remove_handler('files available')
        self.change_activity('')

        self.change_activity('saving hash cache')
        set_config("hash_cache", self.cache)
        self.change_activity('')

        Task_manager.stop(self)

        infos = []
        for entry in self.entries.values():
            if not entry.is_dir:
                infos.append(entry.info)
        self.change_activity('unpublishing %d file infos' % len(infos))
        self.node.unpublish_set(infos)

    def set_roots(self, list):
        """ Pass this function a list [(root_path, root_flags)...] """
        self.lock.acquire()
        #self.roots.append(os.path.abspath(path))
        #self.roots.append((path,flags))
        self.roots = list
        self.fresh = 0
        self.lock.release()

    #def remove_root(self, path):
    #  self.lock.acquire()
    #  try:
    #    if path in self.roots:
#       self.roots.remove(path)
#       self.fresh = 0
#    finally:
#      self.lock.release()

    def add_private(self, path):
        try:
            mtime = os.path.getmtime(path)
        except OSError:
            raise Error('bad file')

        entry = build_file(path,os.path.getmtime(path),flags_full, { },{ })

        self.private_paths[entry.hash] = (path, mtime)

        return entry.info

    def await_refresh(self):
        # Do better...
        self.run_poller_now()
        while 1:
            self.lock.acquire()
            try:
                if self.fresh:
                    return
            finally:
                self.lock.release()
            time.sleep(0.5)

    def get_reference(self, path):
        self.await_refresh()
        self.lock.acquire()
        try:
            try:
                return self.entries[path].info 
            except:
                raise Error('No such file')
        finally:
            self.lock.release()

    def get_name(self, path):
        while 1:
            self.lock.acquire()
            try:
                if self.fresh:
                    break
            finally:
                self.lock.release()
            time.sleep(1)

        self.lock.acquire()
        try:
            if self.paths.has_key(path):
                return self.paths[path][0]
            raise Error('No such path')
        finally:
            self.lock.release()

    def handle(self, request, ignored__address, call_id):
        check.check_matches(request, (types.StringType,))
        check.check_is_af_inet_address(ignored__address)
        
        if request[0] == 'download chunk':
            path, mtime = self.paths.get(request[1],(None,None))
            if not path:
                path, mtime = self.private_paths.get(request[1],(None,None))

            try:
                if not path or \
                     os.path.getmtime(path) <> mtime:
                    return Error("no such file")
                file = open(path,'rb') 
            except IOError:
                return Error("no such file")

            try:
                file.seek(request[2])
                return file.read(request[3])
            finally:
                file.close()
        elif request[0] == 'files available':
            if len(request) == 3:
                list = self.paths.keys()
            else:
                for item in request[3]:
                    if item == '..':
                        return None

                list = [ ]
                for root in self.roots:
                    entry = self.entries.get(apply(os.path.join,[os.path.abspath(root[0])] + request[3]))
                    if not entry:
                        continue

                    if entry.is_dir:
                        for file in entry.files:
                            entry_2 = self.entries.get(file,None)
                            if entry_2:
                                if entry_2.is_dir:
                                    info = entry_2.info.copy()
                                    info['path'] = request[3] + [ info['filename'] ]
                                    list.append(info)
                                else:
                                    list.append(entry_2.hash)
                    else:
                        list.append(entry.hash)

            return list[request[1]:request[2]]


# vim: set expandtab :
