# Gossip exchange system

#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001,2002  Paul Francis Harrison
#    Copyright (C) 2002  Monash University
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import string
import time
import types
import utility
from widgets import *
from settings import *
from utility import get_checked_config, set_config, Task_manager, Task
import safe_pickle
import check
from error import Error
from node import bad_peer
from node import make_timeout
from node import is_timed_out
import search, hash

# A wodge contains:
#
#  'post-time'
#  'author'
#  'subject'
#  'text'
#
# As a map with prepended signature.
#
# Also remember: minimum distance to me, sources

# Todo:

# njh: make gossips appear as they are found, incrementally.

class Wodge:
    #Members:
    #  wodge_string
    #  wodge
    #  distance
    member_tmpls = {
        'wodge': 'any',
        'string': 'any',
        'signature': 'any',
        'initial_time': 'any',
        'initial_distance': 'any', # user's opinion at initial_time, or None
        'unit_decay_time': 'int',
        'opinions': types.DictionaryType,
        'collapsed': types.IntType
    }

    def __init__(self, map):
        self.togglees = [ ] # Calls to make to update display if wodge is folded/unfolded
        self.wodge = None
        self.string = None
        self.signature = None
        self.initial_time = None
        self.initial_distance = None
        self.unit_decay_time = 60*60*24*7
        self.opinions = { }    # acq.name -> distance at initial_time
        self.collapsed = 0
        
        for (key, tmpl) in self.member_tmpls.items():
            if map.has_key(key):
                val = map[key]
                if check.matches(val, tmpl):
                    setattr(self, key, val)

        # .initial_time.
        if type(self.initial_time) in (types.FloatType, types.IntType, types.LongType):
            try:
                self.initial_time = float(self.initial_time)
            except:
                self.initial_time = None
        if type(self.initial_time) != types.FloatType:
            self.initial_time = time.time()

        # .opinions.
        for (key, dist) in self.opinions.items():
            if type(dist) not in (types.FloatType, types.IntType, types.LongType):
                del self.opinions[key]

    def add_togglee(self, togglee):
        self.togglees.append(togglee)

    def remove_togglee(self, togglee):
        self.togglees.remove(togglee)

    def toggle(self):
        self.collapsed = not self.collapsed
        for togglee in self.togglees:
            togglee(self.collapsed)

    def to_map(self):
        map = { }
        for key in self.member_tmpls.keys():
            map[key] = getattr(self,key)
        return map

    def adjust_decays(self):
        "Make sure distance decays at least while we are online if clock is skewing."
        decay = self.decay()
        self.initial_time = time.time()
        if self.initial_distance is not None:
            self.initial_distance += decay
        for key in self.opinions.keys():
            self.opinions[key] += decay

    def decay(self):
        value = (time.time() - self.initial_time) / self.unit_decay_time
        if value < 0.0:
            return 0.0
        else:
            return value

    def distance(self, name_server, perspective=None):
        decay = self.decay()

        if perspective:
            value = self.opinions.get(perspective)
            if value:
                return value + decay
            
            return 1000.0 #Lots

        initial_distance = self.initial_distance
        if initial_distance == None:
            initial_distance = 1000.0 #Lots

            for pair in self.opinions.items():
                acq = name_server.acquaintances.get(pair[0])
                if acq:
                    acq_distance = acq.distance
                    if acq_distance <> None:
                        distance = acq.distance + pair[1]
                        if distance < initial_distance:
                            initial_distance = distance
        
        return initial_distance + decay

    def make_body(self, gossip):
        vbox_body = GtkVBox(FALSE,0)
        
        text_widget = Text_widget(450)
        text_widget.write(_('Topics: '))
        wodge_topics = self.wodge.get('topics',[ ])
        for i in range(len(wodge_topics)):
            text_widget.write_word(wodge_topics[i],1,lambda gossip=gossip,topic=wodge_topics[i]: gossip.show_gossip(topic))
            if i < len(wodge_topics)-1:
                text_widget.write(', ')
        
        if self.wodge['human-name']:
            opinion_list = [ ]
            decay = self.decay()
            for pair in self.opinions.items():
                acq = gossip.app.name_server.acquaintances.get(pair[0])
                if acq:
                    distance = acq.distance
                    if distance <> None:
                        opinion_list.append( (distance+decay+pair[1], acq.nickname, decay+pair[1]) )
            opinion_list.sort()
            
            str = ''
            for item2 in opinion_list[:5]:
                str = str + item2[1] + ' ' + '%.1f'%item2[2] + ', '
            if len(opinion_list) > 5:
                str = str + ' ...'
            else:
                str = str[:-2]
            
            #str = 'Posted ' + time.strftime('%d %b %Y',time.localtime(wodge['post-time'])) + ' by ' + wodge['human-name'] + 
            text_widget.write(_('\nPosted ') + time.strftime(_('%d %b %Y'),time.localtime(self.wodge['post-time'])) + _(' by '))
            text_widget.write_word(self.wodge['human-name'],1,
                lambda gossip=gossip,wodge=self.wodge: search.search_for_name(hash.hash_of('identity-offline '+wodge['name']),'Wodge poster',gossip.app.node,gossip.app)
            )
            
            str = '\n\n' + self.wodge['text'] + _('\n\nTrust distance:\n(')+str+')'
            text_widget.write('\n'+str, gossip.app)
        else:
            str = self.wodge['text'] + _('\n\nTrust distance:\n' + \
                        '(this is an anonymous post, you can help make it harder to track by reducing its distance)') 
            text_widget.write('\n'+str, gossip.app)

        #str = 'Topics: '+string.join(wodge.get('topics',[ ]),', ')+'\n'+str

        #vbox_item.pack_start(text_widget.widget, TRUE,TRUE,0)
        vbox_body.pack_start(text_widget.widget, TRUE,TRUE,0)

        hbox_body = GtkHBox(FALSE,0)
        vbox_body.pack_start(hbox_body, FALSE,FALSE,5)

        distance = self.distance(gossip.app.name_server)
        if distance < 20:
            max = 20
        else:
            max = distance
        distance_adj = GtkAdjustment(distance ,0.0,max + 0.5,0.5,0.5,0.5)

        distance_scale = GtkHScale(distance_adj)
        #vbox_item.pack_start(distance_scale, FALSE,FALSE,0)
        hbox_body.pack_start(distance_scale, TRUE,TRUE,0)

        def value_changed(adjustment, gossip=gossip,wodge=self):
            gossip.lock.acquire()
            wodge.initial_distance = adjustment.value - wodge.decay()
            gossip.lock.release()
        distance_adj.connect("value_changed",value_changed)

        alignment = GtkAlignment(0,1,0,0)
        hbox_body.pack_start(alignment, FALSE,FALSE,10)

        reply = GtkButton(_(' Reply '))
        #hbox_body.pack_start(reply, FALSE,FALSE,10)
        alignment.add(reply)

        def on_click(button, wodge_object=self,gossip=gossip):
            gossip.prompt_for_wodge(wodge_object)
        reply.connect("clicked",on_click)

        return vbox_body
                
def to_fixed(value):
    return int(value*256+0.5)
def from_fixed(value):
    return value / 256.0

def gossip_sync_poller(gossip):
    #while 1:
    
    gossip.sync_gossip()
    gossip.lock.acquire()
    requests = gossip.update_requests
    gossip.update_requests = [ ]
    gossip.lock.release()
        
    for item in requests:
        apply(item,())

    #  # Update hourly
    #  for i in range(60*60):
    #    time.sleep(1)
#
    #    # Atomic reads, so no need to lock
    #    if not gossip.running:
#       return
#      if gossip.update_requests:
#       break

class Gossip(Task_manager):
    def __init__(self, app):
        Task_manager.__init__(self)
    
        self.app = app
        self.node = app.node

        self.any_gossip_gets = 0

        self.update_requests = [ ]

        gossip_list = get_checked_config("gossip", types.ListType, [ ])
        self.gossip = [ ]
        for item in gossip_list:
            if type(item) != types.DictionaryType:
                print _("Warning: ignoring corrupt gossip item: "), `item`
                continue
            wodge = Wodge(item)
            wodge.adjust_decays()
            self.gossip.append(wodge)

    def check_invar(self):
        check.check_has_type(self.gossip, types.ListType)
        for wodge in self.gossip:
            check.check_isinstance(wodge, Wodge)
    
    def start(self):
        Task_manager.start(self)

        self.node.add_handler('gossip list', self,
                              (types.IntType, types.IntType))
        self.node.add_handler("gossip get",self)

        #Task(gossip_sync_poll_task,self).start()
        self.start_poller(20, 60*60, gossip_sync_poller,self)

    def stop(self):
        self.running = 0

        self.node.remove_handler("gossip list")
        self.node.remove_handler("gossip get")

        self.save()

        Task_manager.stop(self)
        
    def save(self):
        self.lock.acquire()
        try:
            gossip_list = [ ]
            for item in self.gossip:
                gossip_list.append(item.to_map())
            set_config("gossip", gossip_list, 1)
        finally:
            self.lock.release()

    def request_update(self, callback):
        self.lock.acquire()
        self.update_requests.append(callback)
        self.lock.release()

        self.run_poller_now()

    def sorted_wodges(self, perspective=None):
        self.lock.acquire()
        list = [ ]
        for item in self.gossip:
            list.append((item.distance(self.app.name_server, perspective),item))
        list.sort()
        self.lock.release()

        return list

    def sync_gossip(self):
        def gossip_fetch_task(self,acq):
            if not self.running:
                return

            if acq.distance <> None:
                #acq.start_watching(self.node, 1)
                
                acq.lock.acquire()
                online = acq.online
                address = acq.address
                distance = acq.distance
                acq.lock.release()

                if distance == None or not online:
                    return

                pos = 0
                fetch_timeout = make_timeout(gossip_fetch_time) 
                while not is_timed_out(fetch_timeout):
                    try:
                        result = self.node.call(address,('gossip list',pos,pos+20))
                        if not check.matches(result,
                                             [(types.IntType, 'any', 'any')]):
                            bad_peer(address,
                                     _("Bad reply to 'gossip list': ") + `result`)
                            break
                        if len(result) == 0:
                            break

                        all_ok = 1
                        # effic: sort self.gossip, sort the returned results,
                        # to speed up searches for signatures.  However, I
                        # haven't yet seen gossip_fetch_task come up in
                        # profiles.
                        for item in result:
                            already_there = 0

                            self.lock.acquire()
                            try:
                                for wodge in self.gossip:
                                    if wodge.signature == item[2]:
                                        #if wodge.distance(self.app.name_server) > from_fixed(item[0])+distance:
                                        #  # TODO: What to do with unit_decay_time?
                                        #  wodge.initial_distance = from_fixed(item[0])+distance
                                        #  wodge.initial_time     = time.time()
                                        wodge.opinions[acq.name] = from_fixed(item[0]) - wodge.decay()
                                        already_there = 1
                                        break
                            finally:
                                self.lock.release()

                            if already_there:
                                continue

                            try:
                                string = self.node.call(address,('gossip get',item[2]))
                                wodgewodge = safe_pickle.loads(string)
                            except Error:
                                all_ok = 0
                                break

                            #TODO: Confirm signature of known people

                            wodge = Wodge({
                                'wodge': wodgewodge,
                                'string': string,
                                'signature': item[2],
                                'initial_time': time.time(),
                                'initial_distance': None,
                                #'initial_distance': from_fixed(item[0]) + distance
                                'unit_decay_time': item[1],
                                'opinions': { acq.name : from_fixed(item[0]) },
                                'collapsed': 0})

                            if not self.insert_wodge(wodge):
                                all_ok = 0
                                break

                        if not all_ok:
                            break
                        
                        pos = pos + 18
                    except Error, error:
                        break
        
        tasks = [ ]
        for acq in self.app.name_server.acquaintances.values():
            if acq.distance <> None:
                #acq.start_watching(self.node, 1)
                #name server starts watching all
                #want start watching to be gradual, see name server start()
                while not acq.watched:
                    time.sleep(1)
                    if not self.running:
                        return

                if acq.online:
                    task = Task(gossip_fetch_task,self,acq)
                    tasks.append(task)

        for task in tasks:
                task.start()

        for task in tasks:
            task.join()


    def handle(self, request, ignored__address, call_id):
        check.check_matches(request, (types.StringType,))
        check.check_is_af_inet_address(ignored__address)
        
        self.lock.acquire()
        try:
            if request[0] == 'gossip list':
                #TODO: Type checking, locking, oh, and this is really inefficient
                self.any_gossip_gets = 1

                list = [ ]
                for item in self.gossip[:]:
                    list.append((to_fixed(item.distance(self.app.name_server)),
                                 item.unit_decay_time,
                                 item.signature))
                list.sort()
                return list[request[1]:request[2]]
            elif request[0] == 'gossip get':
                for item in self.gossip:
                    if item.signature == request[1]:
                        return item.string
                return Error(_("No such wodge."))
        finally:
            self.lock.release()

    def insert_wodge(self, wodge):
        check.check_isinstance(wodge, Wodge)
        
        self.lock.acquire()
        try:
            for item in self.gossip:
                if item.signature == wodge.signature:
                    # TODO: Should probably update trust list here...

                    return 0
                
            if len(self.gossip) < gossip_cache_size:
                self.gossip.append(wodge)
                return 1

            worst_pos      = -1
            worst_distance = wodge.distance(self.app.name_server)
            for i in range(len(self.gossip)):
                distance = self.gossip[i].distance(self.app.name_server)
                if distance < worst_distance:
                    worst_pos = i
                    worst_distance = distance
            
            if worst_pos >= 0:
                self.gossip[worst_pos] = wodge
                return 1

            return 0
        finally:
            self.lock.release()

    def post_wodge(self, topics,subject,text,distance,anonymous,in_reply_to=None):
        if topics == [ ]:
            raise Error(_('Please specify some topics.'))
        if self.app.config['human-name'] == '' and not anonymous:
            raise Error(_('Please specify your name using the "Network/Configure..." menu item.'))
        if subject == '':
            raise Error(_('Please give a subject for your item.'))
        if text == '':
            raise Error(_('Please write some text for your item.'))
    
        wodge_info = {
            'topics'         : topics,
            'subject'        : subject,
            'text'           : text,
            'post-time'      : long(time.time())
        }

        if anonymous:
            wodge_info['name'] = ''
            wodge_info['human-name'] = ''
        else:
            wodge_info['name'] = self.app.name_server.public_key_name
            wodge_info['human-name'] = self.app.config['human-name']

        if in_reply_to:
            wodge_info['in-reply-to'] = in_reply_to.signature

        wodge_str = safe_pickle.dumps(wodge_info)
        if anonymous:
            wodge_sig = hash.hash_of(wodge_str)
        else:
            wodge_sig = self.app.name_server.sign(wodge_str)

        wodge = Wodge({
            'wodge': wodge_info,
            'string': wodge_str,
            'signature': wodge_sig,
            'initial_time': time.time(),
            'initial_distance': distance,
            #Fixme: make unit_decay_time an option
            'unit_decay_time': 60*60*24*7,
            'opinions': { },
            'collapsed': 0})
        # distance = post_distance + (time - post_time)/unit_decay_time

        self.insert_wodge(wodge)
        self.save()

    def prompt_for_wodge(self,in_reply_to=None):
        window = GtkWindow()
        window.set_default_size(500,500)
        window.set_border_width(10)
        window.set_title(_('Post gossip'))
 
        vbox = GtkVBox(FALSE,5)
        window.add(vbox)

        label = Helpful_label_widget(_('Topics discussed:'),\
            _('Topics that your post relates too, separated by commas. '+\
            'You should supply several general topics, plus one specifically '+\
            'relating to your post.\n\n'+\
            'For example, if your post is about a GIMP kernel module, type:\n'+\
            '  linux kernel, gimp, gimp linux kernel module'))
        vbox.pack_start(label.widget, FALSE,FALSE,0)

        topics = GtkEntry()
        vbox.pack_start(topics, FALSE,FALSE,0)

        if in_reply_to:
            topics.set_text(string.join(in_reply_to.wodge.get('topics',[ ]),', '))

        label = Helpful_label_widget(_('Text:'),\
            _('Your actual post.\n\n'+\
            'Your text will be formatted automatically, so you don\'t need to press ' +\
            'enter at the end of each line. Also, people will be able to click on ' +\
            'any URLs in the text.'))
        vbox.pack_start(label.widget, FALSE,FALSE,0)
        
        text = GtkText()
        text.set_editable(TRUE)
        text.set_word_wrap(TRUE)

        if in_reply_to:
            if in_reply_to.wodge['human-name']:
                str = in_reply_to.wodge['human-name'] + _(' wrote that ')
            else:
                str = _('An anonymous author wrote that ')
            str = str + '"' + in_reply_to.wodge['subject'] + '"\n\n' 
            text.insert_defaults(str)
        
        scrolly = GtkScrolledWindow()
        scrolly.set_policy(POLICY_NEVER,POLICY_AUTOMATIC)
        scrolly.add(text)
        vbox.pack_start(scrolly, TRUE,TRUE,0)

        label = Helpful_label_widget(_('One line summary:'),\
            _('A short summary of your post.'))
        vbox.pack_start(label.widget, FALSE,FALSE,0)

        subject = GtkEntry()
        vbox.pack_start(subject, FALSE,FALSE,0)


        #al = GtkAlignment(0,0,0,0)
        #vbox.pack_start(al, FALSE,FALSE,0)
        #label = GtkLabel('Trust distance:')
        #label.set_justify(JUSTIFY_LEFT)
        #al.add(label)
        
        #label = Helpful_label_widget('Trust distance:',\
        #  'How much you yourself trust what you are saying in your post. '+\
        #  'A low distance implies high trust and vice versa. So if you believe '+\
        #  'the contents of your post highly, set the distance close to zero. But '+\
        #  'if you don\'t believe it so much, for example you are just passing on '+\
        #  'a rumour, set the distance to a larger value.',self.app.tooltips)
        #vbox.pack_start(label.widget, FALSE,FALSE,0)

        #distance_adj = GtkAdjustment(0.5 ,0.0,2.5,0.5,0.5,0.5)

        #distance_scale = GtkHScale(distance_adj)
        #vbox.pack_start(distance_scale, FALSE,FALSE,0)

        anonymous = GtkCheckButton(_("Anonymous"))
        vbox.pack_start(anonymous, FALSE,FALSE,0)

        # ... not quite there yet, gossip is only gotten hourly...
        #if not self.any_gossip_gets:
        #  label = GtkLabel('Note: No-one has requested gossip from you yet in this session.\n'+\
#                     'This might be because none of your friends are online, or because you are\n'+\
#                     'new to the Circle network and no one has added you to their trusted gossip\n'+\
#                     'sources.\n'+\
#                     '\n'+\
#                     'This post will only propagate once you and people who trust your news are\n'+\
#                     'online at the same time.')
#      label.set_alignment(0,0)
#      label.set_justify(JUSTIFY_LEFT)
#      vbox.pack_start(label, FALSE,FALSE,0)

        hbox = GtkHBox(FALSE,5)
        vbox.pack_end(hbox, FALSE,FALSE,0)

        def on_post(button, self=self,topics=topics,subject=subject,text=text,window=window,anonymous=anonymous,in_reply_to=in_reply_to):
            try:
                topic_list_unstripped = string.split(topics.get_text(),',')
                topic_list = [ ]
                for item in topic_list_unstripped:
                    str = string.lower(string.strip(item))
                    if str:
                        topic_list.append(str)

                self.post_wodge(topic_list,subject.get_text(),text.get_chars(0,-1),0.0,anonymous.get_active(),in_reply_to)
            except Error, error:
                error.show(window)
                return
            window.destroy()

        button = GtkButton(_("Post"))
        button.connect("clicked",on_post)
        hbox.pack_start(button, TRUE,TRUE,0)

        button = GtkButton(_("Cancel"))
        button.connect("clicked",lambda _b, _window=window: _window.destroy()) 
        hbox.pack_start(button, TRUE,TRUE,0)

        show_window(window, 'post_gossip')

    def update_and_show_gossip(self, topic=None):
        #status = Status()
        #status.show("Fetching gossip","Fetching the latest gossip...")

        #def on_fetched(self=self, status=status, topic=topic):
        #  mainthread_call(status.close)
        #  mainthread_call(self.show_gossip,topic)

        #self.request_update(on_fetched)

        self.show_gossip(update=1)

    def show_gossip(self, topic=None,person=None,perspective=None, update=0):
        window = GtkWindow()
        window.set_default_size(500,500)
        #window.set_border_width(10)
        window.set_title('Gossip')
 
        vbox = GtkVBox(FALSE,5)
        window.add(vbox)

        scrolly = GtkScrolledWindow()
        scrolly.set_policy(POLICY_NEVER,POLICY_ALWAYS)
        vbox.pack_start(scrolly, TRUE,TRUE,0)

        message = GtkLabel(_('Fetching gossip...'))
        scrolly.add_with_viewport(message)

        button = GtkButton(_("Close"))
        button.set_border_width(10)
        button.connect("clicked",lambda _b, _window=window: _window.destroy()) 
        vbox.pack_start(button, FALSE,FALSE,0)

        window.connect("destroy",lambda b, self=self: Task(lambda self: self.save(),self).start())

        show_window(window, 'show_gossip')

        while events_pending():
            mainiteration()

        def do_it(self=self, scrolly=scrolly,topic=topic,person=person,perspective=perspective, window=window):
            list = self.sorted_wodges(perspective)

            if not list:
                Error(_('To use gossip, you must first nominate some people whos gossip you would like to hear.\n\n'+\
                            'Click on the name of someone on your contact list, and tick the "Trust this person\'s gossip" box.')).show(window)
        
            vbox_inner = GtkVBox(FALSE,0)

            filtered_list = [ ]
            for item in list:
                if topic <> None and topic not in item[1].wodge.get('topics',[ ]):
                    continue

                if person <> None and item[1].wodge.get('name') <> person:
                    continue

                filtered_list.append(item)

            n_items = 0
            n_open  = 0
            for item in filtered_list:
                if n_items >= 100:
                    break
                if n_open  >= 20:
                    break

                wodge = item[1].wodge

                n_items = n_items + 1

                hbox_item = GtkHBox(FALSE,0)
                vbox_inner.pack_start(hbox_item, FALSE,FALSE,0)

                vbox_item = GtkVBox(FALSE,5)
                hbox_item.pack_start(vbox_item, TRUE,TRUE,0)

                title = Text_widget(450, 1)
                title.write(string.strip(wodge['subject']), self.app)
                vbox_item.pack_start(title.widget, FALSE,FALSE,0)

                if not item[1].collapsed:
                    vbox_body = item[1].make_body(self)
                    vbox_item.pack_start(vbox_body, TRUE,TRUE,0)
                    n_open = n_open+1
                                 
                button = GtkButton()
                button.set_relief(RELIEF_NONE)
                hbox_item.pack_start(button, FALSE,FALSE,0)

                if item[1].collapsed:
                    arrow_direction = ARROW_DOWN
                else:
                    arrow_direction = ARROW_UP
                arrow = GtkArrow(arrow_direction,SHADOW_IN)
                arrow.set_alignment(0,1)

                button.add(arrow)

                def on_toggle(collapsed,wodge=item[1],hbox_item=hbox_item,arrow=arrow,vbox_item=vbox_item,self=self):
                    #if down[0]:
                    if not collapsed:
                        arrow.set(ARROW_UP,SHADOW_IN)
                        vbox_body = wodge.make_body(self)
                        vbox_body.show_all()
                        vbox_item.pack_start(vbox_body, TRUE,TRUE,0) 
                    else:
                        arrow.set(ARROW_DOWN,SHADOW_IN)
                        vbox_item.children()[1].destroy()
                    hbox_item.queue_resize()

                item[1].add_togglee(on_toggle)
                button.connect("clicked",lambda b,item=item: item[1].toggle())
                button.connect("destroy",lambda b,item=item,on_toggle=on_toggle: item[1].remove_togglee(on_toggle))

            scrolly_children = scrolly.children()
            if(len(scrolly_children) > 0):
                scrolly.remove(scrolly_children[0])
            scrolly.add_with_viewport(vbox_inner)
            vbox_inner.show_all()
                        
        if update:
            self.request_update(lambda do_it=do_it: utility.mainthread_call(do_it))
        else:
            do_it()


# vim: set expandtab :
