import gtk
import pil2drawable

def default_scale_function(output_size, orig_size):
        return reduce(min,
                                    map(lambda a,b:float(a)/b,
                                            output_size,
                                            orig_size),
                                    1)


class GtkPILDisplay(gtk.GtkDrawingArea):
        def refresh(self):
                drawable = self.get_window()
                white_gc = self.get_style().white_gc
                bg_gc = self.get_style().bg_gc
                
                output_size = (drawable.width, drawable.height)

                gtk.draw_rectangle(drawable, bg_gc[0], gtk.TRUE, 0, 0,
                                                                             drawable.width, drawable.height)
                
                if(self.current_image):
                        rasterised = self.current_image
                        
                        orig_size = rasterised.size

                        s = self.scale_func(output_size, orig_size)
                        
                        scaled_size = map(lambda a,s=s:a*s, orig_size)

                        # PIL sometimes crashes
                        try:
                            if rasterised.mode != "RGBA":
                                    rasterised = rasterised.convert("RGB")
                            
                            if(scaled_size != orig_size):
                                    rasterised = rasterised.resize(scaled_size)
                        except:
                            return 1
                        
                        pil2drawable.pil2drawable(rasterised.im.id,
                                                                            (0, 0), (0, 0),
                                                                            drawable, white_gc)
                return 1

                
        def __init__(self, im=None):
                self.objects = []
                gtk.GtkDrawingArea.__init__(self)
                
                def expose(widget,event, self):
                        self.refresh()
                        
                self.add_events(gtk.GDK.EXPOSURE_MASK | gtk.GDK.BUTTON_PRESS_MASK)
                self.connect("expose_event",expose, self)
                self.current_image = im
                self.scale_func = default_scale_function
        
        def set_image(self, im):
                self.current_image = im

if __name__ == '__main__':
        wind = gtk.GtkWindow()

        import Image,sys
        
        widget = GtkPILDisplay(Image.open(sys.argv[1]))
        
        wind.add(widget)
        wind.show_all()
        wind.connect("destroy", gtk.mainquit)
        wind.connect("delete_event", gtk.mainquit)
        gtk.mainloop()
