# music specific widgets and mpg123, mpg321 and ogg123 frontend

#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  N Hurst
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from gtk import *
import popen2,os,os.path,select,time
import re,string

from error import Error
import utility

# wait until some reasonable amount of data is gathered (100K?)
# wait for twice as much data (200K?)
# estimate finish time, when finish time is within playing time, start playing.

class musicplayer(GtkWindow):
        def timeout(self):
                if self.playing:
                        while select.select([self.control.childerr],[],[],0)[0] :
                            fn = self.control.childerr.read(80)
                            if not fn:
                                break

                            fn = string.translate(fn, string.maketrans("\r", "\n"))
                            
                            fn = self.frame_matcher.search(fn)
                            if fn:
                                    time_elapsed   = float(fn.group(1)) * 60 + float(fn.group(2))
                                    time_remaining = float(fn.group(3)) * 60 + float(fn.group(4)) 
                                    if time_elapsed+time_remaining <> 0.0:
                                        self.progress.update(time_elapsed / (time_elapsed+time_remaining))
                if os.path.exists(self.song):
                        self.download_progress.set_value(os.path.getsize(self.song))
                return 1

        def pause_music(self, e=None):
                self.playing = 0

                try:
                    os.kill(self.control.pid, 19)
                except OSError:
                    pass
        
        def unpause_music(self, e=None):
                try:
                    os.kill(self.control.pid, 18)
                except OSError:
                    pass
                self.playing = 1

                if self.control.poll() <> -1:
                    self.progress.set_value(0)
                    self.control = popen2.Popen3(self.cmdline, 1)
        
        def restart_music(self, e=None):
                if not self.playing:
                    self.unpause_music()

                #os.kill(self.control.pid, 3)
                try:
                    os.kill(self.control.pid, 9)
                except OSError:
                    pass
                self.control.wait()

                # ogg takes a while to die
                time.sleep(2.0)
                # [pjm]: Why does this exist?  If we need to wait for a process to die,
                # we should be using wait/waitfor, not using some timing-dependent hack.

                self.progress.set_value(0)
                self.control = popen2.Popen3(self.cmdline, 1)
                self.playing = 1
        
        def __init__(self, f, s=0, keep_file_checkbox=1):
                GtkWindow.__init__(self)
                self.set_default_size(300,10)
                self.set_border_width(10)
                
                title = os.path.basename(f)
                self.set_name(title)
                self.set_title(title)

                self.song = f
                self.song_quoted = utility.quote_for_shell(self.song)
                #self.song = f
                #self.song_quoted = string.replace(self.song,'\\','\\\\')
                #self.song_quoted = string.replace(self.song_quoted,'$','\\$')
                #self.song_quoted = string.replace(self.song_quoted,'"','\\"')
                #self.song_quoted = '"'+string.replace(self.song_quoted,'`','\\`')+'"'
                extension = string.split(f, ".")[-1]
                lext = string.lower(extension)
                if lext[:3] == 'mp3':
                        player = os.popen("which mpg321").readline()
                        if not player:
                                player = os.popen("which mpg123").readline()
                        if not player:
                                raise Error("No mp3 player found. Please install mpg123 or mpg321.")
                        self.cmdline = "%s -q -v %s" % (string.strip(player), self.song_quoted)
                        #print "using %s" % self.cmdline
                elif lext[:3] == 'ogg':
                        self.cmdline = "ogg123 -q -v %s" % self.song_quoted
                self.vbox = GtkVBox(FALSE, 5)
                self.add(self.vbox)

                self.progress_label = GtkLabel(os.path.basename(self.song))
                self.vbox.pack_start(self.progress_label)
                
                self.progress = GtkProgressBar()
                self.vbox.pack_start(self.progress)
                #self.adjust = GtkAdjustment(0,0,10000,1)
                #self.progress.set_adjustment(self.adjust)      
                #self.progress.set_show_text(1)
                #self.progress.set_format_string(self.song)
                
                self.download_progress = GtkProgressBar()
                self.vbox.pack_start(self.download_progress)
                self.adjust = GtkAdjustment(0,0,s,1)
                self.download_progress.set_adjustment(self.adjust)      
                #self.download_progress.set_show_text(1)
                #self.download_progress.set_format_string("total length: %s" % utility.human_size(s))
                
                if keep_file_checkbox:
                    self.keep_file_button = GtkCheckButton("Keep file after playing")
                    self.vbox.pack_start(self.keep_file_button)
                #self.keep_file_button.connect("clicked", self.toggle_keep_playing)
                
                self.hbox = GtkHBox(FALSE, 5)
                self.vbox.pack_start(self.hbox)
                
                self.start_button = GtkButton("Play")
                self.hbox.pack_start(self.start_button)
                self.start_button.connect("clicked", self.unpause_music)
                self.playing = 1
                
                self.stop_button = GtkButton("Pause")
                self.hbox.pack_start(self.stop_button)
                self.stop_button.connect("clicked", self.pause_music)
                
                self.restart_button = GtkButton("Restart")
                self.hbox.pack_start(self.restart_button)
                self.restart_button.connect("clicked", self.restart_music)

                self.close_button = GtkButton("Close")
                self.hbox.pack_start(self.close_button)
                self.close_button.connect("clicked", self.destroy)
                
                self.show_all()
                
                self.control = popen2.Popen3(self.cmdline, 1)

                #print self.control, self.control.pid
                #self.frame_matcher = re.compile("Frame. *([0-9]+)")
                
                self.frame_matcher = re.compile("Time:\s*([0-9]+):([0-9.]+)\s*\[([0-9]+):([0-9.]+)\]")

                self.connect("destroy", self.on_destroy)

                #Multiple calls to timeout_add crash windows gtk
                #timeout_add(1000, self.timeout)
                self.timeout()
                utility.schedule_mainthread(1000.0, self.timeout)

        def on_destroy(self, event):
                import os

                #if self.playing:
                        #os.kill(self.control.pid, 3)
                if not self.playing:
                    self.unpause_music()

                try:
                    os.kill(self.control.pid, 9)
                except OSError:
                    pass
                self.control.wait()

if __name__ == '__main__':
        import sys
        f = sys.argv[1]
        music = musicplayer(f, os.path.getsize(f))
        music.connect('destroy', mainquit)

        #music.connect('delete_event', music.destroy)

        mainloop()


# vim: set expandtab :
