# UDP Proxy

#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  Paul Francis Harrison
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys,os,string,socket,select,cPickle,popen2

from utility import *
import utility
from error import Error
import settings

proxy_program = """
import os,socket,sys,string,select,cPickle

sock = socket.socket(socket.AF_INET,socket.SOCK_DGRAM)

try:
  SOL_SOCKET = socket.SOL_SOCKET
except:
  SOL_SOCKET = 1
try:
  MSG_DONTWAIT = socket.MSG_DONTWAIT
except:
  MSG_DONTWAIT = 64
try:
  SO_SNDBUF = socket.SO_SNDBUF
except:
  SO_SNDBUF = 7
try:
  SO_RCVBUF = socket.SO_RCVBUF
except:
  SO_RCVBUF = 8

for port in """+`settings.default_ports`+""":
    try:
        sock.bind(('',port))
        break
    except:
        pass
else:
    sock.bind(('',socket.INADDR_ANY))

for i in range(20,12,-1):
    try:
        sock.setsockopt(SOL_SOCKET,SO_SNDBUF,1<<i)
        sock.setsockopt(SOL_SOCKET,SO_RCVBUF,1<<i)
        break
    except socket.error:
        pass

try:
    sock.setsockopt(0,11,1)
except:
    pass
            
if sys.platform == 'linux2':
    sock.setsockopt(0,10,0)

print 'PORT',sock.getsockname()[1]
sys.stdout.flush()

def do_error():
    try:
        message, address = sock.recvfrom(65536,0x2000 | MSG_DONTWAIT)
    except:
        return
    cPickle.dump((1,address,message),sys.stdout,1)
    sys.stdout.flush()

while 1:
    result = select.select([sock, sys.stdin],[ ],[sock])
    if sys.stdin in result[0]:
        request = cPickle.load(sys.stdin) 
        if not request:
            sys.exit(0)
        for i in range(100):
            try:
                sock.sendto(request[1],request[0])
            except:
                do_error()
                continue
            break
    elif sock in result[0]:
        try:
            message, address = sock.recvfrom(65536,MSG_DONTWAIT)
            cPickle.dump((0,address,message),sys.stdout,1)
            sys.stdout.flush()
        except:
            do_error()
    elif sock in result[2]:
        do_error()
"""
#    queue = [ ]
#    while select.select([sys.stdin],[ ],[ ],0)[0]:
#      queue.append( cPickle.load(sys.stdin) )
#    for request in queue:
#      if not request:
#        sys.exit(0)
#      try:
#        sock.sendto(request[1],request[0])
#      except:
#        pass


#proxy_program = string.replace(proxy_program,"\n","\\n")

class Proxy(Synchronous):
    def __init__(self, host,password, transient_for=None):
        import gtk

        Synchronous.__init__(self)

        try:
            if sys.platform == 'win32':
                raise Error('definitely not')
            import gnome.zvt
            have_zvt = 1
        except:
            have_zvt = 0
        
        split = string.split(host,'@') 
        
        #try:
        #  address = socket.gethostbyname(split[-1])
        #except:
        #  raise Error("Couldn't find host " + host + '.')

        self.window = gtk.GtkWindow()
        self.window.set_title('Proxy login')
        #self.window.set_modal(1)
        if transient_for:
            self.window.set_transient_for(transient_for)

        if have_zvt:
            zvt = gnome.zvt.ZvtTerm(80,10)
            # TODO raise error on unexpected death during operation...
            died_tag = zvt.connect("child_died",threadnice_mainquit)
            self.window.add(zvt)
        elif sys.platform == 'win32':
            self.window.set_border_width(10)
            self.window.set_default_size(300,100)
            self.window.add(gtk.GtkLabel("Starting proxy."))
        else:
            self.window.set_border_width(10)
            self.window.add(gtk.GtkLabel("Starting proxy.\n\nYou may need to supply a password at the command line."))
        
        self.window.show_all()

        if have_zvt:
            zvt.feed("\nConnecting to " + split[-1] + " ...\n\n\r")
        
        if sys.platform <> 'win32':
            read_stdin, write_stdin  = os.pipe()
            read_stdout, write_stdout = os.pipe()

            if have_zvt:
                pid = zvt.forkpty()
            else:
                pid = os.fork()

            if pid == 0:
                os.close(0)
                os.dup(read_stdin)
                os.close(1)
                os.dup(write_stdout)
                os.close(write_stdin)
                os.close(read_stdin)
                os.close(write_stdout)
                os.close(read_stdout)

                #os.execlp('ssh','ssh',host,'python -c "exec \\"'+proxy_program+'\\""')
                os.execlp('ssh','ssh',host,'python','-u','-c','"exec input()"')
                os._exit(0)

            os.close(read_stdin)
            os.close(write_stdout)
            self.write_stdin = os.fdopen(write_stdin,'wb')
            self.read_stdout = os.fdopen(read_stdout,'rb', 0)

            self.write_stdin.write(repr(proxy_program)+'\n')
            self.write_stdin.flush()

            destroy_tag = self.window.connect("destroy",threadnice_mainquit)
            input_tag = gtk.input_add(self.read_stdout,5,threadnice_mainquit)
            threadnice_mainloop()
            gtk.input_remove(input_tag)
            self.window.disconnect(destroy_tag)
        else:
            command = 'plink -batch '+host
            if password:
                command = command + ' -pw "'+password+'"'
            command = command + " python -u -c 'exec input()'"
            try:
                self.read_stdout, self.write_stdin = popen2.popen2(command,mode='b')
            except:
                self.window.hide()
                raise Error('Could not start plink.')
            
            self.write_stdin.write(repr(proxy_program)+'\n')
            self.write_stdin.flush()

            while gtk.events_pending():
                gtk.mainiteration()

        if have_zvt:
            zvt.disconnect(died_tag)

        #On win32 newline is \n\r, confusion ensues
        #line = self.read_stdout.readline()
        line = ''
        while 1:
            char = self.read_stdout.read(1)
            if char == '\n' or not char:
                break
            line = line + char

        self.window.hide()
    
        if line[:5] <> 'PORT ':
            self.write_stdin.close()
            self.read_stdout.close()

            if sys.platform == 'win32':
                message = 'Attempt to start the proxy failed.\n\n' + \
                                    'Check the PuTTY session name and password.\n\n' + \
                                    'Make sure you have installed plink.exe somewhere on your path.\n' + \
                                    '(see "Help" for details)'
            else:
                message = 'Attempt to start the proxy failed.\n\n'+\
                                    'Check your username, server address and password.'
            raise Error(message)

        #self.address = (address, string.atoi(line[5:]))
        self.address = ('127.0.0.1', string.atoi(line[5:]))

        #print "Proxy started at %s:%d" % self.address
        self.broken = 0
        self.running = 1
        
        check.check_is_af_inet_address(self.address)  #=@E3

    def recvfrom(self):
        """ Returns (is_error, address, message).
                Only one thread should call this at a time!"""
        if self.broken:
            raise Error("proxy broken")
                
        try:
            result = cPickle.load(self.read_stdout)
        except:
            self.become_broken()
        return result

    def sendto(self, message, address):
        self.lock.acquire()
        try:
            if self.broken:
                raise Error("proxy broken")

            try:
                cPickle.dump((address,message),self.write_stdin,1)
                self.write_stdin.flush()
            except:
                self.become_broken()
        finally:
            self.lock.release()

    def stop(self):
        self.lock.acquire()
        try:
            self.running = 0
            cPickle.dump(None,self.write_stdin,1)
            self.write_stdin.close()
            self.read_stdout.close()
        finally:
            self.lock.release()

    def become_broken(self):
        self.lock.acquire()
        try:
            if not self.broken and self.running:
                error = Error('The connection to the firewall has gone down, or the proxy has crashed.\n\n'+\
                                            'Circle is now disconnected from the network. Sorry.')
                utility.mainthread_call(error.show)
            self.broken = 1

            raise Error('proxy broken')
        finally:
            self.lock.release()

if __name__ == '__main__':
     Proxy('localhost').stop()


# vim: set expandtab :
