# Searching

#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  Paul Francis Harrison
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Searching for various things in the hashtable
"""

# TODO
#   "availabe from n sources including njh"
#   "trust person x's taste music"

import sys,os,string,types

from settings import *
from utility import *
from name_server import *
from file_server import *
from error import Error
import safe_pickle, crypto.RSA, hash

if use_gtk:
    from widgets import *
    from gtk import *
    #import GDK #, GtkExtra

def search_browse_task(pipe,node,address,path):
    """Note: there can be duplicates written to pipe."""
    position = 0
    while pipe.running:
        try:
            names = node.call(address,('files available',position,position+20,path))
        except Error:
            return

        for name in names:
            if type(name) == types.StringType:
                Task(search_subsubtask,pipe,node,name,address,None).start()
            else:
                pipe.write((address,name))

        if len(names) < 20:
            break

        position = position + len(names)

def search_subsubtask(pipe,node,search_for,link,source):
    """Note: there can be duplicates written to pipe."""
    if link == None:
        link = source

    data_position = 0
    while pipe.running:
        try:
            # TODO: Why request only one item per packet?
            data = node.call(link,('query data',search_for,data_position,data_position+1) )
            # TODO: Check that data is of sequence type.
        except Error:
            break

        for datum in data:
            pipe.write((link, datum))

        if len(data) < 1:
            break

        data_position = data_position + len(data)

# these import lines should really be hidden in class Player
if use_gtk:
    import mp3
    import textviewer

    try:
        import imageviewer
    except:
        imageviewer = None

class Players:
    def __init__(self, filename):
        self.players = {}

        if sys.platform == 'win32':
            return
    
        extension = string.split(filename, ".")[-1]
        lext = string.lower(extension)

        if lext in ['mp3','ogg']:
            self.players['Play'] = mp3.musicplayer

        if lext in ['jpg','jpeg','gif','png','xpm','bmp']:
            if imageviewer:
                self.players['View'] = imageviewer.imageviewer

        if lext == 'txt':
            self.players['Read'] = textviewer.textviewer
            
        #self.playable = len(self.players.items()) != 0

    def actions(self):
        return self.players.keys()
        
    def play(self, action, filename, size, keep_file_checkbox):
        return self.players[action](filename, size, keep_file_checkbox)
        
        #if len(self.players.keys()) == 1:
        #  return self.players.values()[0](f, s)
        #else: # display a list of players
        #  pass

def search_browse_files(address, node,app):
    pipe = Pipe()
    pipe.start(search_browse_task,node,address, [ ])

    Searcher('Files',pipe,[ ], node,app).start()

def search_for_files(str, node,app):
    temp_list = string.split(string.lower(str))
    list = [ ]
    for item in temp_list:
        if len(item) >= min_search_keyword_len:
            list.append(item)

    if len(list) == 0:
        raise Error(_("Please specify some keywords to search for.\n" +\
                                "(keywords must contain at least %d letters)") % min_search_keyword_len)

    largest = list[0]
    for item in list[1:]:
        if len(item) > len(largest):
            largest = item

    #pipe = Pipe()
    #pipe.start(search_task,node, hash.hash_of(largest))
    pipe = node.retrieve(hash.hash_of(largest))
    
    if app.config['use_donkey']:
        app.donkey.retrieve(largest,pipe)
    if app.config['use_overnet']:
        app.overnet.retrieve(largest,pipe)
    Searcher(_('Files matching: ')+string.join(list),pipe,list, node,app).start()
    #todo : use several pipes in the searcher

def search_for_name(name,title, node,app):
    pipe = node.retrieve(name)

    Searcher(title,pipe,[ ], node,app).start()

def search_for_people(str, node,app):
    list = string.split(string.lower(str))
    
    if len(list) == 0:
        raise Error(_("Please specify some keywords to search for.\n" +\
                                "For example, their name or username."))
    
    largest = list[0]
    for item in list[1:]:
        if len(item) > len(largest):
            largest = item
    
    #pipe = Pipe()
    #pipe.start(search_task,node, hash.hash_of('identity-name '+largest))
    pipe = node.retrieve(hash.hash_of('identity-name '+largest), identity_redundancy)
    
    Searcher(_('People matching: ')+string.join(list),pipe,list, node,app).start()

def search_for_all_people(node,app):
    pipe = node.retrieve(hash.hash_of('service identity'), identity_redundancy)
    
    Searcher(_('All people'),pipe,[ ], node,app).start()

def search_for_all_channels(node,app):
    pipe = node.retrieve(hash.hash_of('channel exists'), channel_redundancy)
    
    Searcher(_('All channels'),pipe,[ ], node,app).start()

def search_for_people_on_channel(channel, node,app):
    pipe = node.retrieve(hash.hash_of('channel subscribe '+channel), channel_redundancy)
    
    Searcher(_('People on %s')%channel,pipe,[ ], node,app).start()

def search_address_list_task(pipe, list,node):
    for item in list:
        Task(search_address_list_subtask,pipe,item,node).start()

def search_address_list_subtask(pipe, item,node):
    try:
        result = node.call(item,('identity query',))
        pipe.write((item,result))
    except Error, error:
        pipe.write((item,{ \
            'type' : 'daemonic node', \
            'name' : hostname_of_ip(item[0]) + ':%d' % item[1], \
            'keywords' : [ ] \
        }))

def search_show_sources(source_list, node,app, title='Sources'):
    pipe = Pipe()
    pipe.start(search_address_list_task,source_list,node)
     
    Searcher(title,pipe,[ ], node,app).start()

def check_and_demangle_item(item):
    """Return demangled item if it is a valid mangled item,
       otherwise throw an exception (not necessarily of class Error).

       Ensures: either an exception is thrown or ret is a dictionary
       whose keys include 'type'."""

    item_type = item['type']
    if item_type == 'identity offline demangled':
        raise Error('impostor')
    
    elif item_type == 'identity offline':
        demangle = safe_pickle.loads(item['package'])
        if not crypto.RSA.construct(demangle['key']).verify(hash.hash_of(item['package']),item['signature']):
            raise Error('bad signature')

        demangle['type'] = 'identity offline demangled'
        return demangle

    else:
        return item


# Search tree classes

# All such classes must provide "get_children", "get_text", "get_description", "get_actions", "get_widget", "get_comparator", "update" and "stop"

class Search_tree_object:
    def __init__(self):
        pass

    def stop(self):
        pass

    def get_children(self):
        return None

    def update(self, node):
        return 0

    def get_description(self):
        return ''

    def get_actions(self,app):
        return [ ]

    def get_widget(self):
        return None

    def get_comparator(self):
        return (-1e10, 0, string.lower(self.text)) 

    def matches(self, item,address):
        return 0

class Search_tree_interior(Search_tree_object):
    def __init__(self, text, pipe_getter, keywords, empty_text = ' (empty)'):
        Search_tree_object.__init__(self)

        self.sources = None #thomas

        self.text = text
        self.empty_text = empty_text
        self.pipe_getter = pipe_getter
        self.pipe = None
        self.keywords = keywords
        self.children = [ ]
        self.still_searching = 1

    def get_children(self):
        return self.children 

    def get_text(self):
        if self.still_searching and self.pipe:
            return self.text + ' ...'
        elif not self.still_searching and self.children == [ ]:
            return self.text + self.empty_text
        else:
            return self.text

    def get_comparator(self):
        return (-1e10, 0, string.lower(self.text)) 

    def update(self, node):
        any = 0

        if not self.pipe:
            self.pipe = self.pipe_getter()
            any = 1

        if self.pipe.finished() and self.still_searching:
            self.still_searching = 0
            return 1

        list = self.pipe.read_all()
        list.sort()  # 

        prev_pair = None
        for pair in list:
            if pair == prev_pair:
                continue
            link, item = prev_pair = pair

            try:
                item = check_and_demangle_item(item)
            except:
                continue

            keywords = item.get('keywords',[ ])

            if type(keywords) <> types.ListType:
                continue

            bad = 0
            for key in self.keywords:
                if key not in keywords:
                    bad = 1
            if bad:
                continue

            any = 1

            if item['type'] == 'directory':
                def pipe_getter(path=item['path'],node=node,link=link):
                    pipe = Pipe()
                    pipe.start(search_browse_task,node,link,path)
                    return pipe

                new_interior = Search_tree_interior(item['filename'], pipe_getter,[ ])
                self.children.append( new_interior )
            else:
                for child in self.children:
                    if child.matches(item,link):
                        child.add_source(item,link)
                        break
                else: 
                    try:
                        if not search_item_table.has_key(item['type']):
                            raise Error(_('unknown type'))

                        self.children.append( search_item_table[item['type']](item,link) )
                    except Error:
                        pass

        return any

    def stop(self):
        Search_tree_object.stop(self)

        if self.pipe:
            self.pipe.stop(1)

class Search_tree_leaf(Search_tree_object):
    def __init__(self, item,source):
        Search_tree_object.__init__(self)

        self.item    = item
        self.sources = [ source ]

        self.widget  = None

        self._setup()

    def get_text(self):
        return self.title

    def get_comparator(self):
        return (-len(self.sources), 0, string.lower(self.title))

    def get_description(self):
        return self.description

    def get_widget(self):
        return self.widget

    def add_source(self,item,address):
        if address in self.sources:
            return 0

        self.sources.append(address)

        return 1
 
    def browse(self, app):
        search_browse_files(self.sources[0], app.node,app)

class Search_tree_identity(Search_tree_leaf):
    def __init__(self, item,source):
        self.address = None

        Search_tree_leaf.__init__(self, item,source)

    def _setup(self):
        if self.item['type'] == 'identity':
            self.address = self.sources[0]

        hash = key_name(self.item['key'])

        #self.name = self.item.get('human-name','')
        #if not self.name:
        #  self.name = self.item['name']
        self.name = self.item['name'] + ' (' + self.item.get('human-name','') + ')'
        
        self.widget = Signature_widget(hash)
        # Proof of @R43: hash assigned from key_name; @E25; and not shared
        # with any other thread or written to.
        
        self.update_description()

    def update_description(self):
        self.title = self.name
        self.description = _('Name: ') + self.item.get('human-name','') + \
                                             '\n  '+_('Username: ') + self.item['name']
                                             
        if self.address:
            self.title = self.title + _(' (online)')
            self.description = self.description + \
                                             '\n  '+_('Location: ')+ self.address[0]+':'+repr(self.address[1]) 

    def get_comparator(self):
        if self.address == None:
            value = 0
        else:
            value = -1
        return (value, 0, string.lower(self.title))

    def add_source(self,item,address):
        if not Search_tree_leaf.add_source(self,item,address):
            return 0

        if item['type'] == 'identity':
            self.address = address
            self.update_description()
        return 1

    def get_actions(self,app):
        if app.name_server.acquaintances.has_key(key_name(self.item['key'])):
            str = _("View")
        else:
            str = _("Add to contact list")
        button1 = GtkButton(str)
        button1.connect("clicked",\
            lambda _b, _self=self,_app=app: _app.name_server.edit_acquaintance(_self.item,_self.address))

        if not self.address:
            return [ button1 ]

        button2 = GtkButton(_("Browse their files"))
        button2.connect("clicked",\
            lambda _b, _self=self,_app=app: _self.browse(_app))

        return [ button1, button2 ]

    def matches(self, item,address):
        return item['type'] in ['identity','identity offline demangled'] and self.item['key'] == item['key']

class Search_tree_daemonic_node(Search_tree_leaf):
    def _setup(self):
        self.title = _('Daemon at %s')%self.item['name']
        self.description = self.title

    def get_comparator(self):
        return (-len(self.sources), 1, string.lower(self.title))
    
    def get_actions(self,app):
        button1 = GtkButton(_("Browse their files"))
        button1.connect("clicked",\
            lambda _b, _self=self,_app=app: _self.browse(_app))

        return [ button1 ]
 
class Search_tree_channel_exists(Search_tree_leaf):
    def _setup(self):
        self.title = self.item['name']
        self.description = _("1 subscriber on-line.")

    def add_source(self, item,address):
        if not Search_tree_leaf.add_source(self,item,address):
            return 0

        self.description = _("%d subscribers on-line.") % len(self.sources)
        return 1

    def get_actions(self,app):
        if app.channels.list.has_key(self.item['name']):
            button1 = GtkButton(_("Unsubscribe"))
            button1.connect("clicked",\
                lambda _b, _self=self,_app=app:
                    _app.channels.unsub(_self.item['name']))
        else:
            button1 = GtkButton(_("Subscribe"))
            button1.connect("clicked",\
                lambda _b, _self=self,_app=app:
                    _app.channels.sub(_self.item['name']))

        button2 = GtkButton(_("List subscribers"))
        button2.connect("clicked",\
            lambda _b, _self=self,_app=app: search_for_people_on_channel(_self.item['name'],_app.node,_app))

        return [ button1, button2 ]
 
    def matches(self, item,address):
        return self.item['type'] == 'channel exists' and self.item['name'] == item['name']

class Search_tree_file(Search_tree_leaf):
    def _setup(self):
        if not self.item.has_key('filename') or \
             not self.item.has_key('length') or \
             not self.item.has_key('name'):
            raise Error(_('Bad datum'))

        if self.item.get('music_title'):
            self.title = self.item['music_title']
            if self.item.get('music_artist'):
                self.title = self.title + _(' by ') + self.item['music_artist']
        else:
            self.title = self.item['filename']

        self.player = Players(self.item['filename'])
        #self.is_music = self.player.playable

        self.title = self.title + "  (%s)" % human_size(self.item['length'])
        #print self.title
        
        keys = [('filename',''),('music_title',_('   Title: ')),('music_album',_('   Album: ')),('music_artist',_('   Artist: '))]
        self.description = ''
        for key in keys:
            if self.item.has_key(key[0]):
                self.description = self.description+key[1]+self.item[key[0]]+'\n'
        self.description_template = self.description

        if self.sources == [ 'eDonkey' ]:
            self.description = self.description + _('File available from eDonkey.')
        elif self.sources == [ 'Overnet' ]:
            self.description = self.description + _('File available from Overnet.')
        else:
            self.description = self.description + _('File available from 1 source.')
            self.widget = GtkHBox(FALSE,0)
            self.widget.pack_start(GtkLabel(_('URL: ')), FALSE,FALSE,0)
            entry = GtkEntry()
            entry.set_text(hash.hash_to_url(self.item['name']))
            self.widget.pack_start(entry, TRUE,TRUE,0) 
            self.widget.show_all()

    def add_source(self, item,address):
        if not Search_tree_leaf.add_source(self,item,address):
            return 0

        self.description = self.description_template + \
            (_("File available from %d sources.") % len(self.sources))
        return 1
 
    def get_actions(self,app):
        button1 = GtkButton(_("Download"))
        button1.connect("clicked",\
            lambda _b, _self=self,_app=app: _self.download(_app))

        buttons = [button1]
        if self.sources <> [ 'eDonkey' ] and self.sources <> [ 'Overnet' ]:
            button2 = GtkButton(_("View sources"))
            button2.connect("clicked",\
               lambda _b, _self=self,_app=app: _self.view_sources(_app))
            buttons.insert(0,button2)
            for item in self.player.actions():
                button = GtkButton(item)
                button.connect("clicked", lambda b, self=self,app=app,item=item: self.play(item,app))
                buttons.insert(0,button)

        return buttons

    def matches(self, item,address):
        return self.item['name'] == item['name']

    def download(self, app):
        try:
            #downloader = Downloader(app.node,self.item,self.sources,app)
            #downloader.start()
            app.download_manager.download(self.item,self.sources)
        except Error, error:
            error.show()

    def view_sources(self, app):
        search_show_sources(self.sources, app.node,app)

    def play(self, action, app):
        app.file_server.lock.acquire()
        try:
            if app.file_server.paths.has_key(self.item['name']):
                try:
                    player = self.player.play(action, app.file_server.paths[self.item['name']][0], self.item['length'],0)
                except Error, error:
                    error.show()
                # and don't show checkbox...
                return
        finally:
            app.file_server.lock.release()
        
        # on download completion: note need to unlink
        download_complete = [ 0 ]
        def note_completion(download_complete=download_complete):
            download_complete[0] = 1
        try:
            downloader = Circle_Downloader(app.node,self.item,self.sources,app.config['download'],note_completion)
            downloader.start()
        except Error, error:
            error.show()
            return

        # TODO: wait until enough is buffered NICELY
        # the problem is that different media have different ideas of what
        # this means.  for mp3s a possible strategy is to download a few
        # chunks, then estimate how long the whole file will take.  Image
        # files are less satisfactory.  progressive image formats probably
        # should compare download times with display times.  other formats
        # may be able to partially construct images as the data arrives.
        # Multistream formats such as quicktime need to do their own
        # buffering.  Tar files probably should get the whole file.
        
        # time.sleep(1)  All the players now provide a better method of gathering files.

        #player = mp3player(downloader.filename)
        try:
            player = self.player.play(action, downloader.filename, self.item['length'], 1)
        except Error, error:
            error.show()
            if downloader.running:
                downloader.stop()
            try:
                os.unlink(downloader.filename)
            except OSError:
                pass
            return
        
        # on player close: unlink file (maybe)
        def on_close(window,player=player,downloader=downloader,download_complete=download_complete,filename=downloader.filename):
            # Note: this only works because downloader stops in the gtk thread
            if downloader.running:
                downloader.stop()
            if download_complete[0] and not player.keep_file_button.get_active():
                try:
                    os.unlink(filename)
                except OSError:
                    pass

        player.connect("destroy",on_close)

search_item_table = {
    'identity' : Search_tree_identity,
    'identity offline demangled' : Search_tree_identity,
    'daemonic node' : Search_tree_daemonic_node,
    'channel exists' : Search_tree_channel_exists,
    'file' : Search_tree_file
}

class Searcher:
    def __init__(self, title,pipe,keywords, node,app):
        self.title = title
        self.pipe  = pipe
        self.list  = keywords
        self.node  = node
        self.app   = app
        #self.results = [ ] #thomas
        self.current_selection = None
        self.actions = [ ]

        self.running = 0

    def start(self):
        self.running = 1

        self.window = GtkWindow(WINDOW_TOPLEVEL) 
        self.window.set_default_size(400,400)
        #self.window.set_border_width(10)        
        self.window.set_title(self.title)
        self.window.connect("destroy",\
            lambda _b, _self=self: _self.stop())

        vbox = GtkVBox(FALSE,5)
        self.window.add(vbox)

        #self.label = GtkLabel('Searching...')
        #self.label.set_alignment(0,0)
        #vbox.pack_start(self.label, FALSE,FALSE,0)

        def on_selection_change(new_selection, self=self):
            self.bin.foreach(lambda c, self=self: self.bin.remove(c))

            for item in self.actions:
                self.hbox.remove(item)
            self.actions = [ ]

            if not new_selection:
                self.description_label.set_text('')
                self.vbox_2.queue_resize()
                return

            self.actions = new_selection[-1].get_actions(self.app)
            for item in self.actions:
                item.set_flags(CAN_DEFAULT)
                self.hbox.pack_start(item, TRUE,TRUE,0)
            self.hbox.show_all()
            
            if self.actions:
                self.actions[0].grab_default()

            self.description_label.set_text(new_selection[-1].get_description())
            widget = new_selection[-1].get_widget()
            if widget:
                self.bin.add(widget)
                widget.show()

            self.vbox_2.queue_resize()

        def child_getter(path):
            children = path[-1].get_children()
            if not children:
                return children

            children = children[:]
            children.sort(lambda x,y: cmp(x.get_comparator(),y.get_comparator()))
            return children

        self.root = Search_tree_interior(self.title, lambda self=self: self.pipe, self.list,_(': couldn\'t find anything, sorry'))
        self.tree_widget = Tree_widget([self.root],child_getter,lambda x: x.get_text(),on_selection_change)

        vbox.pack_start(self.tree_widget.widget, TRUE,TRUE,0)

        #def on_select(_b, row,column,event, _self=self): 
        #  _self.current_selection = row
        #  str = _self.results[row].description
        #  _self.description_label.set_text(str)
# 
#      _self.bin.foreach(lambda c, _self=_self: _self.bin.remove(c))
#      if _self.results[row].widget:
#        _self.bin.add(_self.results[row].widget)
#        _self.results[row].widget.show()
#
#      for item in _self.actions:
#        _self.hbox.remove(item)
#      _self.actions = _self.results[row].get_action_bar(_self.app)
#      for item in _self.actions:
#        _self.hbox.pack_start(item, TRUE,TRUE,0)
#      _self.hbox.show_all()
#      
#      if event and event.type == GDK._2BUTTON_PRESS:
#        if len(_self.actions):
#          _self.actions[0].emit("clicked",None)
#
#    scrolly = GtkScrolledWindow()
#    scrolly.set_policy(POLICY_NEVER,POLICY_AUTOMATIC)
#    vbox.pack_start(scrolly)
#    
#    self.list_box = GtkCList(1)
#    self.list_box.set_selection_mode(SELECTION_BROWSE)
#    self.list_box.connect("select-row", on_select)
#    #self.list_box.connect("unselect-row", on_unselect)
#    scrolly.add(self.list_box)

        self.vbox_2 = GtkVBox(FALSE, 5)
        self.vbox_2.set_border_width(10)
        vbox.pack_start(self.vbox_2, FALSE,FALSE,0)
        
        self.description_label = GtkLabel('')
        self.description_label.set_justify(JUSTIFY_LEFT)
        self.description_label.set_alignment(0,0)
        self.vbox_2.pack_start(self.description_label, FALSE,FALSE,0)

        self.bin = GtkAlignment(1,1,0.5,1)
        self.vbox_2.pack_start(self.bin,FALSE,FALSE,0)

        # Obscure GTK sizing bug workaround
        self.alignment = GtkAlignment(0,0,1,1)
        self.vbox_2.pack_end(self.alignment, FALSE,FALSE,0)
        self.hbox = GtkHBox(FALSE,5)
        self.alignment.add(self.hbox)

        self.close_button = GtkButton(_("Close"))
        self.close_button.connect("clicked",\
            lambda _b, _window=self.window: _window.destroy())
        self.close_button.set_flags(CAN_DEFAULT)
        self.hbox.pack_end(self.close_button, TRUE,TRUE,0)
        
        #button = GtkButton("View sources")
        #button.connect("clicked",\
        #  lambda _b, _self=self: _self.view_sources())
        #hbox.pack_end(button, TRUE,TRUE,0)

        #button = GtkButton("Download")
        #button.connect("clicked",\
        #  lambda _b, _self=self: _self.download())
        #hbox.pack_end(button, TRUE,TRUE,0)

        show_window(self.window, 'search')
    
        self.update(250.0)
        #utility.schedule_mainthread(250.0, self.update)
        #timeout_add(250,self.update)

    def stop(self):
        #self.pipe.stop(1)

        list = [ self.root ]
        while list:
            children = list[0].get_children()
            if children:
                list.extend(children)
            list[0].stop()
            del list[0]

        self.running = 0
 
    def update(self, last_time):
        assert(type(last_time) == types.FloatType)
        if not self.running:
            return

        any = 0
        for item in self.tree_widget.current_path:
            any = item.update(self.node) or any

        if any:
            self.tree_widget.refresh_display()

        new_time = last_time * 5.0 / 4.0
        if new_time > 10000.0:
            new_time = 10000.0
        utility.schedule_mainthread(new_time, self.update, new_time)
        
        #return self.running

    #  for item in self.pipe.read_all():
    #    self.insert(item[0],item[1])

    #  finished = self.pipe.finished()

    #  if finished:
    #    self.label.set_text('Search results:')

    #  return not finished

    #def insert(self, link,item):
    #  if type(item) <> types.DictType or \
    #     not item.has_key('type') or \
    #     not item.has_key('keywords') or \
    #     type(item['keywords']) <> types.ListType:
    #    return
#
#    for key in self.list:
#      if key not in item['keywords']:
#        return
#    
#    for result in self.results:
#      if result.matches(item,link):
#        result.add_source(link)
#       return
#    
#    new_result = Search_result(item,link)
#  
#    self.results.append(new_result)
#    self.list_box.append([new_result.title])


# vim: set expandtab :
