# Custom GTK widgets

#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  Paul Francis Harrison
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# TODO: feedback on link click

import sys
import string, re, os, os.path
from gtk import *

import check
from error import Error
import hash, settings
from utility import browse_url, show_window

# Create a tooltip for general use

tooltips = GtkTooltips()
tooltips.set_delay(1000)
tooltips.enable()

# Create some custom styles

#def GdkColor_fixed(a,b,c):
#  try:
#    return GdkColor(a,b,c)
#  except:
#    # Some versions of gtk expect a signed short (duh)
#    if a > 32767: a = a - 65536
#    if b > 32767: b = b - 65536
#    if c > 32767: c = c - 65536
#    return GdkColor(a,b,c)

w = GtkWindow()
colormap = w.get_colormap()

black      = colormap.alloc(0,0,0)
grey       = colormap.alloc(24500,24500,24500)
light_grey = colormap.alloc(41000,41000,41000)
blue       = colormap.alloc(0,0,65535)
purple     = colormap.alloc(41000,0,41000)
green      = colormap.alloc(0,41000,0) #(0,32767,0)
red        = colormap.alloc(41000,0,0) 
orange     = colormap.alloc(65535,24500,0) 
white      = colormap.alloc(65535,65535,65535) 

w.ensure_style()
text_widget_style = w.get_style().copy()
text_widget_style.fg[0] = black
text_widget_style.fg[1] = blue
text_widget_style.fg[2] = purple

del colormap
del w

# Block-of-text widget

class Text_widget:
    def __init__(self, width, heading=0):
        #self.widget   = GtkDrawingArea() 
        self.widget   = GtkEventBox()
        self.drawing_area = GtkDrawingArea() 
        self.widget.add(self.drawing_area)

        self.width    = width
        self.heading  = heading

        #self.drawing_area.ensure_style()
        #self.style = self.drawing_area.get_style().copy()
        #self.style.fg[0] = black
        #self.style.fg[1] = blue
        self.style = text_widget_style
        self.drawing_area.set_style(self.style)
        self.font  = self.style.font

        self.space_width = self.font.width(' ')
        self.font_height = self.font.ascent + self.font.descent
        
        self.drawing_area.connect("expose_event",self.on_expose)
        self.drawing_area.connect("button-release-event",self.on_click)
        self.drawing_area.connect("configure_event",self.configure);
        self.drawing_area.set_events(GDK.BUTTON_RELEASE_MASK)

        self.clear()

    def clear(self):
        self.words    = [ ]
        self.position = (0,1)
        self.height   = self.font_height + 2

        self.actions  = [ ]

        self.drawing_area.size(self.width,self.height)

    def write_space(self):
        self.position = (self.position[0] + self.space_width, self.position[1])

    def write_tab(self):
        self.position = (self.position[0] + self.space_width*settings.tab_size, self.position[1])

    def write_newline(self):
        self.position = (0, self.height)
        self.height   = self.height + self.font_height + 1
        self.drawing_area.size(self.width,self.height)

    def write_word(self, string, color,action, words_for_action=None):
        width = self.font.width(string)

        if width + self.position[0] >= self.width:
            self.write_newline()

        leftover = ''
        while width > self.width and len(string) > 1 :
            leftover = string[-1] + leftover
            string = string[:-1]
            width = self.font.width(string)

        if action:
            word = [self.position[0],self.position[1],color,string]
            self.words.append(word)
            if words_for_action == None:
                words_for_action = [ ]
            words_for_action.append(word)
            self.actions.append( (self.position[0],
                                  self.position[1],
                                  self.position[0]+width,
                                  self.position[1]+self.font_height,
                                  action,words_for_action) )
        else: 
            self.words.append((self.position[0],self.position[1],color,string))

        self.position = (self.position[0] + width, self.position[1])

        if leftover:
            self.write_word(leftover, color,action, words_for_action)

    def write(self, text, app=None):
        symbols = re.split(r'( |\n|\t)', text)

        for symbol in symbols:
            stripped_symbol = symbol
            while stripped_symbol and stripped_symbol[0] in '([{<':
                stripped_symbol = stripped_symbol[1:]
            while stripped_symbol and stripped_symbol[-1] in ')]}>.,!?':
                stripped_symbol = stripped_symbol[:-1]
        
            if symbol == ' ':
                self.write_space()
            elif symbol == '\t':
                self.write_tab()
            elif symbol == '\n':
                self.write_newline()
            else:
                if stripped_symbol[:5] == 'http:' or stripped_symbol[:4] == 'ftp:' or stripped_symbol[:4] == 'www.':
                    color = 1

                    def action(symbol=stripped_symbol):
                        try:
                            browse_url(symbol)
                        except Error,error:
                            error.show()
                        return 2 # change to color purple
                    #action = lambda symbol=symbol: browse_url(symbol)
                elif stripped_symbol[:12] == 'circle-file:' and app:
                    color = 1

                    def action(symbol=stripped_symbol,app=app):
                        try:
                            import search
                            search.search_for_name(hash.url_to_hash(symbol),'Files matching url', app.node,app)
                        except Error,error:
                            error.show()
                        return 2 # change color to purple
                else:
                    color = 0
                    action = None
                self.write_word(symbol, color,action)

        # queue redraw?

    def on_expose(self, drawing_area,event):
        if self.heading:
            drawing_area.draw_rectangle(self.style.bg_gc[1],1,0,0,self.width,self.height)
        else:
            drawing_area.draw_rectangle(self.style.bg_gc[0],1,0,0,self.width,self.height)

        for word in self.words:
            drawing_area.draw_text(self.font,
                                   self.style.fg_gc[word[2]],
                                   word[0],word[1]+self.font.ascent,word[3])

    def on_click(self, drawing_area,event):
        if event.button <> 1:
            return

        for item in self.actions:
            if event.x >= item[0] and event.y >= item[1] and \
                 event.x <  item[2] and event.y <  item[3]:
                new_color = apply(item[4],())
                if new_color <> None:
                    for word in item[5]:
                        word[2] = new_color
                    drawing_area.queue_draw()
    
    def configure(self, drawing_area, event):
        # this should really cause the widget to redraw itself.
        # - set width
        # - re-layout
        # - requisition required height

        #drawable = self.widget.get_window()
        #self.width, height = drawable.width, drawable.height
        pass

# Human hashable representation of a name
        
def mid_point(a,b):
    return ((a[0]+b[0])/2.0,(a[1]+b[1])/2.0)

def bezier_curve(*list):
    for i in range(6):
        new_list = [ list[0] ]

        for j in range(0,len(list)-1):
            new_list.append( mid_point(list[j],list[j+1]) )

        new_list.append(list[-1])
        list = new_list

    return list

def quadratic(p0, p1, p2):
        x0, y0 = p0
        x1, y1 = p1
        x2, y2 = p2

        return "%f %f lineto %f %f %f %f %f %f curveto\n" % (x0, y0,
                              x0 + 2*(x1 - x0)/3.0,
                              y0 + 2*(y1 - y0)/3.0,
                              x1 + 1*(x2 - x1)/3.0,
                              y1 + 1*(y2 - y1)/3.0,
                              x2, y2)
    
class Signature_widget(GtkButton):
    height = 100
    width  = height * 3

    # FIXME: In python2.1, this results in 6.  In future python versions
    # it will result in 6.25, which means signatures look different
    # depending on python version, which reduces the efficacy of using
    # signatures for authentication.  One fix is to force to integer
    # division (explicit floor operation); though from the code it would
    # appear that 6.25 is the value that we actually want here.
    # Same comments apply to on_click below.  (Thould these be using
    # common code?)

    # [pfh] added int( ), will maintain backward compatability

    step   = int( (width-height) / (settings.name_bytes*2) )
    
    def __init__(self, name):
        check.check_is_name(name)  #=@R43
        
        GtkButton.__init__(self)
        self.connect_after("expose-event",self.on_expose)
        self.connect_after("draw",self.on_expose)
        #self.add_events(GDK.BUTTON_PRESS_MASK)
        #self.connect("button_press_event",self.on_click, name)
        self.set_relief(RELIEF_NONE)
        self.connect("clicked",self.on_click, name)
        self.set_usize(self.width,self.height)

        if sys.platform == 'win32':
            self.set_sensitive(FALSE)

        points = [ ]
        for i in range(settings.name_bytes*2):
            if i & 1 :
                x = ord(name[i >> 1]) >> 4
            else:
                x = ord(name[i >> 1]) & 0xf
            a = ([0.0, 1.25, 1.75, 3.0][x & 3]+1)/5.0
            b = ([0.0, 2.0, 2.5, 3.0][x >> 2]+1)/5.0
            points.append( (i*self.step + a*self.height,b*self.height) )

        self.lines = bezier_curve(points[0],points[1],mid_point(points[1],points[2]))[:-1]

        for i in range(1,len(points)-1):
            self.lines.extend(bezier_curve(mid_point(points[i-1],points[i]),points[i],mid_point(points[i],points[i+1]))[:-1])
        
        self.lines.extend(bezier_curve(mid_point(points[-3],points[-2]),points[-2],points[-1]))
    
    def on_expose(self, widget,event=None):
        draw_lines(widget.get_window(),widget.get_style().fg_gc[0], self.lines)

    def on_click(self, widget, name):
        height = 100
        width  = height * 3
        
        step   = int( (width-height) / (settings.name_bytes*2) )
        
        fname = os.tmpnam() # because we want to run gv on the file,
        # we need to manage the file ourselves.  This could open us to
        # a symlink vuln.
        
        f = open(fname, "wt")
        
        f.write("""%%!PS-Adobe-2.0 EPSF-2.0
%%%%Title: signature
%%%%Creator: Circle
%%%%CreationDate: 
%%%%For: 
%%%%Orientation: Portrait
%%%%BoundingBox: 0 0 %f %f
%%%%Pages: 0
%%%%BeginSetup
%%%%EndSetup
%%%%Magnification: 1.0000
%%%%EndComments

/Times-Roman findfont
10 scalefont
setfont

newpath
2 2 moveto
(circle-person:%s) show

0 %f translate
1 -1 scale

newpath
""" % (width, height, string.join(map(lambda a:hex(ord(a))[2:], name), ''), height))
        
        points = [ ]
        for i in range(settings.name_bytes*2):
                if i & 1 :
                        x = ord(name[i >> 1]) >> 4
                else:
                        x = ord(name[i >> 1]) & 0xf
                a = ([0.0, 1.25, 1.75, 3.0][x & 3]+1)/5.0
                b = ([0.0, 2.0, 2.5, 3.0][x >> 2]+1)/5.0
                points.append( (i*self.step + a*self.height,b*self.height) )
        
        f.write("%f %f moveto\n" % (points[0][0], points[0][1]))
        f.write(quadratic(points[0],points[1],mid_point(points[1],points[2])))

        for i in range(1,len(points)-1):
            f.write(quadratic(mid_point(points[i-1],points[i]),points[i],mid_point(points[i],points[i+1])))
                
        f.write(quadratic(mid_point(points[-3],points[-2]),points[-2],points[-1]))
        f.write("stroke")
        f.close()
        
        try:
            os.system("( gv %s ; rm %s ) &" % (fname,fname))
        except OSError:
            pass

# Title label with drop-down help

class Helpful_label_widget:
    def __init__(self, title,help,width=300):
        self.widget = GtkHBox(FALSE,0)
        self.text_packed  = 0
        self.text_visible = 0

        vbox = GtkVBox(FALSE,0)
        self.widget.pack_start(vbox, TRUE,TRUE,0)

        if type(title) == type(''):
            label = GtkLabel(title)
            label.set_alignment(0,0)
        else:
            # It's a widget
            label = title
        vbox.pack_start(label, TRUE,TRUE,0)

        text = Text_widget(width,1)
        text.write(help)

        alignment = GtkAlignment(1,0,0,1)
        alignment.add(text.widget)

        button = GtkButton()
        button.set_relief(RELIEF_NONE)
        tooltips.set_tip(button,"help","")
        self.widget.pack_start(button, FALSE,FALSE,0)

        arrow = GtkArrow(ARROW_DOWN,SHADOW_IN)
        arrow.set_alignment(0,1)
        button.add(arrow)

        def on_click(button, self=self,alignment=alignment,vbox=vbox,arrow=arrow):
            if not self.text_packed:
                vbox.pack_start(alignment, TRUE,TRUE,0)
                self.text_packed = 1
            if not self.text_visible:
                alignment.show_all()
                self.text_visible = 1
                arrow.set(ARROW_UP,SHADOW_IN)
            else:
                alignment.hide()
                self.text_visible = 0
                arrow.set(ARROW_DOWN,SHADOW_IN)
            self.widget.queue_resize()
        button.connect("clicked",on_click)


# A widget for navigation a directory and selecting files

class Tree_widget:
    def __init__(self, initial_path,query_func,repr_func,on_selection_change=None):
        self.query_func = query_func
        self.repr_func = repr_func
        self.on_selection_change = on_selection_change

        self.current_path = initial_path
        self.is_selected = 0
        self.paths = [ ]
        self.in_refresh = 0

        self.widget = GtkScrolledWindow()
        self.widget.set_policy(POLICY_NEVER,POLICY_AUTOMATIC)
        
        self.list_widget = GtkCList()
        self.list_widget.set_selection_mode(SELECTION_SINGLE)
        self.widget.add(self.list_widget)

        self.refresh_display()

        def on_select(list, index,column,event, self=self):
            if self.in_refresh:
                return
            self.is_selected = 1
            self.current_path = self.paths[index]
            self.refresh_display()
            if self.on_selection_change:
                self.on_selection_change(self.current_path)
        def on_unselect(list, index,column,event, self=self):
            if self.in_refresh:
                return
            if self.paths[index] == self.current_path:
                self.is_selected = 0
            if self.on_selection_change:
                self.on_selection_change(None)
 
        self.list_widget.connect("select-row",on_select)
        self.list_widget.connect("unselect-row",on_unselect)

    def get_selection(self):
        if not self.is_selected:
            return None
        else:
            return self.current_path

    def refresh_display(self):
        if self.widget.flags(DESTROYED):
            return

        self.in_refresh = 1

        result = self.query_func(self.current_path)
        if result <> None:
            path = self.current_path
        else:
            path = self.current_path[:-1]
            result = self.query_func(path)
            if result == None: result = [ ]

        old_paths = self.paths
        self.paths = [ ]
        items = [ ]
        for i in range(len(path)):
            self.paths.append(path[:i+1])
            items.append( ('  '*i+self.repr_func(path[i]), black) )

        for item in result:
            self.paths.append(path + [item])
            try:
                if item.sources==['eDonkey']:
                    color=green
                elif item.sources==['Overnet']:
                    color=purple
                else:
                    color=black
            except:
                color=black
            items.append(('  '*len(path)+' '+self.repr_func(item),color))

        adjustment = self.list_widget.get_vadjustment()
        if adjustment.lower == adjustment.upper:
            old_point = None
        else:
            old_point =  (adjustment.value-adjustment.lower)/(adjustment.upper-adjustment.lower) 

        self.list_widget.freeze()
        self.list_widget.unselect_all()
        
        if self.is_selected:

            index = old_paths.index(self.current_path)
            try:
                new_index = self.paths.index(self.current_path)
            except:
                new_index = 0

            while index > new_index:
                #Try to avoid losing scroll position
                self.list_widget.append([''])
                self.list_widget.remove(0)
                index = index - 1
            while index < new_index:
                self.list_widget.insert(0, [''])
                index = index + 1
            
            self.list_widget.select_row(index,0)
        
        while self.list_widget.rows < len(self.paths):
            self.list_widget.append([''])
        while self.list_widget.rows > len(self.paths):
            self.list_widget.remove(self.list_widget.rows-1)

        for i in range(len(items)):
            self.list_widget.set_text(i,0,items[i][0])
            self.list_widget.set_foreground(i,items[i][1])


        if old_point <> None:
            adjustment.set_value(old_point*(adjustment.upper-adjustment.lower)+adjustment.lower)

        self.list_widget.thaw()

        self.in_refresh = 0

def File_selector_widget(initial_path, only_directories=0):
    def query_func(path, only_directories=only_directories):
        path = apply(os.path.join,path)
        
        if sys.platform == 'win32' and path == '':
            list = [ ]
            for i in range(24):
                item = chr(ord('C')+i)+':\\'
                if os.path.exists(item):
                    list.append(item)

            return list

        if os.path.isdir(path):
            try:
                list = filter(lambda x: x[:1] <> '.', os.listdir(path))
            except:
                return [ ]
            if only_directories:
                list = filter(lambda x,path=path: os.path.isdir(os.path.join(path,x)), list)
            list.sort()
            return list
        else:
            return None

    initial_path = os.path.abspath(initial_path)
    path = [ ]
    while 1:
        initial_path, name = os.path.split(initial_path)
        if not name:
            break
        path.insert(0,name)
    path.insert(0,initial_path)

    if sys.platform == 'win32':
        path.insert(0,'')
    
    return Tree_widget(path, query_func, lambda x: x or 'My computer')

def File_selector_window(initial_path, only_directories, title,text,ok_action,parent_window=None):
    window = GtkWindow()
    window.set_title(title)
    window.set_default_size(400,500)
    window.set_border_width(10)
    if parent_window:
        window.set_transient_for(parent_window)

    vbox = GtkVBox(FALSE,5)
    window.add(vbox)

    label = GtkLabel(text)
    label.set_alignment(0,0)
    label.set_justify(JUSTIFY_LEFT)
    vbox.pack_start(label, FALSE,FALSE,0)

    selector = File_selector_widget(initial_path,only_directories)
    vbox.pack_start(selector.widget, TRUE,TRUE,5)

    hbox = GtkHBox(FALSE,5)
    vbox.pack_start(hbox, FALSE,FALSE,0)

    button = GtkButton('OK')
    def ok_clicked(b, window=window,selector=selector,ok_action=ok_action):
        path = selector.get_selection() 
        if not path:
            return
        window.destroy()
        if ok_action:
            ok_action(apply(os.path.join,path))
    button.connect("clicked",ok_clicked)
    hbox.pack_start(button, TRUE,TRUE,0)

    button = GtkButton('Cancel')
    button.connect("clicked",lambda b,window=window: window.destroy())
    hbox.pack_start(button, TRUE,TRUE,0)

    return window

class File_entry_widget:
    def __init__(self, title,text,default='', only_directories=0, window=None):
        self.window = window
        self.only_directories = only_directories
        self.title = title
        self.text = text

        self.widget = GtkHBox(FALSE,5)

        self.entry = GtkEntry()
        self.entry.set_text(default)
        self.widget.pack_start(self.entry, TRUE,TRUE,0)

        button = GtkButton(" Select... ")
        self.widget.pack_start(button, FALSE,FALSE,0)

        def on_clicked(b,self=self):
            def on_ok(path,self=self):
                self.entry.set_text(path)
            show_window(File_selector_window('.',self.only_directories, self.title,self.text,on_ok,self.window),'select')
        button.connect("clicked",on_clicked)

    def get_text(self):
        return self.entry.get_text()

# >>>>

def Toggle_icon_widget(char, extra_space=0):
    widget = GtkToggleButton()
    
    widget.ensure_style()
    font = widget.get_style().font

    height = font.ascent+font.descent+4
    unit = int(height/6)
    mid = int((height-1)/2)
    
    width = unit*2+8+extra_space*2

    x1 = 4+extra_space
    x2 = unit+4+extra_space
    x3 = unit*2+4+extra_space
    y1 = mid-unit
    y2 = mid
    y3 = mid+unit

    if char == '>':
        poly = [(x1,y1),(x2,y1),(x3,y2),(x2,y3),(x1,y3),(x2,y2)]
    else:
        poly = [(x2,y1),(x3,y1),(x2,y2),(x3,y3),(x2,y3),(x1,y2)]

    widget.set_usize(width,height)

    #widget.set_usize(font.width(char)+width-1 +4+extra_space*2,font.ascent+font.descent +4)

    #def on_expose(widget, event, width=width,extra_space=extra_space,char=char):
    def on_expose(widget, event, poly=poly):
        style = widget.get_style()

        if widget.get_active():
            gc = style.dark_gc[3]
        else:
            gc = style.dark_gc[0]

        draw_polygon(widget.get_window(),gc,1,poly)
        draw_polygon(widget.get_window(),gc,0,poly)
        #for i in range(width):
        #  draw_text(widget.get_window(),style.font,gc,i+2+extra_space,style.font.ascent+2,char)

    widget.connect_after("expose-event",on_expose)
    widget.connect_after("draw",on_expose)
    return widget

def Text_button_widget(str):
    widget = GtkButton()
    
    widget.ensure_style()
    font = widget.get_style().font

    width = 4

    widget.set_usize(font.width(str) +4,font.ascent+font.descent +4)

    def on_expose(widget, event, width=width,str=str):
        style = widget.get_style()

        draw_text(widget.get_window(),style.font,style.fg_gc[0],2,style.font.ascent+2,str)

    widget.connect_after("expose-event",on_expose)
    widget.connect_after("draw",on_expose)
    return widget

# Here is a simple test we run if the module is run directly

if __name__ == '__main__':
    wind = GtkWindow()
    
    #widget = Text_widget(100,1)
    #widget.widget.set_state(3)
    
    #widget.write("Hello world! What a wonderful world. Supercalafragalisticexpialadocious Yay Yay.\n\nYay!")
    #timeout_add(3000, widget.write, "Hello world! What a wonderful world. Supercalafragalisticexpialadocious Yay Yay.\n\nYay!")

    widget = Signature_widget(hash.hash_of('bob'))
    # Proof of @R43: @E24.
    
    #widget = Helpful_label_widget('Name:','Enter your name here.\n\neg: John Smith')

    #widget = File_selector_widget('/home/pfh',1)

    wind.add(widget)
    #wind.set_usize(400,400)

    #def action(path):
    #  print path

    #wind = File_selector_window('.',0, 'Selct a file','Please',action,None)

    #wind.add(Text_button_widget('hello'))

    wind.show_all()
    wind.connect("destroy", mainquit)
    wind.connect("delete_event", mainquit)
    mainloop()


# vim: expandtab:shiftwidth=4 :
